@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -11,488 +11,126 @@
 #include "a2dp.h"
 
 #include <errno.h>
+#include <limits.h>
 #include <stdbool.h>
 #include <string.h>
+#include <strings.h>
 
 #include <glib.h>
 
-#include "a2dp-codecs.h"
-#include "bluealsa.h"
+#include "a2dp-aac.h"
+#include "a2dp-aptx-hd.h"
+#include "a2dp-aptx.h"
+#include "a2dp-faststream.h"
+#include "a2dp-lc3plus.h"
+#include "a2dp-ldac.h"
+#include "a2dp-mpeg.h"
+#include "a2dp-sbc.h"
+#include "bluealsa-config.h"
+#include "codec-sbc.h"
 #include "hci.h"
-#include "sbc.h"
+#include "shared/a2dp-codecs.h"
 #include "shared/defs.h"
 #include "shared/log.h"
 
-static const a2dp_sbc_t a2dp_sbc = {
-	.frequency =
-		SBC_SAMPLING_FREQ_16000 |
-		SBC_SAMPLING_FREQ_32000 |
-		SBC_SAMPLING_FREQ_44100 |
-		SBC_SAMPLING_FREQ_48000,
-	.channel_mode =
-		SBC_CHANNEL_MODE_MONO |
-		SBC_CHANNEL_MODE_DUAL_CHANNEL |
-		SBC_CHANNEL_MODE_STEREO |
-		SBC_CHANNEL_MODE_JOINT_STEREO,
-	.block_length =
-		SBC_BLOCK_LENGTH_4 |
-		SBC_BLOCK_LENGTH_8 |
-		SBC_BLOCK_LENGTH_12 |
-		SBC_BLOCK_LENGTH_16,
-	.subbands =
-		SBC_SUBBANDS_4 |
-		SBC_SUBBANDS_8,
-	.allocation_method =
-		SBC_ALLOCATION_SNR |
-		SBC_ALLOCATION_LOUDNESS,
-	.min_bitpool = SBC_MIN_BITPOOL,
-	.max_bitpool = SBC_MAX_BITPOOL,
-};
-
-static const struct a2dp_channel_mode a2dp_sbc_channels[] = {
-	{ A2DP_CHM_MONO, 1, SBC_CHANNEL_MODE_MONO },
-	{ A2DP_CHM_DUAL_CHANNEL, 2, SBC_CHANNEL_MODE_DUAL_CHANNEL },
-	{ A2DP_CHM_STEREO, 2, SBC_CHANNEL_MODE_STEREO },
-	{ A2DP_CHM_JOINT_STEREO, 2, SBC_CHANNEL_MODE_JOINT_STEREO },
-};
-
-static const struct a2dp_sampling_freq a2dp_sbc_samplings[] = {
-	{ 16000, SBC_SAMPLING_FREQ_16000 },
-	{ 32000, SBC_SAMPLING_FREQ_32000 },
-	{ 44100, SBC_SAMPLING_FREQ_44100 },
-	{ 48000, SBC_SAMPLING_FREQ_48000 },
-};
-
-static const a2dp_mpeg_t a2dp_mpeg_source = {
-	.layer =
-		MPEG_LAYER_MP3,
-	.crc = 1,
-	.channel_mode =
-	/* NOTE: LAME does not support dual-channel mode. */
-		MPEG_CHANNEL_MODE_MONO |
-		MPEG_CHANNEL_MODE_STEREO |
-		MPEG_CHANNEL_MODE_JOINT_STEREO,
-	/* NOTE: Since MPF-2 is not required for neither Sink
-	 *       nor Source, we are not going to support it. */
-	.mpf = 0,
-	.frequency =
-		MPEG_SAMPLING_FREQ_16000 |
-		MPEG_SAMPLING_FREQ_22050 |
-		MPEG_SAMPLING_FREQ_24000 |
-		MPEG_SAMPLING_FREQ_32000 |
-		MPEG_SAMPLING_FREQ_44100 |
-		MPEG_SAMPLING_FREQ_48000,
-	.vbr = 1,
-	MPEG_INIT_BITRATE(
-		MPEG_BIT_RATE_INDEX_0 |
-		MPEG_BIT_RATE_INDEX_1 |
-		MPEG_BIT_RATE_INDEX_2 |
-		MPEG_BIT_RATE_INDEX_3 |
-		MPEG_BIT_RATE_INDEX_4 |
-		MPEG_BIT_RATE_INDEX_5 |
-		MPEG_BIT_RATE_INDEX_6 |
-		MPEG_BIT_RATE_INDEX_7 |
-		MPEG_BIT_RATE_INDEX_8 |
-		MPEG_BIT_RATE_INDEX_9 |
-		MPEG_BIT_RATE_INDEX_10 |
-		MPEG_BIT_RATE_INDEX_11 |
-		MPEG_BIT_RATE_INDEX_12 |
-		MPEG_BIT_RATE_INDEX_13 |
-		MPEG_BIT_RATE_INDEX_14
-	)
-};
-
-static const a2dp_mpeg_t a2dp_mpeg_sink = {
-	.layer =
-#if ENABLE_MPG123
-		MPEG_LAYER_MP1 |
-		MPEG_LAYER_MP2 |
+struct a2dp_codec * const a2dp_codecs[] = {
+#if ENABLE_LC3PLUS
+	&a2dp_lc3plus_source,
+	&a2dp_lc3plus_sink,
 #endif
-		MPEG_LAYER_MP3,
-	.crc = 1,
-	.channel_mode =
-	/* NOTE: LAME does not support dual-channel mode. Be aware, that
-	 *       lack of this feature violates A2DP Sink specification. */
-		MPEG_CHANNEL_MODE_MONO |
-#if ENABLE_MPG123
-		MPEG_CHANNEL_MODE_DUAL_CHANNEL |
-#endif
-		MPEG_CHANNEL_MODE_STEREO |
-		MPEG_CHANNEL_MODE_JOINT_STEREO,
-	/* NOTE: Since MPF-2 is not required for neither Sink
-	 *       nor Source, we are not going to support it. */
-	.mpf = 0,
-	.frequency =
-		MPEG_SAMPLING_FREQ_16000 |
-		MPEG_SAMPLING_FREQ_22050 |
-		MPEG_SAMPLING_FREQ_24000 |
-		MPEG_SAMPLING_FREQ_32000 |
-		MPEG_SAMPLING_FREQ_44100 |
-		MPEG_SAMPLING_FREQ_48000,
-	.vbr = 1,
-	MPEG_INIT_BITRATE(
-		MPEG_BIT_RATE_INDEX_0 |
-		MPEG_BIT_RATE_INDEX_1 |
-		MPEG_BIT_RATE_INDEX_2 |
-		MPEG_BIT_RATE_INDEX_3 |
-		MPEG_BIT_RATE_INDEX_4 |
-		MPEG_BIT_RATE_INDEX_5 |
-		MPEG_BIT_RATE_INDEX_6 |
-		MPEG_BIT_RATE_INDEX_7 |
-		MPEG_BIT_RATE_INDEX_8 |
-		MPEG_BIT_RATE_INDEX_9 |
-		MPEG_BIT_RATE_INDEX_10 |
-		MPEG_BIT_RATE_INDEX_11 |
-		MPEG_BIT_RATE_INDEX_12 |
-		MPEG_BIT_RATE_INDEX_13 |
-		MPEG_BIT_RATE_INDEX_14
-	)
-};
-
-static const struct a2dp_channel_mode a2dp_mpeg_channels[] = {
-	{ A2DP_CHM_MONO, 1, MPEG_CHANNEL_MODE_MONO },
-	{ A2DP_CHM_DUAL_CHANNEL, 2, MPEG_CHANNEL_MODE_DUAL_CHANNEL },
-	{ A2DP_CHM_STEREO, 2, MPEG_CHANNEL_MODE_STEREO },
-	{ A2DP_CHM_JOINT_STEREO, 2, MPEG_CHANNEL_MODE_JOINT_STEREO },
-};
-
-static const struct a2dp_sampling_freq a2dp_mpeg_samplings[] = {
-	{ 16000, MPEG_SAMPLING_FREQ_16000 },
-	{ 22050, MPEG_SAMPLING_FREQ_22050 },
-	{ 24000, MPEG_SAMPLING_FREQ_24000 },
-	{ 32000, MPEG_SAMPLING_FREQ_32000 },
-	{ 44100, MPEG_SAMPLING_FREQ_44100 },
-	{ 48000, MPEG_SAMPLING_FREQ_48000 },
-};
-
-static const a2dp_aac_t a2dp_aac = {
-	.object_type =
-		/* NOTE: AAC Long Term Prediction and AAC Scalable are
-		 *       not supported by the FDK-AAC library. */
-		AAC_OBJECT_TYPE_MPEG2_AAC_LC |
-		AAC_OBJECT_TYPE_MPEG4_AAC_LC,
-	AAC_INIT_FREQUENCY(
-			AAC_SAMPLING_FREQ_8000 |
-			AAC_SAMPLING_FREQ_11025 |
-			AAC_SAMPLING_FREQ_12000 |
-			AAC_SAMPLING_FREQ_16000 |
-			AAC_SAMPLING_FREQ_22050 |
-			AAC_SAMPLING_FREQ_24000 |
-			AAC_SAMPLING_FREQ_32000 |
-			AAC_SAMPLING_FREQ_44100 |
-			AAC_SAMPLING_FREQ_48000 |
-			AAC_SAMPLING_FREQ_64000 |
-			AAC_SAMPLING_FREQ_88200 |
-			AAC_SAMPLING_FREQ_96000)
-	.channels =
-		AAC_CHANNELS_1 |
-		AAC_CHANNELS_2,
-	.vbr = 1,
-	AAC_INIT_BITRATE(320000)
-};
-
-static const struct a2dp_channel_mode a2dp_aac_channels[] = {
-	{ A2DP_CHM_MONO, 1, AAC_CHANNELS_1 },
-	{ A2DP_CHM_STEREO, 2, AAC_CHANNELS_2 },
-};
-
-static const struct a2dp_sampling_freq a2dp_aac_samplings[] = {
-	{ 8000, AAC_SAMPLING_FREQ_8000 },
-	{ 11025, AAC_SAMPLING_FREQ_11025 },
-	{ 12000, AAC_SAMPLING_FREQ_12000 },
-	{ 16000, AAC_SAMPLING_FREQ_16000 },
-	{ 22050, AAC_SAMPLING_FREQ_22050 },
-	{ 24000, AAC_SAMPLING_FREQ_24000 },
-	{ 32000, AAC_SAMPLING_FREQ_32000 },
-	{ 44100, AAC_SAMPLING_FREQ_44100 },
-	{ 48000, AAC_SAMPLING_FREQ_48000 },
-	{ 64000, AAC_SAMPLING_FREQ_64000 },
-	{ 88200, AAC_SAMPLING_FREQ_88200 },
-	{ 96000, AAC_SAMPLING_FREQ_96000 },
-};
-
-static const a2dp_aptx_t a2dp_aptx = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(APTX_VENDOR_ID, APTX_CODEC_ID),
-	.channel_mode =
-		/* NOTE: Used apt-X library does not support
-		 *       single channel (mono) mode. */
-		APTX_CHANNEL_MODE_STEREO,
-	.frequency =
-		APTX_SAMPLING_FREQ_16000 |
-		APTX_SAMPLING_FREQ_32000 |
-		APTX_SAMPLING_FREQ_44100 |
-		APTX_SAMPLING_FREQ_48000,
-};
-
-static const struct a2dp_channel_mode a2dp_aptx_channels[] = {
-	{ A2DP_CHM_STEREO, 2, APTX_CHANNEL_MODE_STEREO },
-};
-
-static const struct a2dp_sampling_freq a2dp_aptx_samplings[] = {
-	{ 16000, APTX_SAMPLING_FREQ_16000 },
-	{ 32000, APTX_SAMPLING_FREQ_32000 },
-	{ 44100, APTX_SAMPLING_FREQ_44100 },
-	{ 48000, APTX_SAMPLING_FREQ_48000 },
-};
-
-static const a2dp_faststream_t a2dp_faststream = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(FASTSTREAM_VENDOR_ID, FASTSTREAM_CODEC_ID),
-	.direction = FASTSTREAM_DIRECTION_MUSIC | FASTSTREAM_DIRECTION_VOICE,
-	.frequency_music =
-		FASTSTREAM_SAMPLING_FREQ_MUSIC_44100 |
-		FASTSTREAM_SAMPLING_FREQ_MUSIC_48000,
-	.frequency_voice =
-		FASTSTREAM_SAMPLING_FREQ_VOICE_16000,
-};
-
-static const struct a2dp_sampling_freq a2dp_faststream_samplings_music[] = {
-	{ 44100, FASTSTREAM_SAMPLING_FREQ_MUSIC_44100 },
-	{ 48000, FASTSTREAM_SAMPLING_FREQ_MUSIC_48000 },
-};
-
-static const struct a2dp_sampling_freq a2dp_faststream_samplings_voice[] = {
-	{ 16000, FASTSTREAM_SAMPLING_FREQ_VOICE_16000 },
-};
-
-static const a2dp_aptx_hd_t a2dp_aptx_hd = {
-	.aptx.info = A2DP_SET_VENDOR_ID_CODEC_ID(APTX_HD_VENDOR_ID, APTX_HD_CODEC_ID),
-	.aptx.channel_mode =
-		/* NOTE: Used apt-X HD library does not support
-		 *       single channel (mono) mode. */
-		APTX_CHANNEL_MODE_STEREO,
-	.aptx.frequency =
-		APTX_SAMPLING_FREQ_16000 |
-		APTX_SAMPLING_FREQ_32000 |
-		APTX_SAMPLING_FREQ_44100 |
-		APTX_SAMPLING_FREQ_48000,
-};
-
-static const struct a2dp_channel_mode a2dp_aptx_hd_channels[] = {
-	{ A2DP_CHM_STEREO, 2, APTX_CHANNEL_MODE_STEREO },
-};
-
-static const struct a2dp_sampling_freq a2dp_aptx_hd_samplings[] = {
-	{ 16000, APTX_SAMPLING_FREQ_16000 },
-	{ 32000, APTX_SAMPLING_FREQ_32000 },
-	{ 44100, APTX_SAMPLING_FREQ_44100 },
-	{ 48000, APTX_SAMPLING_FREQ_48000 },
-};
-
-static const a2dp_ldac_t a2dp_ldac = {
-	.info = A2DP_SET_VENDOR_ID_CODEC_ID(LDAC_VENDOR_ID, LDAC_CODEC_ID),
-	.channel_mode =
-		LDAC_CHANNEL_MODE_MONO |
-		LDAC_CHANNEL_MODE_DUAL |
-		LDAC_CHANNEL_MODE_STEREO,
-	.frequency =
-		/* NOTE: Used LDAC library does not support
-		 *       frequencies higher than 96 kHz. */
-		LDAC_SAMPLING_FREQ_44100 |
-		LDAC_SAMPLING_FREQ_48000 |
-		LDAC_SAMPLING_FREQ_88200 |
-		LDAC_SAMPLING_FREQ_96000,
-};
-
-static const struct a2dp_channel_mode a2dp_ldac_channels[] = {
-	{ A2DP_CHM_MONO, 1, LDAC_CHANNEL_MODE_MONO },
-	{ A2DP_CHM_DUAL_CHANNEL, 2, LDAC_CHANNEL_MODE_DUAL },
-	{ A2DP_CHM_STEREO, 2, LDAC_CHANNEL_MODE_STEREO },
-};
-
-static const struct a2dp_sampling_freq a2dp_ldac_samplings[] = {
-	{ 44100, LDAC_SAMPLING_FREQ_44100 },
-	{ 48000, LDAC_SAMPLING_FREQ_48000 },
-	{ 88200, LDAC_SAMPLING_FREQ_88200 },
-	{ 96000, LDAC_SAMPLING_FREQ_96000 },
-};
-
-static const struct a2dp_codec a2dp_codec_source_sbc = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_SBC,
-	.capabilities = &a2dp_sbc,
-	.capabilities_size = sizeof(a2dp_sbc),
-	.channels[0] = a2dp_sbc_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_sbc_channels),
-	.samplings[0] = a2dp_sbc_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_sbc_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_sbc = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_SBC,
-	.capabilities = &a2dp_sbc,
-	.capabilities_size = sizeof(a2dp_sbc),
-	.channels[0] = a2dp_sbc_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_sbc_channels),
-	.samplings[0] = a2dp_sbc_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_sbc_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_source_mpeg = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_MPEG12,
-	.capabilities = &a2dp_mpeg_source,
-	.capabilities_size = sizeof(a2dp_mpeg_source),
-	.channels[0] = a2dp_mpeg_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_mpeg_channels),
-	.samplings[0] = a2dp_mpeg_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_mpeg_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_mpeg = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_MPEG12,
-	.capabilities = &a2dp_mpeg_sink,
-	.capabilities_size = sizeof(a2dp_mpeg_sink),
-	.channels[0] = a2dp_mpeg_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_mpeg_channels),
-	.samplings[0] = a2dp_mpeg_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_mpeg_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_source_aac = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_MPEG24,
-	.capabilities = &a2dp_aac,
-	.capabilities_size = sizeof(a2dp_aac),
-	.channels[0] = a2dp_aac_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_aac_channels),
-	.samplings[0] = a2dp_aac_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_aac_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_aac = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_MPEG24,
-	.capabilities = &a2dp_aac,
-	.capabilities_size = sizeof(a2dp_aac),
-	.channels[0] = a2dp_aac_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_aac_channels),
-	.samplings[0] = a2dp_aac_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_aac_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_source_aptx = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_VENDOR_APTX,
-	.capabilities = &a2dp_aptx,
-	.capabilities_size = sizeof(a2dp_aptx),
-	.channels[0] = a2dp_aptx_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_aptx_channels),
-	.samplings[0] = a2dp_aptx_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_aptx_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_aptx = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_VENDOR_APTX,
-	.capabilities = &a2dp_aptx,
-	.capabilities_size = sizeof(a2dp_aptx),
-	.channels[0] = a2dp_aptx_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_aptx_channels),
-	.samplings[0] = a2dp_aptx_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_aptx_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_source_aptx_hd = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_VENDOR_APTX_HD,
-	.capabilities = &a2dp_aptx_hd,
-	.capabilities_size = sizeof(a2dp_aptx_hd),
-	.channels[0] = a2dp_aptx_hd_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_aptx_hd_channels),
-	.samplings[0] = a2dp_aptx_hd_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_aptx_hd_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_aptx_hd = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_VENDOR_APTX_HD,
-	.capabilities = &a2dp_aptx_hd,
-	.capabilities_size = sizeof(a2dp_aptx_hd),
-	.channels[0] = a2dp_aptx_hd_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_aptx_hd_channels),
-	.samplings[0] = a2dp_aptx_hd_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_aptx_hd_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_source_faststream = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_VENDOR_FASTSTREAM,
-	.backchannel = true,
-	.capabilities = &a2dp_faststream,
-	.capabilities_size = sizeof(a2dp_faststream),
-	.samplings[0] = a2dp_faststream_samplings_music,
-	.samplings_size[0] = ARRAYSIZE(a2dp_faststream_samplings_music),
-	.samplings[1] = a2dp_faststream_samplings_voice,
-	.samplings_size[1] = ARRAYSIZE(a2dp_faststream_samplings_voice),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_faststream = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_VENDOR_FASTSTREAM,
-	.backchannel = true,
-	.capabilities = &a2dp_faststream,
-	.capabilities_size = sizeof(a2dp_faststream),
-	.samplings[0] = a2dp_faststream_samplings_music,
-	.samplings_size[0] = ARRAYSIZE(a2dp_faststream_samplings_music),
-	.samplings[1] = a2dp_faststream_samplings_voice,
-	.samplings_size[1] = ARRAYSIZE(a2dp_faststream_samplings_voice),
-};
-
-static const struct a2dp_codec a2dp_codec_source_ldac = {
-	.dir = A2DP_SOURCE,
-	.codec_id = A2DP_CODEC_VENDOR_LDAC,
-	.capabilities = &a2dp_ldac,
-	.capabilities_size = sizeof(a2dp_ldac),
-	.channels[0] = a2dp_ldac_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_ldac_channels),
-	.samplings[0] = a2dp_ldac_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_ldac_samplings),
-};
-
-static const struct a2dp_codec a2dp_codec_sink_ldac = {
-	.dir = A2DP_SINK,
-	.codec_id = A2DP_CODEC_VENDOR_LDAC,
-	.capabilities = &a2dp_ldac,
-	.capabilities_size = sizeof(a2dp_ldac),
-	.channels[0] = a2dp_ldac_channels,
-	.channels_size[0] = ARRAYSIZE(a2dp_ldac_channels),
-	.samplings[0] = a2dp_ldac_samplings,
-	.samplings_size[0] = ARRAYSIZE(a2dp_ldac_samplings),
-};
-
-const struct a2dp_codec *a2dp_codecs[] = {
 #if ENABLE_LDAC
-	&a2dp_codec_source_ldac,
+	&a2dp_ldac_source,
+# if HAVE_LDAC_DECODE
+	&a2dp_ldac_sink,
+# endif
 #endif
 #if ENABLE_APTX_HD
-	&a2dp_codec_source_aptx_hd,
+	&a2dp_aptx_hd_source,
+# if HAVE_APTX_HD_DECODE
+	&a2dp_aptx_hd_sink,
+# endif
 #endif
 #if ENABLE_APTX
-	&a2dp_codec_source_aptx,
+	&a2dp_aptx_source,
+# if HAVE_APTX_DECODE
+	&a2dp_aptx_sink,
+# endif
 #endif
 #if ENABLE_FASTSTREAM
-	&a2dp_codec_source_faststream,
-	&a2dp_codec_sink_faststream,
+	&a2dp_faststream_source,
+	&a2dp_faststream_sink,
 #endif
 #if ENABLE_AAC
-	&a2dp_codec_source_aac,
-	&a2dp_codec_sink_aac,
+	&a2dp_aac_source,
+	&a2dp_aac_sink,
 #endif
 #if ENABLE_MPEG
 # if ENABLE_MP3LAME
-	&a2dp_codec_source_mpeg,
+	&a2dp_mpeg_source,
 # endif
 # if ENABLE_MP3LAME || ENABLE_MPG123
-	&a2dp_codec_sink_mpeg,
+	&a2dp_mpeg_sink,
 # endif
 #endif
-	&a2dp_codec_source_sbc,
-	&a2dp_codec_sink_sbc,
+	&a2dp_sbc_source,
+	&a2dp_sbc_sink,
 	NULL,
 };
 
+/**
+ * Initialize A2DP codecs. */
+int a2dp_codecs_init(void) {
+	a2dp_sbc_init();
+#if ENABLE_MPEG
+	a2dp_mpeg_init();
+#endif
+#if ENABLE_AAC
+	a2dp_aac_init();
+#endif
+#if ENABLE_APTX
+	a2dp_aptx_init();
+#endif
+#if ENABLE_APTX_HD
+	a2dp_aptx_hd_init();
+#endif
+#if ENABLE_FASTSTREAM
+	a2dp_faststream_init();
+#endif
+#if ENABLE_LC3PLUS
+	a2dp_lc3plus_init();
+#endif
+#if ENABLE_LDAC
+	a2dp_ldac_init();
+#endif
+	return 0;
+}
+
+static int a2dp_codecs_qsort_cmp(const void *a_, const void *b_) {
+	const struct a2dp_codec *a = *(const struct a2dp_codec **)a_;
+	const struct a2dp_codec *b = *(const struct a2dp_codec **)b_;
+	int ret;
+	if ((ret = a->dir - b->dir) != 0)
+		return ret;
+	if (a->codec_id >= A2DP_CODEC_VENDOR &&
+			b->codec_id >= A2DP_CODEC_VENDOR) {
+		const char *a_name = a2dp_codecs_codec_id_to_string(a->codec_id);
+		const char *b_name = a2dp_codecs_codec_id_to_string(b->codec_id);
+		return strcasecmp(a_name, b_name);
+	}
+	return a->codec_id - b->codec_id;
+}
+
+/**
+ * Sort A2DP codecs.
+ *
+ * This function sorts A2DP codecs according to following rules:
+ *  - sort codecs by A2DP direction
+ *  - sort codecs by codec ID
+ *  - sort vendor codecs alphabetically (case insensitive) */
+void a2dp_codecs_qsort(const struct a2dp_codec ** codecs, size_t nmemb) {
+	qsort(codecs, nmemb, sizeof(*codecs), a2dp_codecs_qsort_cmp);
+}
+
 /**
  * Lookup codec configuration for given stream direction.
  *
@@ -501,14 +139,59 @@ const struct a2dp_codec *a2dp_codecs[] = {
  * @return On success this function returns the address of the codec
  *   configuration structure. Otherwise, NULL is returned. */
 const struct a2dp_codec *a2dp_codec_lookup(uint16_t codec_id, enum a2dp_dir dir) {
-	size_t i;
-	for (i = 0; i < ARRAYSIZE(a2dp_codecs) - 1; i++)
+	for (size_t i = 0; a2dp_codecs[i] != NULL; i++)
 		if (a2dp_codecs[i]->dir == dir &&
 				a2dp_codecs[i]->codec_id == codec_id)
 			return a2dp_codecs[i];
 	return NULL;
 }
 
+/**
+ * Lookup number of channels for given capability value.
+ *
+ * @param codec A2DP codec setup.
+ * @param capability_value A2DP codec channel mode configuration value.
+ * @param backchannel If true, lookup in the back-channel configuration.
+ * @return On success this function returns the number of channels. Otherwise,
+ *   if given capability value is not supported (or invalid), 0 is returned. */
+unsigned int a2dp_codec_lookup_channels(
+		const struct a2dp_codec *codec,
+		uint16_t capability_value,
+		bool backchannel) {
+
+	const size_t slot = backchannel ? 1 : 0;
+	size_t i;
+
+	for (i = 0; i < codec->channels_size[slot]; i++)
+		if (capability_value == codec->channels[slot][i].value)
+			return codec->channels[slot][i].channels;
+
+	return 0;
+}
+
+/**
+ * Lookup sampling frequency for given capability value.
+ *
+ * @param codec A2DP codec setup.
+ * @param capability_value A2DP codec sampling configuration value.
+ * @param backchannel If true, lookup in the back-channel configuration.
+ * @return On success this function returns the sampling frequency. Otherwise,
+ *   if given capability value is not supported (or invalid), 0 is returned. */
+unsigned int a2dp_codec_lookup_frequency(
+		const struct a2dp_codec *codec,
+		uint16_t capability_value,
+		bool backchannel) {
+
+	const size_t slot = backchannel ? 1 : 0;
+	size_t i;
+
+	for (i = 0; i < codec->samplings_size[slot]; i++)
+		if (capability_value == codec->samplings[slot][i].value)
+			return codec->samplings[slot][i].frequency;
+
+	return 0;
+}
+
 /**
  * Get A2DP 16-bit vendor codec ID - BlueALSA extension.
  *
@@ -564,14 +247,19 @@ uint16_t a2dp_get_vendor_codec_id(const void *capabilities, size_t size) {
 		switch (codec_id) {
 		case LHDC_CODEC_ID:
 			return A2DP_CODEC_VENDOR_LHDC;
+		case LHDC_LL_CODEC_ID:
+			return A2DP_CODEC_VENDOR_LHDC_LL;
 		case LHDC_V1_CODEC_ID:
 			return A2DP_CODEC_VENDOR_LHDC_V1;
-		case LLAC_CODEC_ID:
-			return A2DP_CODEC_VENDOR_LLAC;
+		} break;
+	case BT_COMPID_FRAUNHOFER_IIS:
+		switch (codec_id) {
+		case LC3PLUS_CODEC_ID:
+			return A2DP_CODEC_VENDOR_LC3PLUS;
 		} break;
 	}
 
-	hexdump("Unknown vendor codec", capabilities, size);
+	hexdump("Unknown vendor codec", capabilities, size, true);
 
 	errno = ENOTSUP;
 	return 0xFFFF;
@@ -728,9 +416,34 @@ uint32_t a2dp_check_configuration(
 
 #if ENABLE_FASTSTREAM
 	case A2DP_CODEC_VENDOR_FASTSTREAM: {
+
 		const a2dp_faststream_t *cap = configuration;
 		cap_freq = cap->frequency_music;
 		cap_freq_bc = cap->frequency_voice;
+
+		if ((cap->direction & (FASTSTREAM_DIRECTION_MUSIC | FASTSTREAM_DIRECTION_VOICE)) == 0) {
+			debug("Invalid FastStream directions: %#x", cap->direction);
+			ret |= A2DP_CHECK_ERR_FASTSTREAM_DIR;
+		}
+
+		break;
+	}
+#endif
+
+#if ENABLE_LC3PLUS
+	case A2DP_CODEC_VENDOR_LC3PLUS: {
+
+		const a2dp_lc3plus_t *cap = configuration;
+		cap_chm = cap->channels;
+		cap_freq = LC3PLUS_GET_FREQUENCY(*cap);
+
+		if (cap->frame_duration != LC3PLUS_FRAME_DURATION_025 &&
+				cap->frame_duration != LC3PLUS_FRAME_DURATION_050 &&
+				cap->frame_duration != LC3PLUS_FRAME_DURATION_100) {
+			debug("Invalid LC3plus frame duration: %#x", cap->frame_duration);
+			ret |= A2DP_CHECK_ERR_LC3PLUS_DURATION;
+		}
+
 		break;
 	}
 #endif
@@ -783,21 +496,21 @@ int a2dp_filter_capabilities(
 		return errno = EINVAL, -1;
 	}
 
-	uint8_t tmp[32];
+	a2dp_t tmp;
 	g_assert_cmpuint(sizeof(tmp), >=, size);
 
 	size_t i;
 	for (i = 0; i < size; i++)
-		tmp[i] = ((uint8_t *)capabilities)[i] & ((uint8_t *)codec->capabilities)[i];
+		((uint8_t *)&tmp)[i] = ((uint8_t *)capabilities)[i] & ((uint8_t *)&codec->capabilities)[i];
 
 	switch (codec->codec_id) {
 	case A2DP_CODEC_SBC:
-		((a2dp_sbc_t *)tmp)->min_bitpool = MAX(
+		tmp.sbc.min_bitpool = MAX(
 			((a2dp_sbc_t *)capabilities)->min_bitpool,
-			((a2dp_sbc_t *)codec->capabilities)->min_bitpool);
-		((a2dp_sbc_t *)tmp)->max_bitpool = MIN(
+			codec->capabilities.sbc.min_bitpool);
+		tmp.sbc.max_bitpool = MIN(
 			((a2dp_sbc_t *)capabilities)->max_bitpool,
-			((a2dp_sbc_t *)codec->capabilities)->max_bitpool);
+			codec->capabilities.sbc.max_bitpool);
 		break;
 #if ENABLE_MPEG
 	case A2DP_CODEC_MPEG12:
@@ -805,9 +518,9 @@ int a2dp_filter_capabilities(
 #endif
 #if ENABLE_AAC
 	case A2DP_CODEC_MPEG24:
-		AAC_SET_BITRATE(*(a2dp_aac_t *)tmp, MIN(
+		AAC_SET_BITRATE(tmp.aac, MIN(
 					AAC_GET_BITRATE(*(a2dp_aac_t *)capabilities),
-					AAC_GET_BITRATE(*(a2dp_aac_t *)codec->capabilities)));
+					AAC_GET_BITRATE(codec->capabilities.aac)));
 		break;
 #endif
 #if ENABLE_APTX
@@ -822,6 +535,10 @@ int a2dp_filter_capabilities(
 	case A2DP_CODEC_VENDOR_FASTSTREAM:
 		break;
 #endif
+#if ENABLE_LC3PLUS
+	case A2DP_CODEC_VENDOR_LC3PLUS:
+		break;
+#endif
 #if ENABLE_LDAC
 	case A2DP_CODEC_VENDOR_LDAC:
 		break;
@@ -830,7 +547,7 @@ int a2dp_filter_capabilities(
 		g_assert_not_reached();
 	}
 
-	memcpy(capabilities, tmp, size);
+	memcpy(capabilities, &tmp, size);
 	return 0;
 }
 
@@ -910,7 +627,8 @@ int a2dp_select_configuration(
 			goto fail;
 		}
 
-		if (config.sbc_quality == SBC_QUALITY_XQ) {
+		if (config.sbc_quality == SBC_QUALITY_XQ ||
+				config.sbc_quality == SBC_QUALITY_XQPLUS) {
 			if (cap_chm & SBC_CHANNEL_MODE_DUAL_CHANNEL)
 				cap->channel_mode = SBC_CHANNEL_MODE_DUAL_CHANNEL;
 			else
@@ -953,8 +671,8 @@ int a2dp_select_configuration(
 			goto fail;
 		}
 
-		cap->min_bitpool = MAX(SBC_MIN_BITPOOL, cap->min_bitpool);
-		cap->max_bitpool = MIN(SBC_MAX_BITPOOL, cap->max_bitpool);
+		cap->min_bitpool = MAX(codec->capabilities.sbc.min_bitpool, cap->min_bitpool);
+		cap->max_bitpool = MIN(codec->capabilities.sbc.max_bitpool, cap->max_bitpool);
 
 		break;
 	}
@@ -1025,8 +743,15 @@ int a2dp_select_configuration(
 			goto fail;
 		}
 
-		if (AAC_GET_BITRATE(*cap) == 0)
-			AAC_SET_BITRATE(*cap, AAC_GET_BITRATE(*(a2dp_aac_t *)codec->capabilities));
+		unsigned int ba_bitrate = AAC_GET_BITRATE(codec->capabilities.aac);
+		unsigned int peer_bitrate = AAC_GET_BITRATE(*cap);
+		if (peer_bitrate == 0)
+			/* fix bitrate value if it was not set */
+			peer_bitrate = UINT_MAX;
+		AAC_SET_BITRATE(*cap, MIN(ba_bitrate, peer_bitrate));
+
+		if (!config.aac_prefer_vbr)
+			cap->vbr = 0;
 
 		break;
 	}
@@ -1091,6 +816,45 @@ int a2dp_select_configuration(
 			goto fail;
 		}
 
+		if ((cap->direction & (FASTSTREAM_DIRECTION_MUSIC | FASTSTREAM_DIRECTION_VOICE)) == 0) {
+			error("FastStream: No supported directions: %#x", cap->direction);
+		}
+
+		break;
+	}
+#endif
+
+#if ENABLE_LC3PLUS
+	case A2DP_CODEC_VENDOR_LC3PLUS: {
+
+		a2dp_lc3plus_t *cap = capabilities;
+		unsigned int cap_chm = cap->channels;
+		unsigned int cap_freq = LC3PLUS_GET_FREQUENCY(*cap);
+
+		if (cap->frame_duration & LC3PLUS_FRAME_DURATION_100)
+			cap->frame_duration = LC3PLUS_FRAME_DURATION_100;
+		else if (cap->frame_duration & LC3PLUS_FRAME_DURATION_050)
+			cap->frame_duration = LC3PLUS_FRAME_DURATION_050;
+		else if (cap->frame_duration & LC3PLUS_FRAME_DURATION_025)
+			cap->frame_duration = LC3PLUS_FRAME_DURATION_025;
+		else {
+			error("LC3plus: No supported frame duration: %#x", cap->frame_duration);
+			goto fail;
+		}
+
+		if ((cap->channels = a2dp_codec_select_channel_mode(codec, cap_chm, false)) == 0) {
+			error("LC3plus: No supported channels: %#x", cap_chm);
+			goto fail;
+		}
+
+		unsigned int freq;
+		if ((freq = a2dp_codec_select_sampling_freq(codec, cap_freq, false)) != 0)
+			LC3PLUS_SET_FREQUENCY(*cap, freq);
+		else {
+			error("LC3plus: No supported sampling frequencies: %#x", cap_freq);
+			goto fail;
+		}
+
 		break;
 	}
 #endif
