@@ -1,6 +1,6 @@
 /*
  * bluealsa-pcm.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2021 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -35,15 +35,37 @@
 
 #include "shared/dbus-client.h"
 #include "shared/defs.h"
+#include "shared/hex.h"
 #include "shared/log.h"
 #include "shared/rt.h"
 
+#define BA_PAUSE_STATE_RUNNING 0
+#define BA_PAUSE_STATE_PAUSED  (1 << 0)
+#define BA_PAUSE_STATE_PENDING (1 << 1)
+
+#if SND_LIB_VERSION >= 0x010104 && SND_LIB_VERSION <= 0x010205
+/**
+ * alsa-lib releases 1.1.4 to 1.2.5.1 inclusive have a bug in the rate plugin
+ * which, when combined with the hw params refinement algorithm used by the
+ * ioplug, can cause snd_pcm_avail() to return bogus values. This, in turn,
+ * can trigger deadlock in applications built on the portaudio library
+ * (e.g. audacity) and possibly cause faults in other applications too.
+ *
+ * This macro enables a work-around for this bug.
+ * */
+# define BLUEALSA_HW_PARAMS_FIX 1
+#endif
 
 struct bluealsa_pcm {
 	snd_pcm_ioplug_t io;
 
 	/* D-Bus connection context */
 	struct ba_dbus_ctx dbus_ctx;
+	/* time of last D-Bus dispatching */
+	struct timespec dbus_dispatch_ts;
+
+	/* IO thread and application thread sync */
+	pthread_mutex_t mutex;
 
 	/* requested BlueALSA PCM */
 	struct ba_pcm ba_pcm;
@@ -70,17 +92,25 @@ struct bluealsa_pcm {
 	/* ALSA operates on frames, we on bytes */
 	size_t frame_size;
 
-	pthread_mutex_t delay_mutex;
 	struct timespec delay_ts;
 	snd_pcm_uframes_t delay_hw_ptr;
 	unsigned int delay_pcm_nread;
-	bool delay_valid;
+	/* In the capture mode, delay_running indicates that frames are being
+	 * transfered to the FIFO by the server. In playback mode it indicates
+	 * that the IO thread is transferring frames to the FIFO. */
+	bool delay_running;
 
 	/* delay accumulated just before pausing */
 	snd_pcm_sframes_t delay_paused;
+	/* maximum delay in FIFO */
+	snd_pcm_sframes_t delay_fifo_size;
 	/* user provided extra delay component */
 	snd_pcm_sframes_t delay_ex;
 
+	/* synchronize threads to begin/end pause */
+	pthread_cond_t pause_cond;
+	unsigned int pause_state;
+
 };
 
 /**
@@ -112,6 +142,7 @@ static snd_pcm_uframes_t snd_pcm_ioplug_hw_avail(const snd_pcm_ioplug_t * const
  * Helper function for closing PCM transport. */
 static int close_transport(struct bluealsa_pcm *pcm) {
 	int rv = 0;
+	pthread_mutex_lock(&pcm->mutex);
 	if (pcm->ba_pcm_fd != -1) {
 		rv |= close(pcm->ba_pcm_fd);
 		pcm->ba_pcm_fd = -1;
@@ -120,6 +151,7 @@ static int close_transport(struct bluealsa_pcm *pcm) {
 		rv |= close(pcm->ba_pcm_ctrl_fd);
 		pcm->ba_pcm_ctrl_fd = -1;
 	}
+	pthread_mutex_unlock(&pcm->mutex);
 	return rv;
 }
 
@@ -131,22 +163,34 @@ static void io_thread_cleanup(struct bluealsa_pcm *pcm) {
 }
 
 /**
- * Some playback applications may start the PCM before they have written the
- * first full period of audio data. To match the behavior of the ALSA hw
- * plug-in we do not report an underrun in this case. Instead we pause the IO
- * thread for the estimated time it should take for a real-time application to
- * write the balance of the period. */
-static void io_thread_wait_first_period(snd_pcm_ioplug_t *io) {
-	while (io->appl_ptr < io->period_size && (
-				io->state == SND_PCM_STATE_RUNNING ||
-				io->state == SND_PCM_STATE_PREPARED)) {
-		uint64_t nsec = (io->period_size - io->appl_ptr) * 1000000000 / io->rate;
-		struct timespec ts = {
-			.tv_sec = nsec / 1000000000,
-			.tv_nsec = nsec % 1000000000,
-		};
-		nanosleep(&ts, NULL);
+ * Helper function for IO thread delay calculation. */
+static void io_thread_update_delay(struct bluealsa_pcm *pcm,
+		snd_pcm_sframes_t hw_ptr) {
+
+	struct timespec now;
+	unsigned int nread = 0;
+
+	gettimestamp(&now);
+	ioctl(pcm->ba_pcm_fd, FIONREAD, &nread);
+
+	pthread_mutex_lock(&pcm->mutex);
+
+	/* stash current time and levels */
+	pcm->delay_ts = now;
+	pcm->delay_pcm_nread = nread;
+	if (hw_ptr == -1) {
+		pcm->delay_hw_ptr = 0;
+		if (pcm->io.stream == SND_PCM_STREAM_PLAYBACK)
+			pcm->delay_running = false;
 	}
+	else {
+		pcm->delay_hw_ptr = hw_ptr;
+		if (pcm->io.stream == SND_PCM_STREAM_PLAYBACK)
+			pcm->delay_running = true;
+	}
+
+	pthread_mutex_unlock(&pcm->mutex);
+
 }
 
 /**
@@ -171,75 +215,79 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 		goto fail;
 	}
 
-	/* Block I/O until a full period of frames is available. */
-	if (io->stream == SND_PCM_STREAM_PLAYBACK) {
-		debug2("Waiting for first period of frames");
-		io_thread_wait_first_period(io);
-	}
-
 	struct asrsync asrs;
 	asrsync_init(&asrs, io->rate);
 
+	/* We update pcm->io_hw_ptr (i.e. the value seen by ioplug) only when
+	 * a period has been completed. We use a temporary copy during the
+	 * transfer procedure. */
+	snd_pcm_sframes_t io_hw_ptr = pcm->io_hw_ptr;
+
 	debug2("Starting IO loop: %d", pcm->ba_pcm_fd);
 	for (;;) {
 
-		if (io->state == SND_PCM_STATE_PAUSED ||
+		if (pcm->pause_state & BA_PAUSE_STATE_PENDING ||
 				pcm->io_hw_ptr == -1) {
+			debug2("Pausing IO thread");
+
+			pthread_mutex_lock(&pcm->mutex);
+			pcm->pause_state = BA_PAUSE_STATE_PAUSED;
+			pthread_cond_signal(&pcm->pause_cond);
+			pthread_mutex_unlock(&pcm->mutex);
+
 			int tmp;
-			debug2("IO thread paused: %d", io->state);
 			sigwait(&sigset, &tmp);
+
+			pthread_mutex_lock(&pcm->mutex);
+			pcm->pause_state = BA_PAUSE_STATE_RUNNING;
+			pthread_mutex_unlock(&pcm->mutex);
+
+			debug2("IO thread resumed");
+
 			if (pcm->io_hw_ptr == -1)
 				continue;
 			if (pcm->ba_pcm_fd == -1)
 				goto fail;
-			if (io->stream == SND_PCM_STREAM_PLAYBACK) {
-				debug2("Waiting for first period of frames");
-				io_thread_wait_first_period(io);
-			}
+
 			asrsync_init(&asrs, io->rate);
-			debug2("IO thread resumed: %d", io->state);
+			io_hw_ptr = pcm->io_hw_ptr;
 		}
 
-		if (io->state == SND_PCM_STATE_DISCONNECTED)
-			goto fail;
-
-		/* We update pcm->io_hw_ptr (i.e. the value seen by ioplug) only when
-		 * a transfer has been completed. We use a temporary copy during the
-		 * transfer procedure. */
-		snd_pcm_uframes_t io_hw_ptr = pcm->io_hw_ptr;
+		/* There are 2 reasons why the number of available frames may be
+		 * zero: XRUN or drained final samples; we set the HW pointer to
+		 * -1 to indicate we have no work to do. */
+		snd_pcm_uframes_t avail;
+		if ((avail = snd_pcm_ioplug_hw_avail(io, io_hw_ptr, io->appl_ptr)) == 0) {
+			pcm->io_hw_ptr = io_hw_ptr = -1;
+			io_thread_update_delay(pcm, io_hw_ptr);
+			eventfd_write(pcm->event_fd, 1);
+			continue;
+		}
 
+		/* current offset of the head pointer in the IO buffer */
 		snd_pcm_uframes_t offset = io_hw_ptr % io->buffer_size;
-		snd_pcm_uframes_t frames = pcm->io_avail_min;
-		char *head = pcm->io_hw_buffer + offset * pcm->frame_size;
 
-		/* If the leftover in the buffer is less than a whole period sizes,
-		 * adjust the number of frames which should be transfered. It has
-		 * turned out, that the buffer might contain fractional number of
-		 * periods - it could be an ALSA bug, though, it has to be handled. */
+		/* Transfer at most 1 period of frames in each iteration ... */
+		snd_pcm_uframes_t frames = io->period_size;
+		/* ... but do not try to transfer more frames than are available in the
+		 * ring buffer! */
+		if (frames > avail)
+			frames = avail;
+
+		/* When used with the rate plugin the buffer might contain a fractional
+		 * number of periods. So if the leftover in the buffer is less than a
+		 * whole period size, adjust the number of frames which should be
+		 * transfered.  */
 		if (io->buffer_size - offset < frames)
 			frames = io->buffer_size - offset;
 
-		/* Do not try to transfer more frames than are available in the ring
-		 * buffer! */
-		snd_pcm_uframes_t hw_avail;
-		if ((hw_avail = snd_pcm_ioplug_hw_avail(io, io_hw_ptr, io->appl_ptr)) < frames)
-			frames = hw_avail;
-
-		/* There are 2 reasons why the number of available frames may be
-		 * zero: xrun or drained final samples; we set the HW pointer to
-		 * -1 to indicate we have no work to do. */
-		if (frames == 0) {
-			io_hw_ptr = -1;
-			goto sync;
-		}
-
 		/* IO operation size in bytes */
 		size_t len = frames * pcm->frame_size;
+		char *head = pcm->io_hw_buffer + offset * pcm->frame_size;
 
-		/* Increment the HW pointer (with boundary wrap) ready for the next
-		 * iteration. */
+		/* Increment the HW pointer (with boundary wrap). */
 		io_hw_ptr += frames;
-		if (io_hw_ptr >= pcm->io_hw_boundary)
+		if ((snd_pcm_uframes_t)io_hw_ptr >= pcm->io_hw_boundary)
 			io_hw_ptr -= pcm->io_hw_boundary;
 
 		ssize_t ret = 0;
@@ -261,6 +309,8 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 			if (ret == 0)
 				goto fail;
 
+			io_thread_update_delay(pcm, io_hw_ptr);
+
 		}
 		else {
 
@@ -277,30 +327,19 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 				len -= ret;
 			} while (len != 0);
 
-			struct timespec now;
-			unsigned int nread = 0;
-
-			gettimestamp(&now);
-			ioctl(pcm->ba_pcm_fd, FIONREAD, &nread);
-
-			/* stash current time and levels */
-			pthread_mutex_lock(&pcm->delay_mutex);
-			pcm->delay_ts = now;
-			pcm->delay_hw_ptr = io_hw_ptr;
-			pcm->delay_pcm_nread = nread;
-			pcm->delay_valid = true;
-			pthread_mutex_unlock(&pcm->delay_mutex);
+			io_thread_update_delay(pcm, io_hw_ptr);
 
 			/* synchronize playback time */
 			asrsync_sync(&asrs, frames);
+
 		}
 
-sync:
 		/* Make the new HW pointer value visible to the ioplug. */
 		pcm->io_hw_ptr = io_hw_ptr;
-		/* Generate poll() event so application is made aware of
-		 * the HW pointer change. */
-		eventfd_write(pcm->event_fd, 1);
+
+		/* Wake application thread if enough space/frames is available. */
+		if (frames + io->buffer_size - avail >= pcm->io_avail_min)
+			eventfd_write(pcm->event_fd, 1);
 	}
 
 fail:
@@ -308,6 +347,7 @@ fail:
 	pthread_cleanup_pop(1);
 	close_transport(pcm);
 	eventfd_write(pcm->event_fd, 0xDEAD0000);
+	pthread_cond_signal(&pcm->pause_cond);
 	return NULL;
 }
 
@@ -323,14 +363,17 @@ static int bluealsa_start(snd_pcm_ioplug_t *io) {
 		return 0;
 	}
 
-	/* initialize delay calculation */
-	pcm->delay_valid = false;
-
 	if (!bluealsa_dbus_pcm_ctrl_send_resume(pcm->ba_pcm_ctrl_fd, NULL)) {
 		debug2("Couldn't start PCM: %s", strerror(errno));
 		return -errno;
 	}
 
+	/* Initialize delay calculation - capture reception begins immediately,
+	 * playback transmission begins only when first period has been written
+	 * by the application. */
+	pcm->delay_running = io->stream == SND_PCM_STREAM_CAPTURE ? true : false;
+	gettimestamp(&pcm->delay_ts);
+
 	/* start the IO thread */
 	pcm->io_started = true;
 	if ((errno = pthread_create(&pcm->io_thread, NULL,
@@ -353,10 +396,13 @@ static int bluealsa_stop(snd_pcm_ioplug_t *io) {
 		pthread_cancel(pcm->io_thread);
 		pthread_join(pcm->io_thread, NULL);
 	}
-	pcm->delay_valid = false;
+
+	pcm->delay_running = false;
+	pcm->delay_pcm_nread = 0;
 
 	/* Bug in ioplug - if pcm->io_hw_ptr == -1 then it reports state
-	 * SND_PCM_STATE_XRUN instead of SND_PCM_STATE_SETUP after pcm is stopped */
+	 * SND_PCM_STATE_XRUN instead of SND_PCM_STATE_SETUP after PCM
+	 * was stopped. */
 	pcm->io_hw_ptr = 0;
 
 	if (!bluealsa_dbus_pcm_ctrl_send_drop(pcm->ba_pcm_ctrl_fd, NULL))
@@ -371,8 +417,20 @@ static int bluealsa_stop(snd_pcm_ioplug_t *io) {
 
 static snd_pcm_sframes_t bluealsa_pointer(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
+
+	/* Any error returned here is translated to -EPIPE, SND_PCM_STATE_XRUN,
+	 * by ioplug; and that prevents snd_pcm_readi() and snd_pcm_writei()
+	 * from returning -ENODEV to the application on device disconnection.
+`	 * Instead, when the device is disconnected, we update the PCM state
+	 * directly here but we do not return an error code. This ensures that
+	 * ioplug does not undo that state change. Both snd_pcm_readi() and
+	 * snd_pcm_writei() return -ENODEV when the PCM state is
+	 * SND_PCM_STATE_DISCONNECTED after their internal call to
+	 * snd_pcm_avail_update(), which will be the case when we set it here.
+	 */
 	if (pcm->ba_pcm_fd == -1)
-		return -ENODEV;
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
+
 #ifndef SND_PCM_IOPLUG_FLAG_BOUNDARY_WA
 	if (pcm->io_hw_ptr != -1)
 		return pcm->io_hw_ptr % io->buffer_size;
@@ -385,37 +443,138 @@ static int bluealsa_close(snd_pcm_ioplug_t *io) {
 	debug2("Closing");
 	bluealsa_dbus_connection_ctx_free(&pcm->dbus_ctx);
 	close(pcm->event_fd);
-	pthread_mutex_destroy(&pcm->delay_mutex);
+	pthread_mutex_destroy(&pcm->mutex);
+	pthread_cond_destroy(&pcm->pause_cond);
 	free(pcm);
 	return 0;
 }
 
+#if BLUEALSA_HW_PARAMS_FIX
+/**
+ * Substitute the period and buffer size produced by the ioplug hw param
+ * refinement algorithm with values that do not trigger the rate plugin
+ * avail() implementation bug.
+ *
+ * It is not possible to expand the configuration within a hw_params
+ * container, only to narrow it. By the time we get to see the container
+ * it has already been reduced to a single configuration, so is effectively
+ * read-only. So in order to fix the problematic buffer size calculated by
+ * the ioplug, we need to completely replace the hw_params container for
+ * the bluealsa pcm.
+ * */
+static int bluealsa_fix_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params) {
+#if DEBUG
+	struct bluealsa_pcm *pcm = io->private_data;
+#endif
+	int ret = 0;
+
+	snd_pcm_uframes_t period_size;
+	if ((ret = snd_pcm_hw_params_get_period_size(params, &period_size, 0)) < 0)
+		return ret;
+	snd_pcm_uframes_t buffer_size;
+	if ((ret =snd_pcm_hw_params_get_buffer_size(params, &buffer_size)) < 0)
+		return ret;
+
+	if (buffer_size % period_size == 0)
+		return 0;
+
+	debug2("Attempting to fix hw params buffer size");
+
+	snd_pcm_hw_params_t *refined_params;
+
+	snd_pcm_hw_params_alloca(&refined_params);
+	if ((ret = snd_pcm_hw_params_any(io->pcm, refined_params)) < 0)
+		return ret;
+
+	snd_pcm_access_mask_t *access = alloca(snd_pcm_access_mask_sizeof());
+	if ((ret = snd_pcm_hw_params_get_access_mask(params, access)) < 0)
+		return ret;
+	if ((ret = snd_pcm_hw_params_set_access_mask(io->pcm, refined_params, access)) < 0)
+		return ret;
+
+	snd_pcm_format_t format;
+	if ((ret = snd_pcm_hw_params_get_format(params, &format)) < 0)
+		return ret;
+	if ((ret = snd_pcm_hw_params_set_format(io->pcm, refined_params, format)) < 0)
+		return ret;
+
+	unsigned int channels;
+	if ((ret = snd_pcm_hw_params_get_channels(params, &channels)) < 0)
+		return ret;
+	if ((ret = snd_pcm_hw_params_set_channels(io->pcm, refined_params, channels)) < 0)
+		return ret;
+
+	unsigned int rate;
+	if ((ret = snd_pcm_hw_params_get_rate(params, &rate, 0)) < 0)
+		return ret;
+	if ((ret = snd_pcm_hw_params_set_rate(io->pcm, refined_params, rate, 0)) < 0)
+		return ret;
+
+	if ((ret = snd_pcm_hw_params_set_period_size(io->pcm, refined_params, period_size, 0)) < 0)
+		return ret;
+
+	if ((ret = snd_pcm_hw_params_set_periods_integer(io->pcm, refined_params)) < 0)
+		return ret;
+
+	buffer_size = (buffer_size / period_size) * period_size;
+	if ((ret = snd_pcm_hw_params_set_buffer_size(io->pcm, refined_params, buffer_size)) < 0)
+		return ret;
+
+	memcpy(params, refined_params, snd_pcm_hw_params_sizeof());
+
+	return ret;
+}
+#endif
+
 static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params) {
 	struct bluealsa_pcm *pcm = io->private_data;
 	(void)params;
 	debug2("Initializing HW");
 
+#if	BLUEALSA_HW_PARAMS_FIX
+	if (bluealsa_fix_hw_params(io, params) < 0)
+		debug2("Warning - unable to fix incorrect buffer size in hw parameters");
+	snd_pcm_uframes_t period_size;
+	int ret;
+	if ((ret = snd_pcm_hw_params_get_period_size(params, &period_size, 0)) < 0)
+		return ret;
+	snd_pcm_uframes_t buffer_size;
+	if ((ret = snd_pcm_hw_params_get_buffer_size(params, &buffer_size)) < 0)
+		return ret;
+#endif
+
 	pcm->frame_size = (snd_pcm_format_physical_width(io->format) * io->channels) / 8;
 
 	DBusError err = DBUS_ERROR_INIT;
-	if (!bluealsa_dbus_open_pcm(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
+	if (!bluealsa_dbus_pcm_open(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
 				&pcm->ba_pcm_fd, &pcm->ba_pcm_ctrl_fd, &err)) {
 		debug2("Couldn't open PCM: %s", err.message);
 		dbus_error_free(&err);
 		return -EBUSY;
 	}
 
-	if (pcm->io.stream == SND_PCM_STREAM_PLAYBACK) {
+	if (pcm->io.stream == SND_PCM_STREAM_PLAYBACK)
 		/* By default, the size of the pipe buffer is set to a too large value for
 		 * our purpose. On modern Linux system it is 65536 bytes. Large buffer in
 		 * the playback mode might contribute to an unnecessary audio delay. Since
 		 * it is possible to modify the size of this buffer we will set is to some
 		 * low value, but big enough to prevent audio tearing. Note, that the size
 		 * will be rounded up to the page size (typically 4096 bytes). */
-		pcm->ba_pcm_buffer_size = fcntl(pcm->ba_pcm_fd, F_SETPIPE_SZ, 2048);
-		debug2("FIFO buffer size: %zd", pcm->ba_pcm_buffer_size);
-	}
+		pcm->delay_fifo_size = fcntl(pcm->ba_pcm_fd, F_SETPIPE_SZ, 2048) / pcm->frame_size;
+	else
+		pcm->delay_fifo_size = fcntl(pcm->ba_pcm_fd, F_GETPIPE_SZ)  / pcm->frame_size;
 
+	debug2("FIFO buffer size: %zd frames", pcm->delay_fifo_size);
+
+#if BLUEALSA_HW_PARAMS_FIX
+	/* ALSA default for avail min is one period. */
+	pcm->io_avail_min = period_size;
+
+	debug2("Selected HW buffer: %zd periods x %zd bytes %c= %zd bytes",
+			buffer_size / period_size, pcm->frame_size * period_size,
+			period_size * (buffer_size / period_size) == buffer_size ? '=' : '<',
+			buffer_size * pcm->frame_size);
+#else
 	/* ALSA default for avail min is one period. */
 	pcm->io_avail_min = io->period_size;
 
@@ -423,6 +582,7 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 			io->buffer_size / io->period_size, pcm->frame_size * io->period_size,
 			io->period_size * (io->buffer_size / io->period_size) == io->buffer_size ? '=' : '<',
 			io->buffer_size * pcm->frame_size);
+#endif
 
 	return 0;
 }
@@ -455,8 +615,10 @@ static int bluealsa_prepare(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
 	/* if PCM FIFO is not opened, report it right away */
-	if (pcm->ba_pcm_fd == -1)
+	if (pcm->ba_pcm_fd == -1) {
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
 		return -ENODEV;
+	}
 
 	/* initialize ring buffer */
 	pcm->io_hw_ptr = 0;
@@ -500,40 +662,85 @@ static snd_pcm_sframes_t bluealsa_calculate_delay(snd_pcm_ioplug_t *io) {
 
 	snd_pcm_sframes_t delay = 0;
 
-	if (!pcm->delay_valid) {
-		/* If we have never transmitted anything, the delay is just based
-		 * on the current buffer length (transfered frames or unread frames
-		 * respectively for playback or capture). */
-		delay = snd_pcm_ioplug_hw_avail(io, io->hw_ptr, io->appl_ptr);
-		if (io->stream == SND_PCM_STREAM_CAPTURE)
-			delay = io->buffer_size - delay;
+	/* if PCM is not started there should be no capture delay */
+	if (!pcm->delay_running && io->stream == SND_PCM_STREAM_CAPTURE)
+		return 0;
+
+	struct timespec now;
+	gettimestamp(&now);
+
+	/* In most cases, dispatching D-Bus messages/signals should be done in the
+	 * poll_revents() callback. However, this mode of operation requires client
+	 * code to use ALSA polling API. If for some reasons, client simply writes
+	 * samples to opened PCM and in the same time wants to know the delay, we
+	 * have to process D-Bus messages here. Otherwise, the BlueALSA component
+	 * of the delay - pcm->ba_pcm.delay - might not be up to date.
+	 *
+	 * This synchronous dispatching will be performed only if the last D-Bus
+	 * dispatching was done more than one second ago - this should prioritize
+	 * asynchronous dispatching in the poll_revents() callback. */
+	if (pcm->dbus_dispatch_ts.tv_sec + 1 < now.tv_sec) {
+		bluealsa_dbus_connection_dispatch(&pcm->dbus_ctx);
+		gettimestamp(&pcm->dbus_dispatch_ts);
+	}
+
+	pthread_mutex_lock(&pcm->mutex);
+
+	struct timespec diff;
+	timespecsub(&now, &pcm->delay_ts, &diff);
+
+	/* the maximum number of frames that can have been
+	 * produced/consumed by the server since pcm->delay_ts */
+	unsigned int tframes =
+		(diff.tv_sec * 1000 + diff.tv_nsec / 1000000) * io->rate / 1000;
+
+	/* the number of frames that were in the FIFO at pcm->delay_ts */
+	snd_pcm_uframes_t fifo_delay = pcm->delay_pcm_nread / pcm->frame_size;
+
+	if (io->stream == SND_PCM_STREAM_CAPTURE) {
+
+		/* Start with maximum frames available in FIFO since pcm->delay_ts. */
+		delay = fifo_delay + tframes;
+
+		/* Adjust by the change in frames in the buffer. */
+		if (io->state != SND_PCM_STATE_XRUN)
+			delay += io->buffer_size - snd_pcm_ioplug_hw_avail(io, pcm->delay_hw_ptr, io->appl_ptr);
+
+		/* impose upper limit */
+		snd_pcm_sframes_t limit = pcm->delay_fifo_size + io->buffer_size;
+		if (delay > limit)
+			delay = limit;
+
 	}
 	else {
-		/* Take the gap between the current input pointer in this thread and
-		 * the last thing transmitted in the I/O thread and adjust that by the
-		 * amount of time that has passed since something was transmitted. */
 
-		struct timespec now;
-		gettimestamp(&now);
+		delay = fifo_delay;
 
-		pthread_mutex_lock(&pcm->delay_mutex);
+		/* The buffer_delay is the number of frames that were in the buffer at
+		 * pcm->delay_ts, adjusted the number written by the application since
+		 * then. */
+		snd_pcm_sframes_t buffer_delay = 0;
+		if (io->state != SND_PCM_STATE_XRUN)
+			buffer_delay = snd_pcm_ioplug_hw_avail(io, pcm->delay_hw_ptr, io->appl_ptr);
 
-		struct timespec diff;
-		difftimespec(&now, &pcm->delay_ts, &diff);
+		/* If the PCM is running, then some frames from the buffer may have been
+		 * consumed, so we add them before adjusting for time elapsed. */
+		if (pcm->delay_running)
+			delay += buffer_delay;
 
-		delay = snd_pcm_ioplug_hw_avail(io, pcm->delay_hw_ptr, io->appl_ptr);
-
-		delay += pcm->delay_pcm_nread / pcm->frame_size;
-
-		unsigned int tsamples =
-			(diff.tv_sec * 1000 + diff.tv_nsec / 1000000) * io->rate / 1000;
-		/* Do not allow us to have a negative number of samples queued! */
-		if ((delay = delay - tsamples) < 0)
+		/* Adjust the total delay by the number of frames consumed. */
+		if ((delay -= tframes) < 0)
 			delay = 0;
 
-		pthread_mutex_unlock(&pcm->delay_mutex);
+		/* If the PCM is not running, then the frames in the buffer will not have
+		 * been consumed since pcm->delay_ts, so we add them after the time
+		 * elapsed adjustment. */
+		if (!pcm->delay_running)
+			delay += buffer_delay;
 	}
 
+	pthread_mutex_unlock(&pcm->mutex);
+
 	/* data transfer (communication) and encoding/decoding */
 	delay += (io->rate / 100) * pcm->ba_pcm.delay / 100;
 
@@ -545,6 +752,16 @@ static snd_pcm_sframes_t bluealsa_calculate_delay(snd_pcm_ioplug_t *io) {
 static int bluealsa_pause(snd_pcm_ioplug_t *io, int enable) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
+	if (enable == 1) {
+		/* Synchronize the IO thread with an application thread to ensure that
+		 * the server will not be paused while we are processing a transfer. */
+		pthread_mutex_lock(&pcm->mutex);
+		pcm->pause_state |= BA_PAUSE_STATE_PENDING;
+		while (!(pcm->pause_state & BA_PAUSE_STATE_PAUSED) && pcm->ba_pcm_fd != -1)
+			pthread_cond_wait(&pcm->pause_cond, &pcm->mutex);
+		pthread_mutex_unlock(&pcm->mutex);
+	}
+
 	if (!bluealsa_dbus_pcm_ctrl_send(pcm->ba_pcm_ctrl_fd,
 				enable ? "Pause" : "Resume", NULL))
 		return -errno;
@@ -581,23 +798,34 @@ static void bluealsa_dump(snd_pcm_ioplug_t *io, snd_output_t *out) {
 static int bluealsa_delay(snd_pcm_ioplug_t *io, snd_pcm_sframes_t *delayp) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
-	if (pcm->ba_pcm_fd == -1)
+	if (pcm->ba_pcm_fd == -1) {
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
 		return -ENODEV;
-
-	/* For capture PCMs only, the delay must be reported as zero when an overrun
-	 * occurs. */
-	if (io->stream == SND_PCM_STREAM_CAPTURE && io->state == SND_PCM_STATE_XRUN) {
-		*delayp = 0;
-		return 0;
 	}
 
-	/* When the PCM is paused, the delay value remains constant. */
-	if (io->state == SND_PCM_STATE_PAUSED)
-		*delayp = pcm->delay_paused;
-	else
-		*delayp = bluealsa_calculate_delay(io);
+	int ret = 0;
+	*delayp = 0;
 
-	return 0;
+	switch (io->state) {
+		case SND_PCM_STATE_PREPARED:
+		case SND_PCM_STATE_RUNNING:
+			*delayp = bluealsa_calculate_delay(io);
+			break;
+		case SND_PCM_STATE_PAUSED:
+			*delayp = pcm->delay_paused;
+			break;
+		case SND_PCM_STATE_XRUN:
+			*delayp = bluealsa_calculate_delay(io);
+			ret = -EPIPE;
+			break;
+		case SND_PCM_STATE_SUSPENDED:
+			ret = -ESTRPIPE;
+			break;
+		default:
+			break;
+	}
+
+	return ret;
 }
 
 static int bluealsa_poll_descriptors_count(snd_pcm_ioplug_t *io) {
@@ -631,9 +859,10 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 	*revents = 0;
 	int ret = 0;
 
-	if (bluealsa_dbus_connection_poll_dispatch(&pcm->dbus_ctx, &pfd[1], nfds - 1))
-		while (dbus_connection_dispatch(pcm->dbus_ctx.conn) == DBUS_DISPATCH_DATA_REMAINS)
-			continue;
+	bluealsa_dbus_connection_poll_dispatch(&pcm->dbus_ctx, &pfd[1], nfds - 1);
+	while (dbus_connection_dispatch(pcm->dbus_ctx.conn) == DBUS_DISPATCH_DATA_REMAINS)
+		continue;
+	gettimestamp(&pcm->dbus_dispatch_ts);
 
 	if (pcm->ba_pcm_fd == -1)
 		goto fail;
@@ -663,6 +892,13 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 				ready = false;
 				*revents = 0;
 				break;
+			case SND_PCM_STATE_PREPARED:
+				/* capture poll should block forever */
+				if (io->stream == SND_PCM_STREAM_CAPTURE) {
+					ready = false;
+					*revents = 0;
+				}
+				break;
 			case SND_PCM_STATE_RUNNING:
 				if ((snd_pcm_uframes_t)avail < pcm->io_avail_min) {
 					ready = false;
@@ -674,14 +910,12 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 			case SND_PCM_STATE_SUSPENDED:
 				*revents |= POLLERR;
 				break;
-			case SND_PCM_STATE_DISCONNECTED:
-				*revents = POLLERR;
-				ret = -ENODEV;
-				break;
 			case SND_PCM_STATE_OPEN:
 				*revents = POLLERR;
 				ret = -EBADF;
 				break;
+			case SND_PCM_STATE_DISCONNECTED:
+				goto fail;
 			default:
 				break;
 		};
@@ -694,6 +928,7 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 	return ret;
 
 fail:
+	snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
 	*revents = POLLERR | POLLHUP;
 	return -ENODEV;
 }
@@ -732,12 +967,43 @@ static int str2bdaddr(const char *str, bdaddr_t *ba) {
 
 static int str2profile(const char *str) {
 	if (strcasecmp(str, "a2dp") == 0)
-		return BA_PCM_FLAG_PROFILE_A2DP;
+		return BA_PCM_TRANSPORT_A2DP_SOURCE | BA_PCM_TRANSPORT_A2DP_SINK;
 	else if (strcasecmp(str, "sco") == 0)
-		return BA_PCM_FLAG_PROFILE_SCO;
+		return BA_PCM_TRANSPORT_HFP_AG | BA_PCM_TRANSPORT_HFP_HF |
+			BA_PCM_TRANSPORT_HSP_AG | BA_PCM_TRANSPORT_HSP_HS;
 	return 0;
 }
 
+/**
+ * Convert volume string to volume level and mute state.
+ *
+ * Mute state is determined by the last character of the volume
+ * string. The '-' at the end indicates to mute; '+' indicates
+ * to unmute. */
+static int str2volume(const char *str, int *volume, int *mute) {
+
+	char *endptr;
+	long v = strtol(str, &endptr, 10);
+	if (endptr != str) {
+		if (v < 0 || v > 100)
+			return -1;
+		*volume = v;
+	}
+
+	if (endptr[0] == '+' && endptr[1] == '\0')
+		*mute = 0;
+	else if (endptr[0] == '-' && endptr[1] == '\0')
+		*mute = 1;
+	else if (endptr[0] != '\0')
+		return -1;
+
+	return 0;
+}
+
+static int str2softvol(const char *str) {
+	return snd_config_get_bool_ascii(str);
+}
+
 static snd_pcm_format_t get_snd_pcm_format(uint16_t format) {
 	switch (format) {
 	case 0x0108:
@@ -756,6 +1022,33 @@ static snd_pcm_format_t get_snd_pcm_format(uint16_t format) {
 	}
 }
 
+static DBusHandlerResult bluealsa_dbus_msg_filter(DBusConnection *conn,
+		DBusMessage *message, void *data) {
+	struct bluealsa_pcm *pcm = (struct bluealsa_pcm *)data;
+	(void)conn;
+
+	if (dbus_message_get_type(message) != DBUS_MESSAGE_TYPE_SIGNAL)
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+	DBusMessageIter iter;
+	if (!dbus_message_iter_init(message, &iter))
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+	if (strcmp(dbus_message_get_path(message), pcm->ba_pcm.pcm_path) != 0 ||
+			strcmp(dbus_message_get_interface(message), DBUS_INTERFACE_PROPERTIES) != 0 ||
+			strcmp(dbus_message_get_member(message), "PropertiesChanged") != 0)
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+	const char *updated_interface;
+	dbus_message_iter_get_basic(&iter, &updated_interface);
+	dbus_message_iter_next(&iter);
+
+	if (strcmp(updated_interface, BLUEALSA_INTERFACE_PCM) == 0)
+		bluealsa_dbus_message_iter_get_pcm_props(&iter, NULL, &pcm->ba_pcm);
+
+	return DBUS_HANDLER_RESULT_HANDLED;
+}
+
 static int bluealsa_set_hw_constraint(struct bluealsa_pcm *pcm) {
 	snd_pcm_ioplug_t *io = &pcm->io;
 
@@ -790,7 +1083,11 @@ static int bluealsa_set_hw_constraint(struct bluealsa_pcm *pcm) {
 		snd_pcm_format_physical_width(get_snd_pcm_format(pcm->ba_pcm.format)) / 8;
 
 	if ((err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_PERIOD_BYTES,
-					min_p, 1024 * 16)) < 0)
+					min_p, 1024 * 1024)) < 0)
+		return err;
+
+	if ((err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_BUFFER_BYTES,
+					2 * min_p, 2 * 1024 * 1024)) < 0)
 		return err;
 
 	if ((err = snd_pcm_ioplug_set_param_minmax(io, SND_PCM_IOPLUG_HW_CHANNELS,
@@ -804,19 +1101,90 @@ static int bluealsa_set_hw_constraint(struct bluealsa_pcm *pcm) {
 	return 0;
 }
 
+static bool bluealsa_select_pcm_codec(struct bluealsa_pcm *pcm, const char *codec, DBusError *err) {
+
+	uint8_t config[64];
+	ssize_t config_len = 0;
+	bool ret = false;
+
+	char *codec_name;
+	if ((codec_name = strdup(codec)) == NULL) {
+		dbus_set_error(err, DBUS_ERROR_NO_MEMORY, NULL);
+		return false;
+	}
+
+	char *config_hex;
+	/* split the given string into name and configuration components */
+	if ((config_hex = strchr(codec_name, ':')) != NULL) {
+		*config_hex++ = '\0';
+
+		size_t config_hex_len;
+		if ((config_hex_len = strlen(config_hex)) > sizeof(config) * 2) {
+			dbus_set_error(err, DBUS_ERROR_FAILED, "Invalid codec configuration: %s", config_hex);
+			goto fail;
+		}
+		if ((config_len = hex2bin(config_hex, config, config_hex_len)) == -1) {
+			dbus_set_error(err, DBUS_ERROR_FAILED, "%s", strerror(errno));
+			goto fail;
+		}
+
+	}
+
+	if (!bluealsa_dbus_pcm_select_codec(&pcm->dbus_ctx, pcm->ba_pcm.pcm_path,
+				bluealsa_dbus_pcm_get_codec_canonical_name(codec_name), config, config_len, err))
+		goto fail;
+
+	ret = true;
+
+fail:
+	free(codec_name);
+	return ret;
+}
+
+static bool bluealsa_update_pcm_volume(struct bluealsa_pcm *pcm,
+		int volume, int mute, DBusError *err) {
+	uint16_t old = pcm->ba_pcm.volume.raw;
+	if (volume >= 0) {
+		const int v = BA_PCM_VOLUME_MAX(&pcm->ba_pcm) * volume / 100;
+		pcm->ba_pcm.volume.ch1_volume = v;
+		if (pcm->ba_pcm.channels == 2)
+			pcm->ba_pcm.volume.ch2_volume = v;
+	}
+	if (mute >= 0) {
+		const bool v = !!mute;
+		pcm->ba_pcm.volume.ch1_muted = v;
+		if (pcm->ba_pcm.channels == 2)
+			pcm->ba_pcm.volume.ch2_muted = v;
+	}
+	if (pcm->ba_pcm.volume.raw == old)
+		return true;
+	return bluealsa_dbus_pcm_update(&pcm->dbus_ctx, &pcm->ba_pcm, BLUEALSA_PCM_VOLUME, err);
+}
+
+static bool bluealsa_update_pcm_softvol(struct bluealsa_pcm *pcm,
+		int softvol, DBusError *err) {
+	if (softvol < 0 || !!softvol == pcm->ba_pcm.soft_volume)
+		return true;
+	pcm->ba_pcm.soft_volume = !!softvol;
+	return bluealsa_dbus_pcm_update(&pcm->dbus_ctx, &pcm->ba_pcm, BLUEALSA_PCM_SOFT_VOLUME, err);
+}
+
 SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	(void)root;
 
-	snd_config_iterator_t i, next;
 	const char *service = BLUEALSA_SERVICE;
 	const char *device = NULL;
 	const char *profile = NULL;
-	struct bluealsa_pcm *pcm;
+	const char *codec = NULL;
+	const char *volume = NULL;
+	const char *softvol = NULL;
 	long delay = 0;
+	struct bluealsa_pcm *pcm;
 	int ret;
 
-	snd_config_for_each(i, next, conf) {
-		snd_config_t *n = snd_config_iterator_entry(i);
+	snd_config_iterator_t pos, next;
+	snd_config_for_each(pos, next, conf) {
+		snd_config_t *n = snd_config_iterator_entry(pos);
 
 		const char *id;
 		if (snd_config_get_id(n, &id) < 0)
@@ -848,6 +1216,33 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 			}
 			continue;
 		}
+		if (strcmp(id, "codec") == 0) {
+			if (snd_config_get_string(n, &codec) < 0) {
+				SNDERR("Invalid type for %s", id);
+				return -EINVAL;
+			}
+			if (strcmp(codec, "unchanged") == 0)
+				codec = NULL;
+			continue;
+		}
+		if (strcmp(id, "volume") == 0) {
+			if (snd_config_get_string(n, &volume) < 0) {
+				SNDERR("Invalid type for %s", id);
+				return -EINVAL;
+			}
+			if (strcmp(volume, "unchanged") == 0)
+				volume = NULL;
+			continue;
+		}
+		if (strcmp(id, "softvol") == 0) {
+			if (snd_config_get_string(n, &softvol) < 0) {
+				SNDERR("Invalid type for %s", id);
+				return -EINVAL;
+			}
+			if (strcmp(softvol, "unchanged") == 0)
+				softvol = NULL;
+			continue;
+		}
 		if (strcmp(id, "delay") == 0) {
 			if (snd_config_get_integer(n, &delay) < 0) {
 				SNDERR("Invalid type for %s", id);
@@ -872,6 +1267,19 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 		return -EINVAL;
 	}
 
+	int pcm_mute = -1;
+	int pcm_volume = -1;
+	if (volume != NULL && str2volume(volume, &pcm_volume, &pcm_mute) != 0) {
+		SNDERR("Invalid volume [0-100][+-]: %s", volume);
+		return -EINVAL;
+	}
+
+	int pcm_softvol = -1;
+	if (softvol != NULL && (pcm_softvol = str2softvol(softvol)) < 0) {
+		SNDERR("Invalid softvol: %s", softvol);
+		return -EINVAL;
+	}
+
 	if ((pcm = calloc(1, sizeof(*pcm))) == NULL)
 		return -ENOMEM;
 
@@ -879,7 +1287,9 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	pcm->ba_pcm_fd = -1;
 	pcm->ba_pcm_ctrl_fd = -1;
 	pcm->delay_ex = delay;
-	pthread_mutex_init(&pcm->delay_mutex, NULL);
+	pthread_mutex_init(&pcm->mutex, NULL);
+	pthread_cond_init(&pcm->pause_cond, NULL);
+	pcm->pause_state = BA_PAUSE_STATE_RUNNING;
 
 	dbus_threads_init_default();
 
@@ -890,15 +1300,26 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 		goto fail;
 	}
 
-	int flags = ba_profile | (
-			stream == SND_PCM_STREAM_PLAYBACK ? BA_PCM_FLAG_SINK : BA_PCM_FLAG_SOURCE);
+	if (!dbus_connection_add_filter(pcm->dbus_ctx.conn, bluealsa_dbus_msg_filter, pcm, NULL)) {
+		SNDERR("Couldn't add D-Bus filter: %s", strerror(ENOMEM));
+		ret = -ENOMEM;
+		goto fail;
+	}
+
 	debug("Getting BlueALSA PCM: %s %s %s", snd_pcm_stream_name(stream), device, profile);
-	if (!bluealsa_dbus_get_pcm(&pcm->dbus_ctx, &ba_addr, flags, &pcm->ba_pcm, &err)) {
+	if (!bluealsa_dbus_get_pcm(&pcm->dbus_ctx, &ba_addr, ba_profile,
+				stream == SND_PCM_STREAM_PLAYBACK ? BA_PCM_MODE_SINK : BA_PCM_MODE_SOURCE,
+				&pcm->ba_pcm, &err)) {
 		SNDERR("Couldn't get BlueALSA PCM: %s", err.message);
 		ret = -ENODEV;
 		goto fail;
 	}
 
+	/* Subscribe for properties-changed signals but for the opened PCM only. */
+	bluealsa_dbus_connection_signal_match_add(&pcm->dbus_ctx, pcm->dbus_ctx.ba_service,
+			pcm->ba_pcm.pcm_path, DBUS_INTERFACE_PROPERTIES, "PropertiesChanged",
+			"arg0='"BLUEALSA_INTERFACE_PCM"'");
+
 	if ((pcm->event_fd = eventfd(0, EFD_CLOEXEC)) == -1) {
 		ret = -errno;
 		goto fail;
@@ -925,11 +1346,39 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	if ((ret = snd_pcm_ioplug_create(&pcm->io, name, stream, mode)) < 0)
 		goto fail;
 
+	if (codec != NULL && codec[0] != '\0') {
+		if (bluealsa_select_pcm_codec(pcm, codec, &err)) {
+			/* Changing the codec may change the audio format, sampling rate and/or
+			 * channels. We need to refresh our cache of PCM properties. */
+			if (!bluealsa_dbus_get_pcm(&pcm->dbus_ctx, &ba_addr, ba_profile,
+						stream == SND_PCM_STREAM_PLAYBACK ? BA_PCM_MODE_SINK : BA_PCM_MODE_SOURCE,
+						&pcm->ba_pcm, &err)) {
+				SNDERR("Couldn't get BlueALSA PCM: %s", err.message);
+				ret = -ENODEV;
+				goto fail;
+			}
+		}
+		else {
+			SNDERR("Couldn't select BlueALSA PCM codec: %s", err.message);
+			dbus_error_free(&err);
+		}
+	}
+
 	if ((ret = bluealsa_set_hw_constraint(pcm)) < 0) {
 		snd_pcm_ioplug_delete(&pcm->io);
 		return ret;
 	}
 
+	if (!bluealsa_update_pcm_softvol(pcm, pcm_softvol, &err)) {
+		SNDERR("Couldn't set BlueALSA PCM soft-volume: %s", err.message);
+		dbus_error_free(&err);
+	}
+
+	if (!bluealsa_update_pcm_volume(pcm, pcm_volume, pcm_mute, &err)) {
+		SNDERR("Couldn't set BlueALSA PCM volume: %s", err.message);
+		dbus_error_free(&err);
+	}
+
 	*pcmp = pcm->io.pcm;
 	return 0;
 
