@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluealsa-dbus.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -27,23 +27,151 @@
 #include <glib.h>
 
 #include "a2dp.h"
-#include "a2dp-codecs.h"
 #include "ba-adapter.h"
 #include "ba-device.h"
+#include "ba-transport.h"
+#include "bluealsa-config.h"
 #include "bluealsa-iface.h"
-#include "bluealsa.h"
+#include "bluealsa-skeleton.h"
 #include "dbus.h"
 #include "hfp.h"
 #include "utils.h"
+#include "shared/a2dp-codecs.h"
 #include "shared/defs.h"
 #include "shared/log.h"
 
-static GVariant *ba_variant_new_device_path(const struct ba_device *d) {
+static const char *bluealsa_dbus_manager_path = "/org/bluealsa";
+static GDBusObjectManagerServer *bluealsa_dbus_manager = NULL;
+
+static GVariant *ba_variant_new_bluealsa_version(void) {
+	return g_variant_new_string(PACKAGE_VERSION);
+}
+
+static GVariant *ba_variant_new_bluealsa_adapters(void) {
+
+	const char *strv[ARRAYSIZE(config.adapters)];
+	GVariant *variant;
+	size_t n = 0;
+
+	pthread_mutex_lock(&config.adapters_mutex);
+
+	for (size_t i = 0; i < ARRAYSIZE(config.adapters); i++)
+		if (config.adapters[i] != NULL)
+			strv[n++] = config.adapters[i]->hci.name;
+
+	variant = g_variant_new_strv(strv, n);
+
+	pthread_mutex_unlock(&config.adapters_mutex);
+
+	return variant;
+}
+
+static GVariant *ba_variant_new_bluealsa_profiles(void) {
+
+	struct {
+		const char *name;
+		bool enabled;
+	} profiles[] = {
+		{ BLUEALSA_TRANSPORT_TYPE_A2DP_SOURCE, config.profile.a2dp_source },
+		{ BLUEALSA_TRANSPORT_TYPE_A2DP_SINK, config.profile.a2dp_sink },
+#if ENABLE_OFONO
+		{ BLUEALSA_TRANSPORT_TYPE_HFP_OFONO, config.profile.hfp_ofono },
+#endif
+		{ BLUEALSA_TRANSPORT_TYPE_HFP_AG, config.profile.hfp_ag },
+		{ BLUEALSA_TRANSPORT_TYPE_HFP_HF, config.profile.hfp_hf },
+		{ BLUEALSA_TRANSPORT_TYPE_HSP_AG, config.profile.hsp_ag },
+		{ BLUEALSA_TRANSPORT_TYPE_HSP_HS, config.profile.hsp_hs },
+	};
+
+	const char *strv[ARRAYSIZE(profiles)];
+	size_t n = 0;
+
+	for (size_t i = 0; i < ARRAYSIZE(profiles); i++)
+		if (profiles[i].enabled)
+			strv[n++] = profiles[i].name;
+
+	return g_variant_new_strv(strv, n);
+}
+
+static GVariant *ba_variant_new_bluealsa_codecs(void) {
+
+	char tmp[64][32];
+	const char *strv[ARRAYSIZE(tmp)];
+	size_t n = 0;
+
+	const struct a2dp_codec * a2dp_codecs_tmp[32];
+	struct a2dp_codec * const * cc = a2dp_codecs;
+	for (const struct a2dp_codec *c = *cc; c != NULL; c = *++cc) {
+		if (!c->enabled)
+			continue;
+		a2dp_codecs_tmp[n] = c;
+		if (++n >= ARRAYSIZE(a2dp_codecs_tmp))
+			break;
+	}
+
+	/* Expose A2DP codecs always in the same order. */
+	a2dp_codecs_qsort(a2dp_codecs_tmp, n);
+
+	for (size_t i = 0; i < n; i++) {
+		const char *profile = a2dp_codecs_tmp[i]->dir == A2DP_SOURCE ?
+				BLUEALSA_TRANSPORT_TYPE_A2DP_SOURCE : BLUEALSA_TRANSPORT_TYPE_A2DP_SINK;
+		const char *name = a2dp_codecs_codec_id_to_string(a2dp_codecs_tmp[i]->codec_id);
+		snprintf(tmp[i], sizeof(tmp[i]), "%s:%s", profile, name);
+		strv[i] = (const char *)&tmp[i];
+	}
+
+	static const char *hfp_profiles[] = {
+#if ENABLE_OFONO
+		BLUEALSA_TRANSPORT_TYPE_HFP_OFONO,
+# endif
+		BLUEALSA_TRANSPORT_TYPE_HFP_AG,
+		BLUEALSA_TRANSPORT_TYPE_HFP_HF,
+	};
+
+	struct {
+		uint16_t codec_id;
+		bool enabled;
+	} hfp_codecs[] = {
+		{ HFP_CODEC_CVSD, config.hfp.codecs.cvsd },
+#if ENABLE_MSBC
+		{ HFP_CODEC_MSBC, config.hfp.codecs.msbc },
+#endif
+	};
+
+	for (size_t i = 0; i < ARRAYSIZE(hfp_profiles); i++)
+		for (size_t ii = 0; ii < ARRAYSIZE(hfp_codecs); ii++)
+			if (hfp_codecs[ii].enabled) {
+				const char *name = hfp_codec_id_to_string(hfp_codecs[ii].codec_id);
+				snprintf(tmp[n], sizeof(tmp[n]), "%s:%s", hfp_profiles[i], name);
+				strv[n] = (const char *)&tmp[n];
+				n++;
+			}
+
+	static const char *hsp_profiles[] = {
+		BLUEALSA_TRANSPORT_TYPE_HSP_AG,
+		BLUEALSA_TRANSPORT_TYPE_HSP_HS,
+	};
+
+	for (size_t i = 0; i < ARRAYSIZE(hsp_profiles); i++) {
+		const char *name = hfp_codec_id_to_string(HFP_CODEC_CVSD);
+		snprintf(tmp[n], sizeof(tmp[n]), "%s:%s", hsp_profiles[i], name);
+		strv[n] = (const char *)&tmp[n];
+		n++;
+	}
+
+	return g_variant_new_strv(strv, n);
+}
+
+GVariant *ba_variant_new_device_path(const struct ba_device *d) {
 	return g_variant_new_object_path(d->bluez_dbus_path);
 }
 
-static GVariant *ba_variant_new_device_battery(const struct ba_device *d) {
-	return g_variant_new_byte(d->battery_level);
+static GVariant *ba_variant_new_device_sequence(const struct ba_device *d) {
+	return g_variant_new_uint32(d->seq);
+}
+
+GVariant *ba_variant_new_device_battery(const struct ba_device *d) {
+	return g_variant_new_byte(d->battery.charge);
 }
 
 static GVariant *ba_variant_new_transport_type(const struct ba_transport *t) {
@@ -89,29 +217,38 @@ static GVariant *ba_variant_new_pcm_codec(const struct ba_transport_pcm *pcm) {
 	const struct ba_transport *t = pcm->t;
 	const char *codec = NULL;
 	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP)
-		codec = ba_transport_codecs_a2dp_to_string(t->type.codec);
+		codec = a2dp_codecs_codec_id_to_string(t->type.codec);
 	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO)
-		codec = ba_transport_codecs_hfp_to_string(t->type.codec);
+		codec = hfp_codec_id_to_string(t->type.codec);
 	if (codec != NULL)
 		return g_variant_new_string(codec);
 	return g_variant_new_string("<null>");
 }
 
 static GVariant *ba_variant_new_pcm_delay(const struct ba_transport_pcm *pcm) {
-	return g_variant_new_uint16(ba_transport_get_delay(pcm->t));
+	return g_variant_new_uint16(ba_transport_pcm_get_delay(pcm));
 }
 
 static GVariant *ba_variant_new_pcm_soft_volume(const struct ba_transport_pcm *pcm) {
 	return g_variant_new_boolean(pcm->soft_volume);
 }
 
+static uint8_t ba_volume_pack_dbus_volume(bool muted, int value) {
+	return (muted << 7) | (((uint8_t)value) & 0x7F);
+}
+
 static GVariant *ba_variant_new_pcm_volume(const struct ba_transport_pcm *pcm) {
-	return g_variant_new_uint16(ba_transport_pcm_get_volume_packed(pcm));
+	uint8_t ch1 = ba_volume_pack_dbus_volume(pcm->volume[0].scale == 0,
+			ba_transport_pcm_volume_level_to_bt(pcm, pcm->volume[0].level));
+	uint8_t ch2 = ba_volume_pack_dbus_volume(pcm->volume[1].scale == 0,
+			ba_transport_pcm_volume_level_to_bt(pcm, pcm->volume[1].level));
+	return g_variant_new_uint16((ch1 << 8) | (pcm->channels == 1 ? 0 : ch2));
 }
 
 static void ba_variant_populate_pcm(GVariantBuilder *props, const struct ba_transport_pcm *pcm) {
 	g_variant_builder_init(props, G_VARIANT_TYPE("a{sv}"));
 	g_variant_builder_add(props, "{sv}", "Device", ba_variant_new_device_path(pcm->t->d));
+	g_variant_builder_add(props, "{sv}", "Sequence", ba_variant_new_device_sequence(pcm->t->d));
 	g_variant_builder_add(props, "{sv}", "Transport", ba_variant_new_transport_type(pcm->t));
 	g_variant_builder_add(props, "{sv}", "Mode", ba_variant_new_pcm_mode(pcm));
 	g_variant_builder_add(props, "{sv}", "Format", ba_variant_new_pcm_format(pcm));
@@ -129,25 +266,23 @@ static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_se
 	if ((codec = a2dp_codec_lookup(sep->codec_id, !sep->dir)) == NULL)
 		return false;
 
-	uint8_t caps[32];
+	/* Do not report SEP if corresponding codec is not enabled
+	 * in BlueALSA - it will be not possible to use it. */
+	if (!codec->enabled)
+		return false;
+
+	uint8_t caps[sizeof(sep->capabilities)];
 	size_t size = MIN(sep->capabilities_size, sizeof(caps));
-	if (a2dp_filter_capabilities(codec, memcpy(caps, sep->capabilities, size), size) != 0) {
+	if (a2dp_filter_capabilities(codec, memcpy(caps, &sep->capabilities, size), size) != 0) {
 		error("Couldn't filter %s capabilities: %s",
-				ba_transport_codecs_a2dp_to_string(sep->codec_id),
+				a2dp_codecs_codec_id_to_string(sep->codec_id),
 				strerror(errno));
 		return false;
 	}
 
 	g_variant_builder_init(props, G_VARIANT_TYPE("a{sv}"));
-
-	GVariantBuilder vcaps;
-	g_variant_builder_init(&vcaps, G_VARIANT_TYPE("ay"));
-
-	size_t i;
-	for (i = 0; i < size; i++)
-		g_variant_builder_add(&vcaps, "y", caps[i]);
-
-	g_variant_builder_add(props, "{sv}", "Capabilities", g_variant_builder_end(&vcaps));
+	g_variant_builder_add(props, "{sv}", "Capabilities", g_variant_new_fixed_array(
+				G_VARIANT_TYPE_BYTE, caps, size, sizeof(uint8_t)));
 
 	switch (codec->codec_id) {
 	case A2DP_CODEC_SBC:
@@ -172,6 +307,10 @@ static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_se
 	case A2DP_CODEC_VENDOR_FASTSTREAM:
 		break;
 #endif
+#if ENABLE_LC3PLUS
+	case A2DP_CODEC_VENDOR_LC3PLUS:
+		break;
+#endif
 #if ENABLE_LDAC
 	case A2DP_CODEC_VENDOR_LDAC:
 		break;
@@ -183,7 +322,8 @@ static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_se
 	return true;
 }
 
-static void bluealsa_manager_get_pcms(GDBusMethodInvocation *inv) {
+static void bluealsa_manager_get_pcms(GDBusMethodInvocation *inv, void *userdata) {
+	(void)userdata;
 
 	GVariantBuilder pcms;
 	g_variant_builder_init(&pcms, G_VARIANT_TYPE("a{oa{sv}}"));
@@ -191,7 +331,7 @@ static void bluealsa_manager_get_pcms(GDBusMethodInvocation *inv) {
 	struct ba_adapter *a;
 	size_t i;
 
-	for (i = 0; i < HCI_MAX_DEV; i++) {
+	for (i = 0; i < ARRAYSIZE(config.adapters); i++) {
 
 		if ((a = ba_adapter_lookup(i)) == NULL)
 			continue;
@@ -211,9 +351,17 @@ static void bluealsa_manager_get_pcms(GDBusMethodInvocation *inv) {
 
 				if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
 
-					ba_variant_populate_pcm(&props, &t->a2dp.pcm);
-					g_variant_builder_add(&pcms, "{oa{sv}}", t->a2dp.pcm.ba_dbus_path, &props);
-					g_variant_builder_clear(&props);
+					if (t->a2dp.pcm.ba_dbus_exported) {
+						ba_variant_populate_pcm(&props, &t->a2dp.pcm);
+						g_variant_builder_add(&pcms, "{oa{sv}}", t->a2dp.pcm.ba_dbus_path, &props);
+						g_variant_builder_clear(&props);
+					}
+
+					if (t->a2dp.pcm_bc.ba_dbus_exported) {
+						ba_variant_populate_pcm(&props, &t->a2dp.pcm_bc);
+						g_variant_builder_add(&pcms, "{oa{sv}}", t->a2dp.pcm_bc.ba_dbus_path, &props);
+						g_variant_builder_clear(&props);
+					}
 
 				}
 				else if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO) {
@@ -242,31 +390,49 @@ static void bluealsa_manager_get_pcms(GDBusMethodInvocation *inv) {
 	g_variant_builder_clear(&pcms);
 }
 
-static void bluealsa_manager_method_call(GDBusConnection *conn, const char *sender,
-		const char *path, const char *interface, const char *method, GVariant *params,
-		GDBusMethodInvocation *invocation, void *userdata) {
-	(void)conn;
-	(void)params;
+static GVariant *bluealsa_manager_get_property(const char *property,
+		GError **error, void *userdata) {
+	(void)error;
 	(void)userdata;
 
+	if (strcmp(property, "Version") == 0)
+		return ba_variant_new_bluealsa_version();
+	if (strcmp(property, "Adapters") == 0)
+		return ba_variant_new_bluealsa_adapters();
+	if (strcmp(property, "Profiles") == 0)
+		return ba_variant_new_bluealsa_profiles();
+	if (strcmp(property, "Codecs") == 0)
+		return ba_variant_new_bluealsa_codecs();
+
+	g_assert_not_reached();
+	return NULL;
+}
+
+/**
+ * Register BlueALSA D-Bus manager interfaces. */
+void bluealsa_dbus_register(void) {
+
 	static const GDBusMethodCallDispatcher dispatchers[] = {
 		{ .method = "GetPCMs",
 			.handler = bluealsa_manager_get_pcms },
-		{ NULL },
+		{ 0 },
 	};
 
-	if (!g_dbus_dispatch_method_call(dispatchers, sender, path, interface, method, invocation))
-		error("Couldn't dispatch D-Bus method call: %s.%s()", interface, method);
+	static const GDBusInterfaceSkeletonVTable vtable = {
+		.dispatchers = dispatchers,
+		.get_property = bluealsa_manager_get_property,
+	};
 
-}
+	debug("Registering D-Bus manager: %s", bluealsa_dbus_manager_path);
+
+	bluealsa_ManagerIfaceSkeleton *ifs_manager;
+	ifs_manager = bluealsa_manager_iface_skeleton_new(&vtable, NULL, NULL);
+	g_dbus_interface_skeleton_export(G_DBUS_INTERFACE_SKELETON(ifs_manager),
+			config.dbus, bluealsa_dbus_manager_path, NULL);
+
+	bluealsa_dbus_manager = g_dbus_object_manager_server_new(bluealsa_dbus_manager_path);
+	g_dbus_object_manager_server_set_connection(bluealsa_dbus_manager, config.dbus);
 
-/**
- * Register BlueALSA D-Bus manager interface. */
-unsigned int bluealsa_dbus_manager_register(GError **error) {
-	static const GDBusInterfaceVTable vtable = {
-		.method_call = bluealsa_manager_method_call };
-	return g_dbus_connection_register_object(config.dbus, "/org/bluealsa",
-			(GDBusInterfaceInfo *)&bluealsa_iface_manager, &vtable, NULL, NULL, error);
 }
 
 static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
@@ -274,7 +440,6 @@ static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
 	(void)condition;
 
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
-	struct ba_transport *t = pcm->t;
 	char command[32];
 	size_t len;
 
@@ -310,8 +475,13 @@ static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
 	case G_IO_STATUS_AGAIN:
 		return TRUE;
 	case G_IO_STATUS_EOF:
+		pthread_mutex_lock(&pcm->mutex);
 		ba_transport_pcm_release(pcm);
-		ba_transport_send_signal(t, BA_TRANSPORT_SIGNAL_PCM_CLOSE);
+		ba_transport_thread_signal_send(pcm->th, BA_TRANSPORT_THREAD_SIGNAL_PCM_CLOSE);
+		pthread_mutex_unlock(&pcm->mutex);
+		/* Check whether we've just closed the last PCM client and in
+		 * such a case schedule transport IO threads termination. */
+		ba_transport_stop_if_no_clients(pcm->t);
 		/* remove channel from watch */
 		return FALSE;
 	}
@@ -319,16 +489,19 @@ static gboolean bluealsa_pcm_controller(GIOChannel *ch, GIOCondition condition,
 	return TRUE;
 }
 
-static void bluealsa_pcm_open(GDBusMethodInvocation *inv) {
+static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 	const bool is_sink = pcm->mode == BA_TRANSPORT_PCM_MODE_SINK;
+	struct ba_transport_thread *th = pcm->th;
 	struct ba_transport *t = pcm->t;
 	int pcm_fds[4] = { -1, -1, -1, -1 };
-	bool locked = false;
 	size_t i;
 
+	/* Prevent two (or more) clients trying to
+	 * open the same PCM at the same time. */
+	pthread_mutex_lock(&pcm->mutex);
+
 	/* preliminary check whether HFP codes is selected */
 	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO &&
 			t->type.codec == HFP_CODEC_UNDEFINED) {
@@ -337,12 +510,6 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv) {
 		goto fail;
 	}
 
-	/* We must ensure that transport release is not in progress before
-	 * accessing transport critical section. Otherwise, we might have
-	 * the IO thread close it in the middle of open procedure! */
-	ba_transport_pthread_cleanup_lock(t);
-	locked = true;
-
 	if (pcm->fd != -1) {
 		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 				G_DBUS_ERROR_FAILED, "%s", strerror(EBUSY));
@@ -370,15 +537,35 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv) {
 	 * the transport is acquired in order to extend battery life. For profiles
 	 * like A2DP Sink and HFP headset, we will wait for incoming connection. */
 	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE ||
-			t->type.profile & BA_TRANSPORT_PROFILE_MASK_AG)
-		if (t->acquire(t) == -1) {
+			t->type.profile & BA_TRANSPORT_PROFILE_MASK_AG) {
+
+		enum ba_transport_thread_state state;
+
+		if (ba_transport_acquire(t) == -1) {
 			g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 					G_DBUS_ERROR_FAILED, "Acquire transport: %s", strerror(errno));
 			goto fail;
 		}
 
+		/* wait until ready to process audio */
+		pthread_mutex_lock(&th->state_mtx);
+		while ((state = th->state) < BA_TRANSPORT_THREAD_STATE_RUNNING)
+			pthread_cond_wait(&th->changed, &th->state_mtx);
+		pthread_mutex_unlock(&th->state_mtx);
+
+		/* bail if something has gone wrong */
+		if (state != BA_TRANSPORT_THREAD_STATE_RUNNING) {
+			g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
+					G_DBUS_ERROR_IO_ERROR, "Acquire transport: %s", strerror(EIO));
+			goto fail;
+		}
+
+	}
+
 	/* get correct PIPE endpoint - PIPE is unidirectional */
 	pcm->fd = pcm_fds[is_sink ? 0 : 1];
+	/* set newly opened PCM as active */
+	pcm->active = true;
 
 	GIOChannel *ch = g_io_channel_unix_new(pcm_fds[2]);
 	g_io_add_watch_full(ch, G_PRIORITY_DEFAULT, G_IO_IN,
@@ -388,11 +575,10 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv) {
 	g_io_channel_set_encoding(ch, NULL, NULL);
 	g_io_channel_unref(ch);
 
-	/* notify our IO thread that the FIFO is ready */
-	ba_transport_send_signal(t, BA_TRANSPORT_SIGNAL_PCM_OPEN);
+	/* notify our audio thread that the FIFO is ready */
+	ba_transport_thread_signal_send(th, BA_TRANSPORT_THREAD_SIGNAL_PCM_OPEN);
 
-	ba_transport_pthread_cleanup_unlock(t);
-	ba_transport_pcm_unref(pcm);
+	pthread_mutex_unlock(&pcm->mutex);
 
 	int fds[2] = { pcm_fds[is_sink ? 1 : 0], pcm_fds[3] };
 	GUnixFDList *fd_list = g_unix_fd_list_new_from_array(fds, 2);
@@ -403,18 +589,15 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv) {
 	return;
 
 fail:
-	if (locked)
-		ba_transport_pthread_cleanup_unlock(t);
-	ba_transport_pcm_unref(pcm);
+	pthread_mutex_unlock(&pcm->mutex);
 	/* clean up created file descriptors */
 	for (i = 0; i < ARRAYSIZE(pcm_fds); i++)
 		if (pcm_fds[i] != -1)
 			close(pcm_fds[i]);
 }
 
-static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv) {
+static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata) {
 
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 	const struct ba_transport *t = pcm->t;
 	const GArray *seps = t->d->seps;
@@ -424,30 +607,53 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv) {
 
 	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
 
+		GArray *codec_ids = g_array_sized_new(FALSE, FALSE, sizeof(uint16_t), 16);
 		size_t i;
+
 		for (i = 0; seps != NULL && i < seps->len; i++) {
 			const struct a2dp_sep *sep = &g_array_index(seps, struct a2dp_sep, i);
 			/* match complementary PCM directions, e.g. A2DP-source with SEP-sink */
 			if (t->a2dp.codec->dir == !sep->dir) {
+
+				bool duplicate = false;
+				size_t j;
+
+				for (j = 0; j < codec_ids->len; j++)
+					if (sep->codec_id == g_array_index(codec_ids, uint16_t, j)) {
+						duplicate = true;
+						break;
+					}
+
+				/* do not return duplicates */
+				if (duplicate)
+					continue;
+
+				g_array_append_val(codec_ids, sep->codec_id);
+
 				GVariantBuilder props;
 				if (ba_variant_populate_sep(&props, sep)) {
 					g_variant_builder_add(&codecs, "{sa{sv}}",
-							ba_transport_codecs_a2dp_to_string(sep->codec_id), &props);
+							a2dp_codecs_codec_id_to_string(sep->codec_id), &props);
 					g_variant_builder_clear(&props);
 				}
+
 			}
 		}
 
+		g_array_free(codec_ids, TRUE);
+
 	}
 	else if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO) {
 
-		g_variant_builder_add(&codecs, "{sa{sv}}",
-				ba_transport_codecs_hfp_to_string(HFP_CODEC_CVSD), NULL);
+		if (config.hfp.codecs.cvsd)
+			g_variant_builder_add(&codecs, "{sa{sv}}",
+					hfp_codec_id_to_string(HFP_CODEC_CVSD), NULL);
 
 #if ENABLE_MSBC
-		if (t->sco.rfcomm->msbc)
+		if (config.hfp.codecs.msbc &&
+				t->sco.rfcomm != NULL && t->sco.rfcomm->codecs.msbc)
 			g_variant_builder_add(&codecs, "{sa{sv}}",
-					ba_transport_codecs_hfp_to_string(HFP_CODEC_MSBC), NULL);
+					hfp_codec_id_to_string(HFP_CODEC_MSBC), NULL);
 #endif
 
 	}
@@ -455,25 +661,23 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv) {
 	g_dbus_method_invocation_return_value(inv, g_variant_new("(a{sa{sv}})", &codecs));
 	g_variant_builder_clear(&codecs);
 
-	ba_transport_pcm_unref(pcm);
 }
 
-static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv) {
+static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata) {
 
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 	struct ba_transport *t = pcm->t;
 	GVariantIter *properties;
 	GVariant *value = NULL;
 	const char *errmsg = NULL;
+	const char *codec_name;
 	const char *property;
-	const char *codec;
 
-	void *a2dp_configuration = NULL;
+	a2dp_t a2dp_configuration = {};
 	size_t a2dp_configuration_size = 0;
 
-	g_variant_get(params, "(sa{sv})", &codec, &properties);
+	g_variant_get(params, "(sa{sv})", &codec_name, &properties);
 	while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
 
 		if (strcmp(property, "Configuration") == 0 &&
@@ -482,8 +686,13 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv) {
 			const void *data = g_variant_get_fixed_array(value,
 					&a2dp_configuration_size, sizeof(char));
 
-			g_free(a2dp_configuration);
-			a2dp_configuration = g_memdup(data, a2dp_configuration_size);
+			if (a2dp_configuration_size > sizeof(a2dp_configuration)) {
+				warn("Configuration blob size exceeded: %zu > %zu",
+						a2dp_configuration_size, sizeof(a2dp_configuration));
+				a2dp_configuration_size = sizeof(a2dp_configuration);
+			}
+
+			memcpy(&a2dp_configuration, data, a2dp_configuration_size);
 
 		}
 
@@ -499,7 +708,7 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv) {
 			goto fail;
 		}
 
-		uint16_t codec_id = ba_transport_codecs_a2dp_from_string(codec);
+		uint16_t codec_id = a2dp_codecs_codec_id_from_string(codec_name);
 		enum a2dp_dir dir = !t->a2dp.codec->dir;
 		const GArray *seps = t->d->seps;
 		struct a2dp_sep *sep = NULL;
@@ -525,18 +734,18 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv) {
 		}
 
 		/* setup default codec configuration */
-		memcpy(sep->configuration, sep->capabilities, sep->capabilities_size);
-		if (a2dp_select_configuration(codec, sep->configuration, sep->capabilities_size) == -1)
+		memcpy(&sep->configuration, &sep->capabilities, sep->capabilities_size);
+		if (a2dp_select_configuration(codec, &sep->configuration, sep->capabilities_size) == -1)
 			goto fail;
 
 		/* use codec configuration blob provided by user */
-		if (a2dp_configuration != NULL) {
-			if (a2dp_check_configuration(codec, a2dp_configuration,
+		if (a2dp_configuration_size != 0) {
+			if (a2dp_check_configuration(codec, &a2dp_configuration,
 						a2dp_configuration_size) != A2DP_CHECK_OK) {
 				errmsg = "Invalid configuration blob";
 				goto fail;
 			}
-			memcpy(sep->configuration, a2dp_configuration, sep->capabilities_size);
+			memcpy(&sep->configuration, &a2dp_configuration, sep->capabilities_size);
 		}
 
 		if (ba_transport_select_codec_a2dp(t, sep) == -1)
@@ -545,7 +754,7 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv) {
 	}
 	else {
 
-		uint16_t codec_id = ba_transport_codecs_hfp_from_string(codec);
+		uint16_t codec_id = hfp_codec_id_from_string(codec_name);
 		if (ba_transport_select_codec_sco(t, codec_id) == -1)
 			goto fail;
 
@@ -557,50 +766,18 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv) {
 fail:
 	if (errmsg == NULL)
 		errmsg = strerror(errno);
-	error("Couldn't select codec: %s: %s", codec, errmsg);
+	error("Couldn't select codec: %s: %s", codec_name, errmsg);
 	g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 			G_DBUS_ERROR_FAILED, "%s", errmsg);
 
 final:
-	ba_transport_pcm_unref(pcm);
-	g_free(a2dp_configuration);
 	g_variant_iter_free(properties);
 	if (value != NULL)
 		g_variant_unref(value);
 }
 
-static void bluealsa_pcm_method_call(GDBusConnection *conn, const char *sender,
-		const char *path, const char *interface, const char *method, GVariant *params,
-		GDBusMethodInvocation *invocation, void *userdata) {
-	(void)conn;
-	(void)params;
+static void bluealsa_rfcomm_open(GDBusMethodInvocation *inv, void *userdata) {
 
-	static const GDBusMethodCallDispatcher dispatchers[] = {
-		{ .method = "Open",
-			.handler = bluealsa_pcm_open,
-			.asynchronous_call = true },
-		{ .method = "GetCodecs",
-			.handler = bluealsa_pcm_get_codecs,
-			.asynchronous_call = true },
-		{ .method = "SelectCodec",
-			.handler = bluealsa_pcm_select_codec,
-			.asynchronous_call = true },
-		{ NULL },
-	};
-
-	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
-	ba_transport_pcm_ref(pcm);
-
-	if (!g_dbus_dispatch_method_call(dispatchers, sender, path, interface, method, invocation)) {
-		error("Couldn't dispatch D-Bus method call: %s.%s()", interface, method);
-		ba_transport_pcm_unref(pcm);
-	}
-
-}
-
-static void bluealsa_rfcomm_open(GDBusMethodInvocation *inv) {
-
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
 	int fds[2] = { -1, -1 };
 
@@ -625,37 +802,28 @@ static void bluealsa_rfcomm_open(GDBusMethodInvocation *inv) {
 	g_object_unref(fd_list);
 }
 
-static void bluealsa_rfcomm_method_call(GDBusConnection *conn, const char *sender,
-		const char *path, const char *interface, const char *method, GVariant *params,
-		GDBusMethodInvocation *invocation, void *userdata) {
-	(void)conn;
-	(void)params;
-	(void)userdata;
+static GVariant *bluealsa_pcm_get_properties(void *userdata) {
 
-	static const GDBusMethodCallDispatcher dispatchers[] = {
-		{ .method = "Open",
-			.handler = bluealsa_rfcomm_open },
-		{ NULL },
-	};
+	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 
-	if (!g_dbus_dispatch_method_call(dispatchers, sender, path, interface, method, invocation))
-		error("Couldn't dispatch D-Bus method call: %s.%s()", interface, method);
+	GVariantBuilder props;
 
+	ba_variant_populate_pcm(&props, pcm);
+
+	return g_variant_builder_end(&props);
 }
 
-static GVariant *bluealsa_pcm_get_property(GDBusConnection *conn,
-		const char *sender, const char *path, const char *interface,
-		const char *property, GError **error, void *userdata) {
-	(void)conn;
-	(void)sender;
-	(void)path;
-	(void)interface;
+static GVariant *bluealsa_pcm_get_property(const char *property,
+		GError **error, void *userdata) {
+	(void)error;
 
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 	struct ba_device *d = pcm->t->d;
 
 	if (strcmp(property, "Device") == 0)
 		return ba_variant_new_device_path(d);
+	if (strcmp(property, "Sequence") == 0)
+		return ba_variant_new_device_sequence(d);
 	if (strcmp(property, "Transport") == 0)
 		return ba_variant_new_transport_type(pcm->t);
 	if (strcmp(property, "Mode") == 0)
@@ -675,42 +843,13 @@ static GVariant *bluealsa_pcm_get_property(GDBusConnection *conn,
 	if (strcmp(property, "Volume") == 0)
 		return ba_variant_new_pcm_volume(pcm);
 
-	*error = g_error_new(G_DBUS_ERROR, G_DBUS_ERROR_NOT_SUPPORTED,
-			"Property not supported '%s'", property);
+	g_assert_not_reached();
 	return NULL;
 }
 
-static GVariant *bluealsa_rfcomm_get_property(GDBusConnection *conn,
-		const char *sender, const char *path, const char *interface,
-		const char *property, GError **error, void *userdata) {
-	(void)conn;
-	(void)sender;
-	(void)path;
-	(void)interface;
-
-	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
-	struct ba_transport *t = r->sco;
-	struct ba_device *d = t->d;
-
-	if (strcmp(property, "Transport") == 0)
-		return ba_variant_new_transport_type(t);
-	if (strcmp(property, "Features") == 0)
-		return ba_variant_new_rfcomm_features(r);
-	if (strcmp(property, "Battery") == 0)
-		return ba_variant_new_device_battery(d);
-
-	*error = g_error_new(G_DBUS_ERROR, G_DBUS_ERROR_NOT_SUPPORTED,
-			"Property not supported '%s'", property);
-	return NULL;
-}
-
-static gboolean bluealsa_pcm_set_property(GDBusConnection *conn,
-		const gchar *sender, const gchar *path, const gchar *interface,
-		const gchar *property, GVariant *value, GError **error, gpointer userdata) {
-	(void)conn;
-	(void)sender;
-	(void)path;
-	(void)interface;
+static bool bluealsa_pcm_set_property(const char *property, GVariant *value,
+		GError **error, void *userdata) {
+	(void)error;
 
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 
@@ -719,43 +858,85 @@ static gboolean bluealsa_pcm_set_property(GDBusConnection *conn,
 		bluealsa_dbus_pcm_update(pcm, BA_DBUS_PCM_UPDATE_SOFT_VOLUME);
 		return TRUE;
 	}
+
 	if (strcmp(property, "Volume") == 0) {
-		ba_transport_pcm_set_volume_packed(pcm, g_variant_get_uint16(value));
+
+		uint16_t packed = g_variant_get_uint16(value);
+		uint8_t ch1 = packed >> 8;
+		uint8_t ch2 = packed & 0xFF;
+
+		int ch1_level = ba_transport_pcm_volume_bt_to_level(pcm, ch1 & 0x7F);
+		bool ch1_muted = !!(ch1 & 0x80);
+		int ch2_level = ba_transport_pcm_volume_bt_to_level(pcm, ch2 & 0x7F);
+		bool ch2_muted = !!(ch2 & 0x80);
+
+		ba_transport_pcm_volume_set(&pcm->volume[0], &ch1_level, &ch1_muted, NULL);
+		ba_transport_pcm_volume_set(&pcm->volume[1], &ch2_level, &ch2_muted, NULL);
+
+		debug("Setting volume: %u [%.2f dB] %c%c %u [%.2f dB]",
+				ch1 & 0x7F, 0.01 * ch1_level, ch1_muted ? 'x' : '<',
+				ch2_muted ? 'x' : '>', ch2 & 0x7F, 0.01 * ch2_level);
+
+		ba_transport_pcm_volume_update(pcm);
 		return TRUE;
 	}
 
-	*error = g_error_new(G_DBUS_ERROR, G_DBUS_ERROR_NOT_SUPPORTED,
-			"Property not supported '%s'", property);
+	g_assert_not_reached();
 	return FALSE;
 }
 
 /**
  * Register BlueALSA D-Bus PCM interface. */
-unsigned int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm, GError **error) {
+int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm) {
 
-	static const GDBusInterfaceVTable vtable = {
-		.method_call = bluealsa_pcm_method_call,
+	static const GDBusMethodCallDispatcher dispatchers[] = {
+		{ .method = "Open",
+			.handler = bluealsa_pcm_open },
+		{ .method = "GetCodecs",
+			.handler = bluealsa_pcm_get_codecs },
+		{ .method = "SelectCodec",
+			.handler = bluealsa_pcm_select_codec },
+		{ 0 },
+	};
+
+	static const GDBusInterfaceSkeletonVTable vtable = {
+		.dispatchers = dispatchers,
+		.get_properties = bluealsa_pcm_get_properties,
 		.get_property = bluealsa_pcm_get_property,
 		.set_property = bluealsa_pcm_set_property,
 	};
 
-	if ((pcm->ba_dbus_id = g_dbus_connection_register_object(config.dbus,
-					pcm->ba_dbus_path, (GDBusInterfaceInfo *)&bluealsa_iface_pcm, &vtable,
-					pcm, (GDestroyNotify)ba_transport_pcm_unref, error)) != 0) {
+	GDBusObjectSkeleton *skeleton = NULL;
+	bluealsa_PCMIfaceSkeleton *ifs_pcm = NULL;
 
-		ba_transport_pcm_ref(pcm);
+	if ((skeleton = g_dbus_object_skeleton_new(pcm->ba_dbus_path)) == NULL)
+		goto fail;
 
-		GVariantBuilder props;
-		ba_variant_populate_pcm(&props, pcm);
+	if ((ifs_pcm = bluealsa_pcm_iface_skeleton_new(&vtable,
+					pcm, (GDestroyNotify)ba_transport_pcm_unref)) == NULL)
+		goto fail;
 
-		g_dbus_connection_emit_signal(config.dbus, NULL,
-				"/org/bluealsa", BLUEALSA_IFACE_MANAGER, "PCMAdded",
-				g_variant_new("(oa{sv})", pcm->ba_dbus_path, &props), NULL);
-		g_variant_builder_clear(&props);
+	ba_transport_pcm_ref(pcm);
 
-	}
+	g_dbus_object_skeleton_add_interface(skeleton, G_DBUS_INTERFACE_SKELETON(ifs_pcm));
+	g_dbus_object_manager_server_export(bluealsa_dbus_manager, skeleton);
+	pcm->ba_dbus_exported = true;
 
-	return pcm->ba_dbus_id;
+	GVariantBuilder props;
+	ba_variant_populate_pcm(&props, pcm);
+	g_dbus_connection_emit_signal(config.dbus, NULL,
+			bluealsa_dbus_manager_path, BLUEALSA_IFACE_MANAGER, "PCMAdded",
+			g_variant_new("(oa{sv})", pcm->ba_dbus_path, &props), NULL);
+	g_variant_builder_clear(&props);
+
+fail:
+
+	if (skeleton != NULL)
+		g_object_unref(skeleton);
+	if (ifs_pcm != NULL)
+		g_object_unref(ifs_pcm);
+
+	return 0;
 }
 
 void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
@@ -778,41 +959,99 @@ void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
 	if (mask & BA_DBUS_PCM_UPDATE_VOLUME)
 		g_variant_builder_add(&props, "{sv}", "Volume", ba_variant_new_pcm_volume(pcm));
 
-	g_dbus_connection_emit_signal(config.dbus, NULL, pcm->ba_dbus_path,
-			DBUS_IFACE_PROPERTIES, "PropertiesChanged",
-			g_variant_new("(sa{sv}as)", BLUEALSA_IFACE_PCM, &props, NULL), NULL);
-
+	g_dbus_connection_emit_properties_changed(config.dbus,
+			pcm->ba_dbus_path, BLUEALSA_IFACE_PCM, &props, NULL);
 	g_variant_builder_clear(&props);
+
 }
 
 void bluealsa_dbus_pcm_unregister(struct ba_transport_pcm *pcm) {
 
-	if (pcm->ba_dbus_id == 0)
+	if (!pcm->ba_dbus_exported)
 		return;
 
-	g_dbus_connection_unregister_object(config.dbus, pcm->ba_dbus_id);
-	pcm->ba_dbus_id = 0;
+	g_dbus_object_manager_server_unexport(bluealsa_dbus_manager, pcm->ba_dbus_path);
+	pcm->ba_dbus_exported = false;
 
 	g_dbus_connection_emit_signal(config.dbus, NULL,
-			"/org/bluealsa", BLUEALSA_IFACE_MANAGER, "PCMRemoved",
+			bluealsa_dbus_manager_path, BLUEALSA_IFACE_MANAGER, "PCMRemoved",
 			g_variant_new("(o)", pcm->ba_dbus_path), NULL);
 
 }
 
+static GVariant *bluealsa_rfcomm_get_properties(void *userdata) {
+
+	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
+	struct ba_transport *t = r->sco;
+	struct ba_device *d = t->d;
+
+	GVariantBuilder props;
+	g_variant_builder_init(&props, G_VARIANT_TYPE("a{sv}"));
+
+	g_variant_builder_add(&props, "{sv}", "Transport", ba_variant_new_transport_type(t));
+	g_variant_builder_add(&props, "{sv}", "Features", ba_variant_new_rfcomm_features(r));
+	g_variant_builder_add(&props, "{sv}", "Battery", ba_variant_new_device_battery(d));
+
+	return g_variant_builder_end(&props);
+}
+
+static GVariant *bluealsa_rfcomm_get_property(const char *property,
+		GError **error, void *userdata) {
+	(void)error;
+
+	struct ba_rfcomm *r = (struct ba_rfcomm *)userdata;
+	struct ba_transport *t = r->sco;
+	struct ba_device *d = t->d;
+
+	if (strcmp(property, "Transport") == 0)
+		return ba_variant_new_transport_type(t);
+	if (strcmp(property, "Features") == 0)
+		return ba_variant_new_rfcomm_features(r);
+	if (strcmp(property, "Battery") == 0)
+		return ba_variant_new_device_battery(d);
+
+	g_assert_not_reached();
+	return NULL;
+}
+
 /**
  * Register BlueALSA D-Bus RFCOMM interface. */
-unsigned int bluealsa_dbus_rfcomm_register(struct ba_rfcomm *r, GError **error) {
+int bluealsa_dbus_rfcomm_register(struct ba_rfcomm *r) {
 
-	static const GDBusInterfaceVTable vtable = {
-		.method_call = bluealsa_rfcomm_method_call,
+	static const GDBusMethodCallDispatcher dispatchers[] = {
+		{ .method = "Open",
+			.handler = bluealsa_rfcomm_open },
+		{ 0 },
+	};
+
+	static const GDBusInterfaceSkeletonVTable vtable = {
+		.dispatchers = dispatchers,
+		.get_properties = bluealsa_rfcomm_get_properties,
 		.get_property = bluealsa_rfcomm_get_property,
 	};
 
-	r->ba_dbus_id = g_dbus_connection_register_object(config.dbus,
-			r->ba_dbus_path, (GDBusInterfaceInfo *)&bluealsa_iface_rfcomm,
-			&vtable, r, NULL, error);
+	GDBusObjectSkeleton *skeleton = NULL;
+	bluealsa_RFCOMMIfaceSkeleton *ifs_rfcomm = NULL;
 
-	return r->ba_dbus_id;
+	if ((skeleton = g_dbus_object_skeleton_new(r->ba_dbus_path)) == NULL)
+		goto fail;
+
+	if ((ifs_rfcomm = bluealsa_rfcomm_iface_skeleton_new(&vtable,
+					r, NULL)) == NULL)
+		goto fail;
+
+	g_dbus_object_skeleton_add_interface(skeleton, G_DBUS_INTERFACE_SKELETON(ifs_rfcomm));
+	g_dbus_object_manager_server_export(bluealsa_dbus_manager, skeleton);
+	r->ba_dbus_exported = true;
+
+fail:
+
+	if (skeleton != NULL)
+		g_object_unref(skeleton);
+	if (ifs_rfcomm != NULL)
+		g_object_unref(ifs_rfcomm);
+
+	return 0;
 }
 
 void bluealsa_dbus_rfcomm_update(struct ba_rfcomm *r, unsigned int mask) {
@@ -825,16 +1064,15 @@ void bluealsa_dbus_rfcomm_update(struct ba_rfcomm *r, unsigned int mask) {
 	if (mask & BA_DBUS_RFCOMM_UPDATE_BATTERY)
 		g_variant_builder_add(&props, "{sv}", "Battery", ba_variant_new_device_battery(r->sco->d));
 
-	g_dbus_connection_emit_signal(config.dbus, NULL, r->ba_dbus_path,
-			DBUS_IFACE_PROPERTIES, "PropertiesChanged",
-			g_variant_new("(sa{sv}as)", BLUEALSA_IFACE_RFCOMM, &props, NULL), NULL);
-
+	g_dbus_connection_emit_properties_changed(config.dbus,
+			r->ba_dbus_path, BLUEALSA_IFACE_RFCOMM, &props, NULL);
 	g_variant_builder_clear(&props);
+
 }
 
 void bluealsa_dbus_rfcomm_unregister(struct ba_rfcomm *r) {
-	if (r->ba_dbus_id == 0)
+	if (!r->ba_dbus_exported)
 		return;
-	g_dbus_connection_unregister_object(config.dbus, r->ba_dbus_id);
-	r->ba_dbus_id = 0;
+	g_dbus_object_manager_server_unexport(bluealsa_dbus_manager, r->ba_dbus_path);
+	r->ba_dbus_exported = false;
 }
