@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluealsa-dbus.h
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2021 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -8,6 +8,7 @@
  *
  */
 
+#pragma once
 #ifndef BLUEALSA_BLUEALSADBUS_H_
 #define BLUEALSA_BLUEALSADBUS_H_
 
@@ -18,6 +19,7 @@
 #include <glib.h>
 
 #include "ba-rfcomm.h"
+#include "ba-device.h"
 #include "ba-transport.h"
 
 #define BA_DBUS_PCM_UPDATE_FORMAT      (1 << 0)
@@ -31,13 +33,16 @@
 #define BA_DBUS_RFCOMM_UPDATE_FEATURES (1 << 0)
 #define BA_DBUS_RFCOMM_UPDATE_BATTERY  (1 << 1)
 
-unsigned int bluealsa_dbus_manager_register(GError **error);
+GVariant *ba_variant_new_device_path(const struct ba_device *d);
+GVariant *ba_variant_new_device_battery(const struct ba_device *d);
 
-unsigned int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm, GError **error);
+void bluealsa_dbus_register(void);
+
+int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm);
 void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask);
 void bluealsa_dbus_pcm_unregister(struct ba_transport_pcm *pcm);
 
-unsigned int bluealsa_dbus_rfcomm_register(struct ba_rfcomm *r, GError **error);
+int bluealsa_dbus_rfcomm_register(struct ba_rfcomm *r);
 void bluealsa_dbus_rfcomm_update(struct ba_rfcomm *r, unsigned int mask);
 void bluealsa_dbus_rfcomm_unregister(struct ba_rfcomm *r);
 
