@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluealsa-iface.h
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -8,6 +8,7 @@
  *
  */
 
+#pragma once
 #ifndef BLUEALSA_BLUEALSAIFACE_H_
 #define BLUEALSA_BLUEALSAIFACE_H_
 
@@ -19,12 +20,16 @@
 #define BLUEALSA_IFACE_PCM     BLUEALSA_SERVICE ".PCM1"
 #define BLUEALSA_IFACE_RFCOMM  BLUEALSA_SERVICE ".RFCOMM1"
 
-#define BLUEALSA_TRANSPORT_TYPE_A2DP_SINK   "A2DP-sink"
-#define BLUEALSA_TRANSPORT_TYPE_A2DP_SOURCE "A2DP-source"
-#define BLUEALSA_TRANSPORT_TYPE_HFP_AG      "HFP-AG"
-#define BLUEALSA_TRANSPORT_TYPE_HFP_HF      "HFP-HF"
-#define BLUEALSA_TRANSPORT_TYPE_HSP_AG      "HSP-AG"
-#define BLUEALSA_TRANSPORT_TYPE_HSP_HS      "HSP-HS"
+#define BLUEALSA_TRANSPORT_TYPE_A2DP        "A2DP"
+#define BLUEALSA_TRANSPORT_TYPE_A2DP_SOURCE BLUEALSA_TRANSPORT_TYPE_A2DP "-source"
+#define BLUEALSA_TRANSPORT_TYPE_A2DP_SINK   BLUEALSA_TRANSPORT_TYPE_A2DP "-sink"
+#define BLUEALSA_TRANSPORT_TYPE_HFP         "HFP"
+#define BLUEALSA_TRANSPORT_TYPE_HFP_OFONO   BLUEALSA_TRANSPORT_TYPE_HFP "-oFono"
+#define BLUEALSA_TRANSPORT_TYPE_HFP_AG      BLUEALSA_TRANSPORT_TYPE_HFP "-AG"
+#define BLUEALSA_TRANSPORT_TYPE_HFP_HF      BLUEALSA_TRANSPORT_TYPE_HFP "-HF"
+#define BLUEALSA_TRANSPORT_TYPE_HSP         "HSP"
+#define BLUEALSA_TRANSPORT_TYPE_HSP_AG      BLUEALSA_TRANSPORT_TYPE_HSP "-AG"
+#define BLUEALSA_TRANSPORT_TYPE_HSP_HS      BLUEALSA_TRANSPORT_TYPE_HSP "-HS"
 
 #define BLUEALSA_PCM_CTRL_DRAIN  "Drain"
 #define BLUEALSA_PCM_CTRL_DROP   "Drop"
