@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluez.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -12,6 +12,7 @@
 
 #include <errno.h>
 #include <pthread.h>
+#include <stdatomic.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -29,43 +30,45 @@
 #include <glib.h>
 
 #include "a2dp.h"
-#include "a2dp-codecs.h"
 #include "ba-adapter.h"
 #include "ba-device.h"
 #include "ba-transport.h"
-#include "bluealsa.h"
+#include "bluealsa-config.h"
 #include "bluealsa-dbus.h"
 #include "bluez-iface.h"
+#include "bluez-skeleton.h"
 #include "dbus.h"
 #include "hci.h"
-#include "sbc.h"
 #include "sco.h"
 #include "utils.h"
+#include "shared/a2dp-codecs.h"
 #include "shared/defs.h"
 #include "shared/log.h"
 
-/* Compatibility patch for glib < 2.42. */
-#ifndef G_DBUS_ERROR_UNKNOWN_OBJECT
-# define G_DBUS_ERROR_UNKNOWN_OBJECT G_DBUS_ERROR_FAILED
+/* Compatibility patch for glib < 2.68. */
+#if !GLIB_CHECK_VERSION(2, 68, 0)
+# define g_memdup2 g_memdup
 #endif
 
 /**
  * Data associated with registered D-Bus object. */
 struct bluez_dbus_object_data {
-	/* D-Bus object registration ID */
-	unsigned int id;
 	/* D-Bus object registration path */
 	char path[64];
+	/* exported interface skeleton */
+	GDBusInterfaceSkeleton *ifs;
 	/* associated adapter */
 	int hci_dev_id;
-	const struct a2dp_codec *codec;
+	/* the type of the transport */
 	struct ba_transport_type ttype;
+	/* media endpoint codec */
+	const struct a2dp_codec *codec;
 	/* determine whether object is registered in BlueZ */
 	bool registered;
 	/* determine whether object is used */
 	bool connected;
 	/* memory self-management */
-	int ref_count;
+	atomic_int ref_count;
 };
 
 /**
@@ -73,22 +76,47 @@ struct bluez_dbus_object_data {
 struct bluez_adapter {
 	/* reference to the adapter structure */
 	struct ba_adapter *adapter;
+	/* manager for battery provider objects */
+	GDBusObjectManagerServer *battery_manager;
 	/* array of end-points for connected devices */
 	GHashTable *device_sep_map;
 };
 
 static pthread_mutex_t bluez_mutex = PTHREAD_MUTEX_INITIALIZER;
 static GHashTable *dbus_object_data_map = NULL;
-static struct bluez_adapter bluez_adapters[HCI_MAX_DEV] = { NULL };
+static struct bluez_adapter bluez_adapters[HCI_MAX_DEV] = { 0 };
 
 #define bluez_adapters_device_lookup(hci_dev_id, addr) \
 	g_hash_table_lookup(bluez_adapters[hci_dev_id].device_sep_map, addr)
 #define bluez_adapters_device_get_sep(seps, i) \
 	g_array_index(seps, struct a2dp_sep, i)
 
+static void bluez_register_a2dp_all(struct ba_adapter *);
+static void bluez_register_battery_provider_manager(struct bluez_adapter *);
+
+static struct bluez_adapter *bluez_adapter_new(struct ba_adapter *a) {
+	bluez_adapters[a->hci.dev_id].adapter = a;
+	bluez_adapters[a->hci.dev_id].device_sep_map = g_hash_table_new_full(
+			g_bdaddr_hash, g_bdaddr_equal, g_free, (GDestroyNotify)g_array_unref);
+	bluez_register_battery_provider_manager(&bluez_adapters[a->hci.dev_id]);
+	bluez_register_a2dp_all(a);
+	return &bluez_adapters[a->hci.dev_id];
+}
+
+static void bluez_adapter_free(struct bluez_adapter *adapter) {
+	if (adapter->adapter == NULL)
+		return;
+	g_hash_table_destroy(adapter->device_sep_map);
+	adapter->device_sep_map = NULL;
+	g_object_unref(adapter->battery_manager);
+	adapter->battery_manager = NULL;
+	ba_adapter_destroy(adapter->adapter);
+	adapter->adapter = NULL;
+}
+
 static void bluez_dbus_object_data_unref(
 		struct bluez_dbus_object_data *obj) {
-	if (--obj->ref_count != 0)
+	if (atomic_fetch_sub_explicit(&obj->ref_count, 1, memory_order_relaxed) > 1)
 		return;
 	free(obj);
 }
@@ -132,52 +160,40 @@ static enum bluez_a2dp_transport_state bluez_a2dp_transport_state_from_string(
 	return BLUEZ_A2DP_TRANSPORT_STATE_IDLE;
 }
 
-static void bluez_endpoint_select_configuration(GDBusMethodInvocation *inv) {
+static void bluez_endpoint_select_configuration(GDBusMethodInvocation *inv, void *userdata) {
 
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 	const struct a2dp_codec *codec = dbus_obj->codec;
 
 	const void *data;
-	void *capabilities;
+	a2dp_t capabilities = {};
 	size_t size = 0;
 
 	params = g_variant_get_child_value(params, 0);
 	data = g_variant_get_fixed_array(params, &size, sizeof(char));
-	capabilities = g_memdup(data, size);
+	memcpy(&capabilities, data, MIN(size, sizeof(capabilities)));
 	g_variant_unref(params);
 
-	if (a2dp_select_configuration(codec, capabilities, size) == -1)
+	hexdump("A2DP peer capabilities blob", &capabilities, size, true);
+	if (a2dp_select_configuration(codec, &capabilities, size) == -1)
 		goto fail;
 
-	GVariantBuilder caps;
-	size_t i;
+	GVariant *rv[] = {
+		g_variant_new_fixed_array(G_VARIANT_TYPE_BYTE, &capabilities, size, sizeof(uint8_t)) };
+	g_dbus_method_invocation_return_value(inv, g_variant_new_tuple(rv, 1));
 
-	g_variant_builder_init(&caps, G_VARIANT_TYPE("ay"));
-	for (i = 0; i < size; i++)
-		g_variant_builder_add(&caps, "y", ((char *)capabilities)[i]);
-
-	g_dbus_method_invocation_return_value(inv, g_variant_new("(ay)", &caps));
-	g_variant_builder_clear(&caps);
-
-	goto final;
+	return;
 
 fail:
 	g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 			G_DBUS_ERROR_INVALID_ARGS, "Invalid capabilities");
-
-final:
-	g_free(capabilities);
 }
 
-static void bluez_register_a2dp_all(struct ba_adapter *adapter);
-
-static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv) {
+static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv, void *userdata) {
 
 	const char *sender = g_dbus_method_invocation_get_sender(inv);
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 	const struct a2dp_codec *codec = dbus_obj->codec;
 	const uint16_t codec_id = codec->codec_id;
@@ -188,7 +204,7 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv) {
 
 	enum bluez_a2dp_transport_state state = 0xFFFF;
 	char *device_path = NULL;
-	void *configuration = NULL;
+	a2dp_t configuration = {};
 	uint16_t volume = 127;
 	uint16_t delay = 150;
 
@@ -222,6 +238,7 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv) {
 
 			size_t size = 0;
 			const void *data = g_variant_get_fixed_array(value, &size, sizeof(char));
+			memcpy(&configuration, data, MIN(size, sizeof(configuration)));
 
 			uint32_t rv;
 			if ((rv = a2dp_check_configuration(codec, data, size)) != A2DP_CHECK_OK) {
@@ -229,9 +246,6 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv) {
 				goto fail;
 			}
 
-			g_free(configuration);
-			configuration = g_memdup(data, size);
-
 		}
 		else if (strcmp(property, "State") == 0 &&
 				g_variant_validate_value(value, G_VARIANT_TYPE_STRING, property)) {
@@ -278,20 +292,30 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv) {
 	}
 
 	if ((t = ba_transport_new_a2dp(d, dbus_obj->ttype,
-					sender, transport_path, codec, configuration)) == NULL) {
+					sender, transport_path, codec, &configuration)) == NULL) {
 		error("Couldn't create new transport: %s", strerror(errno));
 		goto fail;
 	}
 
+	/* Skip volume level initialization in case of A2DP Source
+	 * profile and software volume control. */
+	if (!(t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE &&
+				t->a2dp.pcm.soft_volume)) {
+		int level = ba_transport_pcm_volume_bt_to_level(&t->a2dp.pcm, volume);
+		ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[0], &level, NULL, NULL);
+		ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[1], &level, NULL, NULL);
+	}
+
 	t->a2dp.bluez_dbus_sep_path = dbus_obj->path;
-	t->a2dp.pcm.volume[0].level = volume;
-	t->a2dp.pcm.volume[1].level = volume;
 	t->a2dp.delay = delay;
+	t->a2dp.volume = volume;
 
 	debug("%s configured for device %s",
 			ba_transport_type_to_string(t->type),
 			batostr_(&d->addr));
-	debug("Configuration: channels: %u, sampling: %u",
+	hexdump("A2DP selected configuration blob",
+			&configuration, codec->capabilities_size, true);
+	debug("PCM configuration: channels: %u, sampling: %u",
 			t->a2dp.pcm.channels, t->a2dp.pcm.sampling);
 
 	ba_transport_set_a2dp_state(t, state);
@@ -316,13 +340,11 @@ final:
 	if (value != NULL)
 		g_variant_unref(value);
 	g_free(device_path);
-	g_free(configuration);
 }
 
-static void bluez_endpoint_clear_configuration(GDBusMethodInvocation *inv) {
+static void bluez_endpoint_clear_configuration(GDBusMethodInvocation *inv, void *userdata) {
 
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 
 	struct ba_adapter *a = NULL;
@@ -354,9 +376,8 @@ fail:
 	g_object_unref(inv);
 }
 
-static void bluez_endpoint_release(GDBusMethodInvocation *inv) {
+static void bluez_endpoint_release(GDBusMethodInvocation *inv, void *userdata) {
 
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 
 	debug("Releasing media endpoint: %s", dbus_obj->path);
@@ -366,30 +387,6 @@ static void bluez_endpoint_release(GDBusMethodInvocation *inv) {
 	g_object_unref(inv);
 }
 
-static void bluez_endpoint_method_call(GDBusConnection *conn, const char *sender,
-		const char *path, const char *interface, const char *method, GVariant *params,
-		GDBusMethodInvocation *invocation, void *userdata) {
-	(void)conn;
-	(void)params;
-	(void)userdata;
-
-	static const GDBusMethodCallDispatcher dispatchers[] = {
-		{ .method = "SelectConfiguration",
-			.handler = bluez_endpoint_select_configuration },
-		{ .method = "SetConfiguration",
-			.handler = bluez_endpoint_set_configuration },
-		{ .method = "ClearConfiguration",
-			.handler = bluez_endpoint_clear_configuration },
-		{ .method = "Release",
-			.handler = bluez_endpoint_release },
-		{ NULL },
-	};
-
-	if (!g_dbus_dispatch_method_call(dispatchers, sender, path, interface, method, invocation))
-		error("Couldn't dispatch D-Bus method call: %s.%s()", interface, method);
-
-}
-
 /**
  * Register media endpoint in BlueZ. */
 static int bluez_register_media_endpoint(
@@ -401,26 +398,20 @@ static int bluez_register_media_endpoint(
 	const struct a2dp_codec *codec = dbus_obj->codec;
 	GDBusMessage *msg = NULL, *rep = NULL;
 	int ret = 0;
-	size_t i;
 
 	debug("Registering media endpoint: %s", dbus_obj->path);
 
 	msg = g_dbus_message_new_method_call(BLUEZ_SERVICE, adapter->bluez_dbus_path,
 			BLUEZ_IFACE_MEDIA, "RegisterEndpoint");
 
-	GVariantBuilder caps;
 	GVariantBuilder properties;
-
-	g_variant_builder_init(&caps, G_VARIANT_TYPE("ay"));
 	g_variant_builder_init(&properties, G_VARIANT_TYPE("a{sv}"));
 
-	for (i = 0; i < codec->capabilities_size; i++)
-		g_variant_builder_add(&caps, "y", ((char *)codec->capabilities)[i]);
-
 	g_variant_builder_add(&properties, "{sv}", "UUID", g_variant_new_string(uuid));
 	g_variant_builder_add(&properties, "{sv}", "DelayReporting", g_variant_new_boolean(TRUE));
 	g_variant_builder_add(&properties, "{sv}", "Codec", g_variant_new_byte(codec->codec_id));
-	g_variant_builder_add(&properties, "{sv}", "Capabilities", g_variant_builder_end(&caps));
+	g_variant_builder_add(&properties, "{sv}", "Capabilities", g_variant_new_fixed_array(
+				G_VARIANT_TYPE_BYTE, &codec->capabilities, codec->capabilities_size, sizeof(uint8_t)));
 
 	g_dbus_message_set_body(msg, g_variant_new("(oa{sv})", dbus_obj->path, &properties));
 	g_variant_builder_clear(&properties);
@@ -455,8 +446,20 @@ static void bluez_register_a2dp(
 		const struct a2dp_codec *codec,
 		const char *uuid) {
 
-	static const GDBusInterfaceVTable vtable = {
-		.method_call = bluez_endpoint_method_call,
+	static const GDBusMethodCallDispatcher dispatchers[] = {
+		{ .method = "SelectConfiguration",
+			.handler = bluez_endpoint_select_configuration },
+		{ .method = "SetConfiguration",
+			.handler = bluez_endpoint_set_configuration },
+		{ .method = "ClearConfiguration",
+			.handler = bluez_endpoint_clear_configuration },
+		{ .method = "Release",
+			.handler = bluez_endpoint_release },
+		{ 0 },
+	};
+
+	static const GDBusInterfaceSkeletonVTable vtable = {
+		.dispatchers = dispatchers,
 	};
 
 	struct ba_transport_type ttype = {
@@ -500,9 +503,16 @@ static void bluez_register_a2dp(
 			dbus_obj->ttype = ttype;
 			dbus_obj->ref_count = 2;
 
-			if ((dbus_obj->id = g_dbus_connection_register_object(config.dbus,
-							path, (GDBusInterfaceInfo *)&bluez_iface_endpoint, &vtable,
-							dbus_obj, (GDestroyNotify)bluez_dbus_object_data_unref, &err)) == 0) {
+			bluez_MediaEndpointIfaceSkeleton *ifs_endpoint;
+			if ((ifs_endpoint = bluez_media_endpoint_iface_skeleton_new(&vtable,
+							dbus_obj, (GDestroyNotify)bluez_dbus_object_data_unref)) == NULL) {
+				free(dbus_obj);
+				goto fail;
+			}
+
+			dbus_obj->ifs = G_DBUS_INTERFACE_SKELETON(ifs_endpoint);
+			if (!g_dbus_interface_skeleton_export(dbus_obj->ifs, config.dbus, path, &err)) {
+				g_object_unref(ifs_endpoint);
 				free(dbus_obj);
 				goto fail;
 			}
@@ -537,17 +547,15 @@ fail:
  * Register A2DP endpoints. */
 static void bluez_register_a2dp_all(struct ba_adapter *adapter) {
 
-	const struct a2dp_codec **cc = a2dp_codecs;
-
-	while (*cc != NULL) {
-		const struct a2dp_codec *c = *cc++;
+	struct a2dp_codec * const * cc = a2dp_codecs;
+	for (const struct a2dp_codec *c = *cc; c != NULL; c = *++cc) {
 		switch (c->dir) {
 		case A2DP_SOURCE:
-			if (config.enable.a2dp_source)
+			if (config.profile.a2dp_source && c->enabled)
 				bluez_register_a2dp(adapter, c, BLUETOOTH_UUID_A2DP_SOURCE);
 			break;
 		case A2DP_SINK:
-			if (config.enable.a2dp_sink)
+			if (config.profile.a2dp_sink && c->enabled)
 				bluez_register_a2dp(adapter, c, BLUETOOTH_UUID_A2DP_SINK);
 			break;
 		}
@@ -555,12 +563,162 @@ static void bluez_register_a2dp_all(struct ba_adapter *adapter) {
 
 }
 
-static void bluez_profile_new_connection(GDBusMethodInvocation *inv) {
+static GVariant *bluez_battery_provider_iface_skeleton_get_properties(
+		void *userdata) {
+
+	const struct ba_device *d = userdata;
+
+	GVariantBuilder props;
+	g_variant_builder_init(&props, G_VARIANT_TYPE("a{sv}"));
+
+	g_variant_builder_add(&props, "{sv}", "Device", ba_variant_new_device_path(d));
+	g_variant_builder_add(&props, "{sv}", "Percentage", ba_variant_new_device_battery(d));
+	g_variant_builder_add(&props, "{sv}", "Source", g_variant_new_string("BlueALSA"));
+
+	return g_variant_builder_end(&props);
+}
+
+static GVariant *bluez_battery_provider_iface_skeleton_get_property(
+		const char *property, GError **error, void *userdata) {
+	(void)error;
+
+	const struct ba_device *d = userdata;
+
+	if (strcmp(property, "Device") == 0)
+		return ba_variant_new_device_path(d);
+	if (strcmp(property, "Percentage") == 0)
+		return ba_variant_new_device_battery(d);
+	if (strcmp(property, "Source") == 0)
+		return g_variant_new_string("BlueALSA");
+
+	g_assert_not_reached();
+	return NULL;
+}
+
+/**
+ * Add battery to battery provider. */
+static bool bluez_battery_provider_manager_add(struct ba_device *device) {
+
+	static const GDBusInterfaceSkeletonVTable vtable = {
+		.get_properties = bluez_battery_provider_iface_skeleton_get_properties,
+		.get_property = bluez_battery_provider_iface_skeleton_get_property,
+	};
+
+	struct ba_adapter *a = device->a;
+	GDBusObjectManagerServer *manager = bluez_adapters[a->hci.dev_id].battery_manager;
+
+	if (device->ba_battery_dbus_path != NULL)
+		return true;
+
+	GDBusObjectSkeleton *skeleton = NULL;
+	bluez_BatteryProviderIfaceSkeleton *ifs_battery_provider = NULL;
+
+	char *path = g_strdup_printf("/org/bluez/%s/battery/%s",
+			a->hci.name, device->addr_dbus_str);
+	if ((skeleton = g_dbus_object_skeleton_new(path)) == NULL)
+		goto fail;
+	if ((ifs_battery_provider = bluez_battery_provider_iface_skeleton_new(&vtable,
+				device, (GDestroyNotify)ba_device_unref)) == NULL)
+		goto fail;
+
+	ba_device_ref(device);
+
+	g_dbus_object_skeleton_add_interface(skeleton,
+			G_DBUS_INTERFACE_SKELETON(ifs_battery_provider));
+	g_object_unref(ifs_battery_provider);
+
+	debug("Adding battery to battery provider: %s", path);
+
+	device->ba_battery_dbus_path = path;
+
+	g_dbus_object_manager_server_export(manager, skeleton);
+	g_object_unref(skeleton);
+
+	return true;
+
+fail:
+	if (skeleton != NULL)
+		g_object_unref(skeleton);
+	if (ifs_battery_provider != NULL)
+		g_object_unref(ifs_battery_provider);
+	g_free(path);
+	return false;
+}
+
+/**
+ * Remove battery from battery provider. */
+static bool bluez_battery_provider_manager_remove(struct ba_device *device) {
+
+	struct ba_adapter *a = device->a;
+	GDBusObjectManagerServer *manager = bluez_adapters[a->hci.dev_id].battery_manager;
+
+	if (device->ba_battery_dbus_path == NULL)
+		return true;
+
+	char *path = device->ba_battery_dbus_path;
+	device->ba_battery_dbus_path = NULL;
+
+	debug("Removing battery from battery provider: %s", path);
+	g_dbus_object_manager_server_unexport(manager, path);
+	g_free(path);
+
+	return true;
+}
+
+/**
+ * Register battery provider in BlueZ. */
+static void bluez_register_battery_provider_manager(struct bluez_adapter *b_adapter) {
+
+	struct ba_adapter *a = b_adapter->adapter;
+	GDBusMessage *msg = NULL, *rep = NULL;
+	GError *err = NULL;
+
+	char path[64];
+	snprintf(path, sizeof(path), "/org/bluez/%s/battery", a->hci.name);
+
+	debug("Registering battery provider: %s", path);
+
+	GDBusObjectManagerServer *manager = g_dbus_object_manager_server_new(path);
+	g_dbus_object_manager_server_set_connection(manager, config.dbus);
+	b_adapter->battery_manager = manager;
+
+	msg = g_dbus_message_new_method_call(BLUEZ_SERVICE, a->bluez_dbus_path,
+			BLUEZ_IFACE_BATTERY_PROVIDER_MANAGER, "RegisterBatteryProvider");
+
+	g_dbus_message_set_body(msg, g_variant_new("(o)", path));
+
+	if ((rep = g_dbus_connection_send_message_with_reply_sync(config.dbus, msg,
+					G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &err)) == NULL)
+		goto fail;
+
+	if (g_dbus_message_get_message_type(rep) == G_DBUS_MESSAGE_TYPE_ERROR) {
+		g_dbus_message_to_gerror(rep, &err);
+		if (err->code == G_DBUS_ERROR_UNKNOWN_METHOD) {
+			/* Suppress warning message in case when BlueZ has no battery provider
+			 * support enabled, because it's not a mandatory feature. */
+			debug("BlueZ battery provider support not available");
+			g_error_free(err);
+			err = NULL;
+		}
+		goto fail;
+	}
+
+fail:
+	if (err != NULL) {
+		warn("Couldn't register battery provider: %s", err->message);
+		g_error_free(err);
+	}
+	if (msg != NULL)
+		g_object_unref(msg);
+	if (rep != NULL)
+		g_object_unref(rep);
+}
+
+static void bluez_profile_new_connection(GDBusMethodInvocation *inv, void *userdata) {
 
 	GDBusMessage *msg = g_dbus_method_invocation_get_message(inv);
 	const char *sender = g_dbus_method_invocation_get_sender(inv);
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 
 	struct ba_adapter *a = NULL;
@@ -610,7 +768,6 @@ static void bluez_profile_new_connection(GDBusMethodInvocation *inv) {
 			ba_transport_type_to_string(t->type),
 			batostr_(&d->addr));
 
-	ba_transport_start(t);
 	dbus_obj->connected = true;
 
 	g_dbus_method_invocation_return_value(inv, NULL);
@@ -634,10 +791,9 @@ final:
 		g_error_free(err);
 }
 
-static void bluez_profile_request_disconnection(GDBusMethodInvocation *inv) {
+static void bluez_profile_request_disconnection(GDBusMethodInvocation *inv, void *userdata) {
 
 	GVariant *params = g_dbus_method_invocation_get_parameters(inv);
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 
 	debug("Disconnecting hands-free profile: %s", dbus_obj->path);
@@ -670,9 +826,8 @@ fail:
 	g_object_unref(inv);
 }
 
-static void bluez_profile_release(GDBusMethodInvocation *inv) {
+static void bluez_profile_release(GDBusMethodInvocation *inv, void *userdata) {
 
-	void *userdata = g_dbus_method_invocation_get_user_data(inv);
 	struct bluez_dbus_object_data *dbus_obj = userdata;
 
 	debug("Releasing hands-free profile: %s", dbus_obj->path);
@@ -682,28 +837,6 @@ static void bluez_profile_release(GDBusMethodInvocation *inv) {
 	g_object_unref(inv);
 }
 
-static void bluez_profile_method_call(GDBusConnection *conn, const char *sender,
-		const char *path, const char *interface, const char *method, GVariant *params,
-		GDBusMethodInvocation *invocation, void *userdata) {
-	(void)conn;
-	(void)params;
-	(void)userdata;
-
-	static const GDBusMethodCallDispatcher dispatchers[] = {
-		{ .method = "NewConnection",
-			.handler = bluez_profile_new_connection },
-		{ .method = "RequestDisconnection",
-			.handler = bluez_profile_request_disconnection },
-		{ .method = "Release",
-			.handler = bluez_profile_release },
-		{ NULL },
-	};
-
-	if (!g_dbus_dispatch_method_call(dispatchers, sender, path, interface, method, invocation))
-		error("Couldn't dispatch D-Bus method call: %s.%s()", interface, method);
-
-}
-
 /**
  * Register hands-free profile in BlueZ. */
 static int bluez_register_profile(
@@ -763,8 +896,18 @@ static void bluez_register_hfp(
 		uint16_t version,
 		uint16_t features) {
 
-	static const GDBusInterfaceVTable vtable = {
-		.method_call = bluez_profile_method_call,
+	static const GDBusMethodCallDispatcher dispatchers[] = {
+		{ .method = "NewConnection",
+			.handler = bluez_profile_new_connection },
+		{ .method = "RequestDisconnection",
+			.handler = bluez_profile_request_disconnection },
+		{ .method = "Release",
+			.handler = bluez_profile_release },
+		{ 0 },
+	};
+
+	static const GDBusInterfaceSkeletonVTable vtable = {
+		.dispatchers = dispatchers,
 	};
 
 	struct ba_transport_type ttype = {
@@ -794,9 +937,16 @@ static void bluez_register_hfp(
 		dbus_obj->ttype = ttype;
 		dbus_obj->ref_count = 2;
 
-		if ((dbus_obj->id = g_dbus_connection_register_object(config.dbus,
-						path, (GDBusInterfaceInfo *)&bluez_iface_profile, &vtable,
-						dbus_obj, (GDestroyNotify)bluez_dbus_object_data_unref, &err)) == 0) {
+		bluez_ProfileIfaceSkeleton *ifs_profile;
+		if ((ifs_profile = bluez_profile_iface_skeleton_new(&vtable,
+						dbus_obj, (GDestroyNotify)bluez_dbus_object_data_unref)) == NULL) {
+			free(dbus_obj);
+			goto fail;
+		}
+
+		dbus_obj->ifs = G_DBUS_INTERFACE_SKELETON(ifs_profile);
+		if (!g_dbus_interface_skeleton_export(dbus_obj->ifs, config.dbus, path, &err)) {
+			g_object_unref(ifs_profile);
 			free(dbus_obj);
 			goto fail;
 		}
@@ -829,36 +979,23 @@ fail:
  * is controlled by the global configuration structure - if none is enabled,
  * this function will do nothing. */
 static void bluez_register_hfp_all(void) {
-	if (config.enable.hsp_hs)
+	if (config.profile.hsp_hs)
 		bluez_register_hfp(BLUETOOTH_UUID_HSP_HS, BA_TRANSPORT_PROFILE_HSP_HS,
 				0x0102 /* HSP 1.2 */, 0x1 /* remote audio volume control */);
-	if (config.enable.hsp_ag)
+	if (config.profile.hsp_ag)
 		bluez_register_hfp(BLUETOOTH_UUID_HSP_AG, BA_TRANSPORT_PROFILE_HSP_AG,
 				0x0102 /* HSP 1.2 */, 0x0);
-	if (config.enable.hfp_hf)
+	if (config.profile.hfp_hf)
 		bluez_register_hfp(BLUETOOTH_UUID_HFP_HF, BA_TRANSPORT_PROFILE_HFP_HF,
 				0x0107 /* HFP 1.7 */, config.hfp.features_sdp_hf);
-	if (config.enable.hfp_ag)
+	if (config.profile.hfp_ag)
 		bluez_register_hfp(BLUETOOTH_UUID_HFP_AG, BA_TRANSPORT_PROFILE_HFP_AG,
 				0x0107 /* HFP 1.7 */, config.hfp.features_sdp_ag);
 }
 
-static void bluez_sep_array_free(GArray *seps) {
-	size_t i;
-	for (i = 0; i < seps->len; i++) {
-		g_free(bluez_adapters_device_get_sep(seps, i).capabilities);
-		g_free(bluez_adapters_device_get_sep(seps, i).configuration);
-	}
-	g_array_unref(seps);
-}
-
 /**
  * Register to the BlueZ service. */
-void bluez_register(void) {
-
-	if (dbus_object_data_map == NULL)
-		dbus_object_data_map = g_hash_table_new_full(g_str_hash, g_str_equal,
-				NULL, (GDestroyNotify)bluez_dbus_object_data_unref);
+static void bluez_register(void) {
 
 	GError *err = NULL;
 	GVariantIter *objects = NULL;
@@ -898,10 +1035,7 @@ void bluez_register(void) {
 	for (i = 0; i < ARRAYSIZE(adapters); i++)
 		if (adapters[i] &&
 				(a = ba_adapter_new(i)) != NULL) {
-			bluez_adapters[a->hci.dev_id].adapter = a;
-			bluez_adapters[a->hci.dev_id].device_sep_map = g_hash_table_new_full(
-					g_bdaddr_hash, g_bdaddr_equal, g_free, (GDestroyNotify)bluez_sep_array_free);
-			bluez_register_a2dp_all(a);
+			bluez_adapter_new(a);
 		}
 
 	/* HFP has to be registered globally */
@@ -935,14 +1069,24 @@ static void bluez_signal_interfaces_added(GDBusConnection *conn, const char *sen
 
 	g_variant_get(params, "(&oa{sa{sv}})", &object_path, &interfaces);
 	while (g_variant_iter_next(interfaces, "{&sa{sv}}", &interface, &properties)) {
-		if (strcmp(interface, BLUEZ_IFACE_ADAPTER) == 0)
+		if (strcmp(interface, BLUEZ_IFACE_ADAPTER) == 0) {
 			while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
 				if (strcmp(property, "Address") == 0 &&
+						/* make sure that this new BT adapter matches our HCI filter */
 						bluez_match_dbus_adapter(object_path, g_variant_get_string(value, NULL)))
 					hci_dev_id = g_dbus_bluez_object_path_to_hci_dev_id(object_path);
 				g_variant_unref(value);
 			}
-		else if (strcmp(interface, BLUEZ_IFACE_MEDIA_ENDPOINT) == 0)
+		}
+		else if (strcmp(interface, BLUEZ_IFACE_MEDIA_ENDPOINT) == 0) {
+
+			/* Check whether this new media endpoint interface was added in the HCI
+			 * which exists in our local BlueZ adapter cache - HCI that matches our
+			 * HCI filter. */
+			int dev_id = g_dbus_bluez_object_path_to_hci_dev_id(object_path);
+			if (dev_id == -1 || bluez_adapters[dev_id].adapter == NULL)
+				continue;
+
 			while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
 				if (strcmp(property, "UUID") == 0) {
 					const char *uuid = g_variant_get_string(value, NULL);
@@ -956,12 +1100,19 @@ static void bluez_signal_interfaces_added(GDBusConnection *conn, const char *sen
 					const void *data = g_variant_get_fixed_array(value,
 							&sep.capabilities_size, sizeof(char));
 
-					g_free(sep.capabilities);
-					sep.capabilities = g_memdup(data, sep.capabilities_size);
+					if (sep.capabilities_size > sizeof(sep.capabilities)) {
+						warn("Capabilities blob size exceeded: %zu > %zu",
+								sep.capabilities_size, sizeof(sep.capabilities));
+						sep.capabilities_size = sizeof(sep.capabilities);
+					}
+
+					memcpy(&sep.capabilities, data, sep.capabilities_size);
 
 				}
 				g_variant_unref(value);
+
 			}
+		}
 		g_variant_iter_free(properties);
 	}
 	g_variant_iter_free(interfaces);
@@ -969,10 +1120,7 @@ static void bluez_signal_interfaces_added(GDBusConnection *conn, const char *sen
 	struct ba_adapter *a;
 	if (hci_dev_id != -1 &&
 			(a = ba_adapter_new(hci_dev_id)) != NULL) {
-		bluez_adapters[a->hci.dev_id].adapter = a;
-		bluez_adapters[a->hci.dev_id].device_sep_map = g_hash_table_new_full(
-				g_bdaddr_hash, g_bdaddr_equal, g_free, (GDestroyNotify)bluez_sep_array_free);
-		bluez_register_a2dp_all(a);
+		bluez_adapter_new(a);
 	}
 
 	/* HFP has to be registered globally */
@@ -983,20 +1131,19 @@ static void bluez_signal_interfaces_added(GDBusConnection *conn, const char *sen
 
 		bdaddr_t addr;
 		g_dbus_bluez_object_path_to_bdaddr(object_path, &addr);
-		int hci_dev_id = g_dbus_bluez_object_path_to_hci_dev_id(object_path);
+		int dev_id = g_dbus_bluez_object_path_to_hci_dev_id(object_path);
 
 		GArray *seps;
-		if ((seps = bluez_adapters_device_lookup(hci_dev_id, &addr)) == NULL)
-			g_hash_table_insert(bluez_adapters[hci_dev_id].device_sep_map,
-					g_memdup(&addr, sizeof(addr)), seps = g_array_new(FALSE, FALSE, sizeof(sep)));
+		if ((seps = bluez_adapters_device_lookup(dev_id, &addr)) == NULL)
+			g_hash_table_insert(bluez_adapters[dev_id].device_sep_map,
+					g_memdup2(&addr, sizeof(addr)), seps = g_array_new(FALSE, FALSE, sizeof(sep)));
 
 		strncpy(sep.bluez_dbus_path, object_path, sizeof(sep.bluez_dbus_path) - 1);
 		if (sep.codec_id == A2DP_CODEC_VENDOR)
-			sep.codec_id = a2dp_get_vendor_codec_id(sep.capabilities, sep.capabilities_size);
-		sep.configuration = g_malloc(sep.capabilities_size);
+			sep.codec_id = a2dp_get_vendor_codec_id(&sep.capabilities, sep.capabilities_size);
 
 		debug("Adding new Stream End-Point: %s: %s", batostr_(&addr),
-				ba_transport_codecs_a2dp_to_string(sep.codec_id));
+				a2dp_codecs_codec_id_to_string(sep.codec_id));
 		g_array_append_val(seps, sep);
 
 	}
@@ -1007,6 +1154,7 @@ static void bluez_signal_interfaces_removed(GDBusConnection *conn, const char *s
 		const char *path, const char *interface_, const char *signal, GVariant *params,
 		void *userdata) {
 	debug("Signal: %s.%s()", interface_, signal);
+	(void)conn;
 	(void)sender;
 	(void)path;
 	(void)interface_;
@@ -1031,16 +1179,12 @@ static void bluez_signal_interfaces_removed(GDBusConnection *conn, const char *s
 			g_hash_table_iter_init(&iter, dbus_object_data_map);
 			while (g_hash_table_iter_next(&iter, NULL, (gpointer)&dbus_obj))
 				if (dbus_obj->hci_dev_id == hci_dev_id) {
-					g_dbus_connection_unregister_object(conn, dbus_obj->id);
+					g_dbus_interface_skeleton_unexport(dbus_obj->ifs);
+					g_object_unref(dbus_obj->ifs);
 					g_hash_table_iter_remove(&iter);
 				}
 
-			if (bluez_adapters[hci_dev_id].adapter != NULL) {
-				ba_adapter_destroy(bluez_adapters[hci_dev_id].adapter);
-				bluez_adapters[hci_dev_id].adapter = NULL;
-				g_hash_table_destroy(bluez_adapters[hci_dev_id].device_sep_map);
-				bluez_adapters[hci_dev_id].device_sep_map = NULL;
-			}
+			bluez_adapter_free(&bluez_adapters[hci_dev_id]);
 
 		}
 		else if (strcmp(interface, BLUEZ_IFACE_MEDIA_ENDPOINT) == 0) {
@@ -1065,7 +1209,6 @@ static void bluez_signal_interfaces_removed(GDBusConnection *conn, const char *s
 static void bluez_signal_transport_changed(GDBusConnection *conn, const char *sender,
 		const char *transport_path, const char *interface_, const char *signal, GVariant *params,
 		void *userdata) {
-	debug("Signal: %s.%s()", interface_, signal);
 	(void)conn;
 	(void)sender;
 	(void)interface_;
@@ -1101,7 +1244,7 @@ static void bluez_signal_transport_changed(GDBusConnection *conn, const char *se
 
 	g_variant_get(params, "(&sa{sv}as)", &interface, &properties, NULL);
 	while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
-		debug("Signal: %s: %s: %s", signal, interface, property);
+		debug("Signal: %s.%s(): %s: %s", interface_, signal, interface, property);
 
 		if (strcmp(property, "State") == 0 &&
 				g_variant_validate_value(value, G_VARIANT_TYPE_STRING, property)) {
@@ -1116,9 +1259,17 @@ static void bluez_signal_transport_changed(GDBusConnection *conn, const char *se
 		else if (strcmp(property, "Volume") == 0 &&
 				g_variant_validate_value(value, G_VARIANT_TYPE_UINT16, property)) {
 			/* received volume is in range [0, 127] */
-			const uint16_t volume = g_variant_get_uint16(value);
-			t->a2dp.pcm.volume[0].level = t->a2dp.pcm.volume[1].level = volume;
-			bluealsa_dbus_pcm_update(&t->a2dp.pcm, BA_DBUS_PCM_UPDATE_VOLUME);
+			uint16_t volume = t->a2dp.volume = g_variant_get_uint16(value);
+			if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE &&
+					t->a2dp.pcm.soft_volume)
+				debug("Skipping A2DP volume update: %u", volume);
+			else {
+				int level = ba_transport_pcm_volume_bt_to_level(&t->a2dp.pcm, volume);
+				debug("Updating A2DP volume: %u [%.2f dB]", volume, 0.01 * level);
+				ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[0], &level, NULL, NULL);
+				ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[1], &level, NULL, NULL);
+				bluealsa_dbus_pcm_update(&t->a2dp.pcm, BA_DBUS_PCM_UPDATE_VOLUME);
+			}
 		}
 
 		g_variant_unref(value);
@@ -1135,58 +1286,39 @@ final:
 }
 
 /**
- * Monitor BlueZ service availability.
+ * Monitor BlueZ service disappearance.
  *
  * When BlueZ is properly shutdown, we are notified about adapter removal via
  * the InterfacesRemoved signal. Here, we get the opportunity to perform some
  * cleanup if BlueZ service was killed. */
-static void bluez_signal_name_owner_changed(GDBusConnection *conn, const char *sender,
-		const char *path, const char *interface, const char *signal, GVariant *params,
+static void bluez_disappeared(GDBusConnection *conn, const char *name,
 		void *userdata) {
 	(void)conn;
-	(void)sender;
-	(void)path;
-	(void)interface;
-	(void)signal;
+	(void)name;
 	(void)userdata;
 
-	const char *name;
-	const char *owner_old;
-	const char *owner_new;
-
-	g_variant_get(params, "(&s&s&s)", &name, &owner_old, &owner_new);
-	if (owner_old != NULL && owner_old[0] != '\0') {
-
-		pthread_mutex_lock(&bluez_mutex);
-
-		GHashTableIter iter;
-		struct bluez_dbus_object_data *dbus_obj;
-		g_hash_table_iter_init(&iter, dbus_object_data_map);
-		while (g_hash_table_iter_next(&iter, NULL, (gpointer)&dbus_obj)) {
-			g_dbus_connection_unregister_object(conn, dbus_obj->id);
-			g_hash_table_iter_remove(&iter);
-		}
-
-		size_t i;
-		for (i = 0; i < ARRAYSIZE(bluez_adapters); i++)
-			if (bluez_adapters[i].adapter != NULL) {
-				ba_adapter_destroy(bluez_adapters[i].adapter);
-				bluez_adapters[i].adapter = NULL;
-				g_hash_table_destroy(bluez_adapters[i].device_sep_map);
-				bluez_adapters[i].device_sep_map = NULL;
-			}
-
-		pthread_mutex_unlock(&bluez_mutex);
+	pthread_mutex_lock(&bluez_mutex);
 
+	GHashTableIter iter;
+	struct bluez_dbus_object_data *dbus_obj;
+	g_hash_table_iter_init(&iter, dbus_object_data_map);
+	while (g_hash_table_iter_next(&iter, NULL, (gpointer)&dbus_obj)) {
+		g_dbus_interface_skeleton_unexport(dbus_obj->ifs);
+		g_object_unref(dbus_obj->ifs);
+		g_hash_table_iter_remove(&iter);
 	}
 
+	size_t i;
+	for (i = 0; i < ARRAYSIZE(bluez_adapters); i++)
+		bluez_adapter_free(&bluez_adapters[i]);
+
+	pthread_mutex_unlock(&bluez_mutex);
+
 }
 
 /**
- * Subscribe to BlueZ signals.
- *
- * @return On success this function returns 0. Otherwise -1 is returned. */
-int bluez_subscribe_signals(void) {
+ * Subscribe to BlueZ signals. */
+static void bluez_subscribe_signals(void) {
 
 	g_dbus_connection_signal_subscribe(config.dbus, BLUEZ_SERVICE,
 			DBUS_IFACE_OBJECT_MANAGER, "InterfacesAdded", NULL, NULL,
@@ -1199,9 +1331,24 @@ int bluez_subscribe_signals(void) {
 			DBUS_IFACE_PROPERTIES, "PropertiesChanged", NULL, BLUEZ_IFACE_MEDIA_TRANSPORT,
 			G_DBUS_SIGNAL_FLAGS_NONE, bluez_signal_transport_changed, NULL, NULL);
 
-	g_dbus_connection_signal_subscribe(config.dbus, DBUS_SERVICE,
-			DBUS_IFACE_DBUS, "NameOwnerChanged", NULL, BLUEZ_SERVICE,
-			G_DBUS_SIGNAL_FLAGS_NONE, bluez_signal_name_owner_changed, NULL, NULL);
+	g_bus_watch_name_on_connection(config.dbus, BLUEZ_SERVICE,
+			G_BUS_NAME_WATCHER_FLAGS_NONE, NULL, bluez_disappeared,
+			NULL, NULL);
+
+}
+
+/**
+ * Initialize integration with BlueZ service.
+ *
+ * @return On success this function returns 0. Otherwise -1 is returned. */
+int bluez_init(void) {
+
+	if (dbus_object_data_map == NULL)
+		dbus_object_data_map = g_hash_table_new_full(g_str_hash, g_str_equal,
+				NULL, (GDestroyNotify)bluez_dbus_object_data_unref);
+
+	bluez_subscribe_signals();
+	bluez_register();
 
 	return 0;
 }
@@ -1264,16 +1411,10 @@ bool bluez_a2dp_set_configuration(
 		goto fail;
 	}
 
-	GVariantBuilder caps;
-	g_variant_builder_init(&caps, G_VARIANT_TYPE("ay"));
-
-	size_t i;
-	for (i = 0; i < sep->capabilities_size; i++)
-		g_variant_builder_add(&caps, "y", ((char *)sep->configuration)[i]);
-
 	GVariantBuilder props;
 	g_variant_builder_init(&props, G_VARIANT_TYPE("a{sv}"));
-	g_variant_builder_add(&props, "{sv}", "Capabilities", g_variant_builder_end(&caps));
+	g_variant_builder_add(&props, "{sv}", "Capabilities", g_variant_new_fixed_array(
+				G_VARIANT_TYPE_BYTE, &sep->configuration, sep->capabilities_size, sizeof(uint8_t)));
 
 	msg = g_dbus_message_new_method_call(BLUEZ_SERVICE,
 			sep->bluez_dbus_path, BLUEZ_IFACE_MEDIA_ENDPOINT, "SetConfiguration");
@@ -1300,3 +1441,31 @@ fail:
 		g_object_unref(rep);
 	return rv;
 }
+
+/**
+ * Signal update on battery provider properties.
+ *
+ * This function makes sure that the battery provider object is
+ * exported or unexported when necessary. */
+void bluez_battery_provider_update(
+		struct ba_device *device) {
+
+	if (device->battery.charge == -1) {
+		bluez_battery_provider_manager_remove(device);
+		return;
+	}
+
+	if (!bluez_battery_provider_manager_add(device))
+		return;
+
+	GVariantBuilder props;
+	g_variant_builder_init(&props, G_VARIANT_TYPE("a{sv}"));
+
+	g_variant_builder_add(&props, "{sv}", "Percentage",
+			ba_variant_new_device_battery(device));
+
+	g_dbus_connection_emit_properties_changed(config.dbus,
+			device->ba_battery_dbus_path, BLUEZ_IFACE_BATTERY_PROVIDER, &props, NULL);
+	g_variant_builder_clear(&props);
+
+}
