@@ -1,6 +1,6 @@
 /*
  * BlueALSA - ofono.h
- * Copyright (c) 2016-2019 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *               2018 Thierry Bultel
  *
  * This file is a part of bluez-alsa.
@@ -9,13 +9,13 @@
  *
  */
 
+#pragma once
 #ifndef BLUEALSA_OFONO_H_
 #define BLUEALSA_OFONO_H_
 
 #include <stdbool.h>
 
-int ofono_register(void);
-int ofono_subscribe_signals(void);
+int ofono_init(void);
 bool ofono_detect_service(void);
 
 #endif
