@@ -1,6 +1,6 @@
 /*
  * BlueALSA - dbus-client.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -11,12 +11,18 @@
 #include "shared/dbus-client.h"
 
 #include <errno.h>
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
+#include <sys/param.h>
 #include <unistd.h>
 
+#include "shared/a2dp-codecs.h"
+#include "shared/defs.h"
+
 static int path2ba(const char *path, bdaddr_t *ba) {
 
 	unsigned int x[6];
@@ -157,6 +163,26 @@ dbus_bool_t bluealsa_dbus_connection_signal_match_clean(
 	return TRUE;
 }
 
+/**
+ * Dispatch D-Bus messages synchronously. */
+dbus_bool_t bluealsa_dbus_connection_dispatch(
+		struct ba_dbus_ctx *ctx) {
+
+	struct pollfd fds[8];
+	nfds_t nfds = ARRAYSIZE(fds);
+
+	bluealsa_dbus_connection_poll_fds(ctx, fds, &nfds);
+	if (poll(fds, nfds, 0) > 0)
+		bluealsa_dbus_connection_poll_dispatch(ctx, fds, nfds);
+
+	/* Dispatch incoming D-Bus messages/signals. The actual dispatching is
+	 * done in a function registered with dbus_connection_add_filter(). */
+	while (dbus_connection_dispatch(ctx->conn) == DBUS_DISPATCH_DATA_REMAINS)
+		continue;
+
+	return TRUE;
+}
+
 dbus_bool_t bluealsa_dbus_connection_poll_fds(
 		struct ba_dbus_ctx *ctx,
 		struct pollfd *fds,
@@ -214,6 +240,154 @@ dbus_bool_t bluealsa_dbus_connection_poll_dispatch(
 	return rv;
 }
 
+/**
+ * Callback function for BlueALSA service properties parser. */
+static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
+		DBusMessageIter *value, void *userdata, DBusError *error) {
+	struct ba_service_props *props = (struct ba_service_props *)userdata;
+
+	char type = dbus_message_iter_get_arg_type(value);
+	char type_expected;
+
+	if (strcmp(key, "Version") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_STRING))
+			goto fail;
+
+		const char *tmp;
+		dbus_message_iter_get_basic(value, &tmp);
+		strncpy(props->version, tmp, sizeof(props->version) - 1);
+
+	}
+	else if (strcmp(key, "Adapters") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+
+		const char *tmp[ARRAYSIZE(props->adapters)];
+		size_t length = ARRAYSIZE(tmp);
+		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+			return FALSE;
+
+		props->adapters_len = MIN(length, ARRAYSIZE(tmp));
+		for (size_t i = 0; i < length; i++)
+			strncpy(props->adapters[i], tmp[i], sizeof(props->adapters[i]) - 1);
+
+	}
+	else if (strcmp(key, "Profiles") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+
+		const char *tmp[32];
+		size_t length = ARRAYSIZE(tmp);
+		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+			return FALSE;
+
+		props->profiles = malloc(length * sizeof(*props->profiles));
+		props->profiles_len = MIN(length, ARRAYSIZE(tmp));
+		for (size_t i = 0; i < length; i++)
+			props->profiles[i] = strdup(tmp[i]);
+
+	}
+	else if (strcmp(key, "Codecs") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+
+		const char *tmp[64];
+		size_t length = ARRAYSIZE(tmp);
+		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+			return FALSE;
+
+		props->codecs = malloc(length * sizeof(*props->codecs));
+		props->codecs_len = MIN(length, ARRAYSIZE(tmp));
+		for (size_t i = 0; i < length; i++)
+			props->codecs[i] = strdup(tmp[i]);
+
+	}
+
+	return TRUE;
+
+fail:
+	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+			"Incorrect variant for '%s': %c != %c", key, type, type_expected);
+	return FALSE;
+}
+
+/**
+ * Get properties of BlueALSA service.
+ *
+ * This function allocates resources within the properties structure, which
+ * shall be freed with the bluealsa_dbus_props_free() function. */
+dbus_bool_t bluealsa_dbus_get_props(
+		struct ba_dbus_ctx *ctx,
+		struct ba_service_props *props,
+		DBusError *error) {
+
+	static const char *interface = BLUEALSA_INTERFACE_MANAGER;
+	DBusMessage *msg = NULL, *rep = NULL;
+	dbus_bool_t rv = FALSE;
+
+	props->profiles = NULL;
+	props->profiles_len = 0;
+	props->codecs = NULL;
+	props->codecs_len = 0;
+
+	if ((msg = dbus_message_new_method_call(ctx->ba_service, "/org/bluealsa",
+					DBUS_INTERFACE_PROPERTIES, "GetAll")) == NULL) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	DBusMessageIter iter;
+	dbus_message_iter_init_append(msg, &iter);
+	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &interface)) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
+					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL)
+		goto fail;
+
+	if (!dbus_message_iter_init(rep, &iter)) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE, "Empty response message");
+		goto fail;
+	}
+
+	if (!bluealsa_dbus_message_iter_dict(&iter, error,
+				bluealsa_dbus_message_iter_get_props_cb, props))
+		goto fail;
+
+	rv = TRUE;
+
+fail:
+	if (rep != NULL)
+		dbus_message_unref(rep);
+	if (msg != NULL)
+		dbus_message_unref(msg);
+	return rv;
+}
+
+/**
+ * Free BlueALSA service properties structure. */
+void bluealsa_dbus_props_free(
+		struct ba_service_props *props) {
+	if (props->profiles != NULL) {
+		for (size_t i = 0; i < props->profiles_len; i++)
+			free(props->profiles[i]);
+		free(props->profiles);
+		props->profiles = NULL;
+	}
+	if (props->codecs != NULL) {
+		for (size_t i = 0; i < props->codecs_len; i++)
+			free(props->codecs[i]);
+		free(props->codecs);
+		props->codecs = NULL;
+	}
+}
+
 dbus_bool_t bluealsa_dbus_get_pcms(
 		struct ba_dbus_ctx *ctx,
 		struct ba_pcm **pcms,
@@ -222,15 +396,14 @@ dbus_bool_t bluealsa_dbus_get_pcms(
 
 	DBusMessage *msg;
 	if ((msg = dbus_message_new_method_call(ctx->ba_service, "/org/bluealsa",
-					BLUEALSA_INTERFACE_MANAGER, "GetPCMs")) == NULL) {
+					DBUS_INTERFACE_OBJECT_MANAGER, "GetManagedObjects")) == NULL) {
 		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
 		return FALSE;
 	}
 
 	dbus_bool_t rv = TRUE;
 	struct ba_pcm *_pcms = NULL;
-	char *signature;
-	size_t i;
+	size_t _length = 0;
 
 	DBusMessage *rep;
 	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
@@ -243,44 +416,50 @@ dbus_bool_t bluealsa_dbus_get_pcms(
 		goto fail;
 	}
 
-	DBusMessageIter iter_pcms;
-	for (dbus_message_iter_recurse(&iter, &iter_pcms), i = 0;
-			dbus_message_iter_get_arg_type(&iter_pcms) != DBUS_TYPE_INVALID;
-			dbus_message_iter_next(&iter_pcms), i++) {
+	DBusMessageIter iter_objects;
+	for (dbus_message_iter_recurse(&iter, &iter_objects);
+			dbus_message_iter_get_arg_type(&iter_objects) != DBUS_TYPE_INVALID;
+			dbus_message_iter_next(&iter_objects)) {
 
-		if (dbus_message_iter_get_arg_type(&iter_pcms) != DBUS_TYPE_DICT_ENTRY)
-			goto fail_signature;
+		if (dbus_message_iter_get_arg_type(&iter_objects) != DBUS_TYPE_DICT_ENTRY) {
+			char *signature = dbus_message_iter_get_signature(&iter);
+			dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+					"Incorrect signature: %s != a{oa{sa{sv}}}", signature);
+			dbus_free(signature);
+			goto fail;
+		}
+
+		DBusMessageIter iter_object_entry;
+		dbus_message_iter_recurse(&iter_objects, &iter_object_entry);
+
+		struct ba_pcm pcm;
+		DBusError err = DBUS_ERROR_INIT;
+		if (!bluealsa_dbus_message_iter_get_pcm(&iter_object_entry, &err, &pcm)) {
+			dbus_set_error(error, err.name, "Get PCM: %s", err.message);
+			dbus_error_free(&err);
+			goto fail;
+		}
+
+		if (pcm.transport == BA_PCM_TRANSPORT_NONE)
+			continue;
 
 		struct ba_pcm *tmp = _pcms;
-		if ((tmp = realloc(tmp, (i + 1) * sizeof(*tmp))) == NULL) {
+		if ((tmp = realloc(tmp, (_length + 1) * sizeof(*tmp))) == NULL) {
 			dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
 			goto fail;
 		}
 
 		_pcms = tmp;
 
-		DBusMessageIter iter_pcms_entry;
-		dbus_message_iter_recurse(&iter_pcms, &iter_pcms_entry);
-
-		DBusError err = DBUS_ERROR_INIT;
-		if (!bluealsa_dbus_message_iter_get_pcm(&iter_pcms_entry, &err, &_pcms[i])) {
-			dbus_set_error(error, err.name, "Get PCM: %s", err.message);
-			dbus_error_free(&err);
-			goto fail;
-		}
+		memcpy(&_pcms[_length++], &pcm, sizeof(*_pcms));
 
 	}
 
 	*pcms = _pcms;
-	*length = i;
+	*length = _length;
 
 	goto success;
 
-fail_signature:
-	signature = dbus_message_iter_get_signature(&iter);
-	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
-			"Incorrect signature: %s != a{oa{sv}}", signature);
-
 fail:
 	if (_pcms != NULL)
 		free(_pcms);
@@ -296,36 +475,53 @@ success:
 dbus_bool_t bluealsa_dbus_get_pcm(
 		struct ba_dbus_ctx *ctx,
 		const bdaddr_t *addr,
-		unsigned int flags,
+		unsigned int transports,
+		unsigned int mode,
 		struct ba_pcm *pcm,
 		DBusError *error) {
 
+	const bool get_last = bacmp(addr, BDADDR_ANY) == 0;
 	struct ba_pcm *pcms = NULL;
+	struct ba_pcm *match = NULL;
 	dbus_bool_t rv = TRUE;
 	size_t length = 0;
+	uint32_t seq = 0;
 	size_t i;
 
 	if (!bluealsa_dbus_get_pcms(ctx, &pcms, &length, error))
 		return FALSE;
 
-	for (i = 0; i < length; i++)
-		if (bacmp(&pcms[i].addr, addr) == 0 &&
-				(pcms[i].flags & flags) == flags) {
-			memcpy(pcm, &pcms[i], sizeof(*pcm));
-			goto final;
+	for (i = 0; i < length; i++) {
+		if (get_last) {
+			if (pcms[i].sequence >= seq &&
+					pcms[i].transport & transports &&
+					pcms[i].mode == mode) {
+				seq = pcms[i].sequence;
+				match = &pcms[i];
+			}
 		}
+		else if (bacmp(&pcms[i].addr, addr) == 0 &&
+				pcms[i].transport & transports &&
+				pcms[i].mode == mode) {
+			match = &pcms[i];
+			break;
+		}
+	}
 
-	rv = FALSE;
-	dbus_set_error(error, DBUS_ERROR_FILE_NOT_FOUND, "PCM not found");
+	if (match != NULL)
+		memcpy(pcm, match, sizeof(*pcm));
+	else {
+		dbus_set_error(error, DBUS_ERROR_FILE_NOT_FOUND, "PCM not found");
+		rv = FALSE;
+	}
 
-final:
 	free(pcms);
 	return rv;
 }
 
 /**
  * Open BlueALSA PCM stream. */
-dbus_bool_t bluealsa_dbus_open_pcm(
+dbus_bool_t bluealsa_dbus_pcm_open(
 		struct ba_dbus_ctx *ctx,
 		const char *pcm_path,
 		int *fd_pcm,
@@ -357,6 +553,84 @@ dbus_bool_t bluealsa_dbus_open_pcm(
 	return rv;
 }
 
+const char *bluealsa_dbus_pcm_get_codec_canonical_name(
+		const char *alias) {
+
+	static const char *sco_codecs[] = { "CVSD", "mSBC" };
+	for (size_t i = 0; i < ARRAYSIZE(sco_codecs); i++)
+		if (strcasecmp(sco_codecs[i], alias) == 0)
+			return sco_codecs[i];
+
+	return a2dp_codecs_get_canonical_name(alias);
+}
+
+/**
+ * Select BlueALSA PCM Bluetooth codec. */
+dbus_bool_t bluealsa_dbus_pcm_select_codec(
+		struct ba_dbus_ctx *ctx,
+		const char *pcm_path,
+		const char *codec,
+		const void *configuration,
+		size_t configuration_len,
+		DBusError *error) {
+
+	DBusMessage *msg = NULL, *rep = NULL;
+	dbus_bool_t rv = FALSE;
+
+	if ((msg = dbus_message_new_method_call(ctx->ba_service, pcm_path,
+					BLUEALSA_INTERFACE_PCM, "SelectCodec")) == NULL) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	DBusMessageIter iter;
+	DBusMessageIter props;
+
+	dbus_message_iter_init_append(msg, &iter);
+	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &codec) ||
+			!dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY, "{sv}", &props)) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	if (configuration != NULL &&
+			configuration_len > 0) {
+		const char *property = "Configuration";
+		DBusMessageIter dict;
+		DBusMessageIter config;
+		DBusMessageIter array;
+		if (!dbus_message_iter_open_container(&props, DBUS_TYPE_DICT_ENTRY, NULL, &dict) ||
+				!dbus_message_iter_append_basic(&dict, DBUS_TYPE_STRING, &property) ||
+				!dbus_message_iter_open_container(&dict, DBUS_TYPE_VARIANT, "ay", &config) ||
+				!dbus_message_iter_open_container(&config, DBUS_TYPE_ARRAY, "y", &array) ||
+				!dbus_message_iter_append_fixed_array(&array, DBUS_TYPE_BYTE, &configuration, configuration_len) ||
+				!dbus_message_iter_close_container(&config, &array) ||
+				!dbus_message_iter_close_container(&dict, &config) ||
+				!dbus_message_iter_close_container(&props, &dict)) {
+			dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+			goto fail;
+		}
+	}
+
+	if (!dbus_message_iter_close_container(&iter, &props)) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
+					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL)
+		goto fail;
+
+	rv = TRUE;
+
+fail:
+	if (msg != NULL)
+		dbus_message_unref(msg);
+	if (rep != NULL)
+		dbus_message_unref(rep);
+	return rv;
+}
+
 /**
  * Open BlueALSA RFCOMM socket for dispatching AT commands. */
 dbus_bool_t bluealsa_dbus_open_rfcomm(
@@ -480,6 +754,42 @@ dbus_bool_t bluealsa_dbus_pcm_ctrl_send(
 	return TRUE;
 }
 
+/**
+ * Extract strings from the string array. */
+dbus_bool_t bluealsa_dbus_message_iter_array_get_strings(
+		DBusMessageIter *iter,
+		DBusError *error,
+		const char **strings,
+		size_t *length) {
+
+	char *signature;
+	size_t i;
+
+	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_ARRAY)
+		goto fail;
+
+	DBusMessageIter iter_array;
+	for (dbus_message_iter_recurse(iter, &iter_array), i = 0;
+			dbus_message_iter_get_arg_type(&iter_array) != DBUS_TYPE_INVALID;
+			dbus_message_iter_next(&iter_array)) {
+		if (dbus_message_iter_get_arg_type(&iter_array) != DBUS_TYPE_STRING)
+			goto fail;
+		if (i < *length)
+			dbus_message_iter_get_basic(&iter_array, &strings[i]);
+		i++;
+	}
+
+	*length = i;
+	return TRUE;
+
+fail:
+	signature = dbus_message_iter_get_signature(iter);
+	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+			"Incorrect signature: %s != as", signature);
+	dbus_free(signature);
+	return FALSE;
+}
+
 /**
  * Call the given function for each key/value pairs. */
 dbus_bool_t bluealsa_dbus_message_iter_dict(
@@ -538,22 +848,48 @@ dbus_bool_t bluealsa_dbus_message_iter_get_pcm(
 	const char *path;
 	char *signature;
 
-	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_OBJECT_PATH)
-		goto fail;
-
 	memset(pcm, 0, sizeof(*pcm));
 
+	if (dbus_message_iter_get_arg_type(iter) != DBUS_TYPE_OBJECT_PATH)
+		goto fail;
 	dbus_message_iter_get_basic(iter, &path);
-	strncpy(pcm->pcm_path, path, sizeof(pcm->pcm_path) - 1);
 
 	if (!dbus_message_iter_next(iter))
 		goto fail;
 
-	DBusError err = DBUS_ERROR_INIT;
-	if (!bluealsa_dbus_message_iter_get_pcm_props(iter, &err, pcm)) {
-		dbus_set_error(error, err.name, "Get properties: %s", err.message);
-		dbus_error_free(&err);
-		return FALSE;
+	DBusMessageIter iter_ifaces;
+	for (dbus_message_iter_recurse(iter, &iter_ifaces);
+			dbus_message_iter_get_arg_type(&iter_ifaces) != DBUS_TYPE_INVALID;
+			dbus_message_iter_next(&iter_ifaces)) {
+
+		if (dbus_message_iter_get_arg_type(&iter_ifaces) != DBUS_TYPE_DICT_ENTRY)
+			goto fail;
+
+		DBusMessageIter iter_iface_entry;
+		dbus_message_iter_recurse(&iter_ifaces, &iter_iface_entry);
+
+		const char *iface_name;
+		if (dbus_message_iter_get_arg_type(&iter_iface_entry) != DBUS_TYPE_STRING)
+			goto fail;
+		dbus_message_iter_get_basic(&iter_iface_entry, &iface_name);
+
+		if (strcmp(iface_name, BLUEALSA_INTERFACE_PCM) == 0) {
+
+			strncpy(pcm->pcm_path, path, sizeof(pcm->pcm_path) - 1);
+
+			if (!dbus_message_iter_next(&iter_iface_entry))
+				goto fail;
+
+			DBusError err = DBUS_ERROR_INIT;
+			if (!bluealsa_dbus_message_iter_get_pcm_props(&iter_iface_entry, &err, pcm)) {
+				dbus_set_error(error, err.name, "Get properties: %s", err.message);
+				dbus_error_free(&err);
+				return FALSE;
+			}
+
+			break;
+		}
+
 	}
 
 	return TRUE;
@@ -561,7 +897,7 @@ dbus_bool_t bluealsa_dbus_message_iter_get_pcm(
 fail:
 	signature = dbus_message_iter_get_signature(iter);
 	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
-			"Incorrect signature: %s != oa{sv}", signature);
+			"Incorrect signature: %s != oa{sa{sv}}", signature);
 	dbus_free(signature);
 	return FALSE;
 }
@@ -583,23 +919,36 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_pcm_props_cb(const char *key,
 		strncpy(pcm->device_path, tmp, sizeof(pcm->device_path) - 1);
 		path2ba(tmp, &pcm->addr);
 	}
+	else if (strcmp(key, "Sequence") == 0) {
+		if (type != (type_expected = DBUS_TYPE_UINT32))
+			goto fail;
+		dbus_message_iter_get_basic(variant, &pcm->sequence);
+	}
 	else if (strcmp(key, "Transport") == 0) {
 		if (type != (type_expected = DBUS_TYPE_STRING))
 			goto fail;
 		dbus_message_iter_get_basic(variant, &tmp);
-		if (strstr(tmp, "A2DP") != NULL)
-			pcm->flags |= BA_PCM_FLAG_PROFILE_A2DP;
-		else
-			pcm->flags |= BA_PCM_FLAG_PROFILE_SCO;
+		if (strstr(tmp, "A2DP-source") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_A2DP_SOURCE;
+		else if (strstr(tmp, "A2DP-sink") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_A2DP_SINK;
+		else if (strstr(tmp, "HFP-AG") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HFP_AG;
+		else if (strstr(tmp, "HFP-HF") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HFP_HF;
+		else if (strstr(tmp, "HSP-AG") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HSP_AG;
+		else if (strstr(tmp, "HSP-HS") != NULL)
+			pcm->transport = BA_PCM_TRANSPORT_HSP_HS;
 	}
 	else if (strcmp(key, "Mode") == 0) {
 		if (type != (type_expected = DBUS_TYPE_STRING))
 			goto fail;
 		dbus_message_iter_get_basic(variant, &tmp);
 		if (strcmp(tmp, "source") == 0)
-			pcm->flags |= BA_PCM_FLAG_SOURCE;
+			pcm->mode = BA_PCM_MODE_SOURCE;
 		else if (strcmp(tmp, "sink") == 0)
-			pcm->flags |= BA_PCM_FLAG_SINK;
+			pcm->mode = BA_PCM_MODE_SINK;
 	}
 	else if (strcmp(key, "Format") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
