@@ -35,11 +35,6 @@ bool g_variant_validate_value(GVariant *value, const GVariantType *type,
 unsigned int g_bdaddr_hash(const void *v);
 gboolean g_bdaddr_equal(const void *v1, const void *v2);
 
-uint16_t ba_transport_codecs_a2dp_from_string(const char *str);
-const char *ba_transport_codecs_a2dp_to_string(uint16_t codec);
-uint16_t ba_transport_codecs_hfp_from_string(const char *str);
-const char *ba_transport_codecs_hfp_to_string(uint16_t codec);
-
 const char *ba_transport_type_to_string(struct ba_transport_type type);
 
 #if ENABLE_MP3LAME
@@ -54,10 +49,9 @@ const char *aacdec_strerror(AAC_DECODER_ERROR err);
 const char *aacenc_strerror(AACENC_ERROR err);
 #endif
 
-#if ENABLE_APTX || ENABLE_APTX_HD
-# include <openaptx.h>
-void aptxbtenc_destroy_free(APTXENC enc);
-void aptxhdbtenc_destroy_free(APTXENC enc);
+#if ENABLE_LC3PLUS
+# include <lc3.h>
+const char *lc3plus_strerror(LC3PLUS_Error err);
 #endif
 
 #if ENABLE_LDAC
