@@ -1,6 +1,6 @@
 /*
  * test-alsa-pcm.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2021 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -30,13 +30,13 @@
 #include <check.h>
 #include <alsa/asoundlib.h>
 
+#include "shared/defs.h"
+#include "shared/log.h"
+#include "shared/rt.h"
+
 #include "inc/preload.inc"
 #include "inc/server.inc"
 #include "inc/sine.inc"
-#include "../src/shared/defs.h"
-#include "../src/shared/ffb.c"
-#include "../src/shared/log.c"
-#include "../src/shared/rt.c"
 
 #define dumprv(fn) fprintf(stderr, #fn " = %d\n", (int)fn)
 
@@ -47,7 +47,12 @@ static snd_pcm_format_t pcm_format = SND_PCM_FORMAT_S16_LE;
 /* big enough buffer to keep one period of data */
 static int16_t buffer[1024 * 8];
 
-static int snd_pcm_open_bluealsa(snd_pcm_t **pcmp, const char *service, snd_pcm_stream_t stream, int mode) {
+static int snd_pcm_open_bluealsa(
+		snd_pcm_t **pcmp,
+		const char *service,
+		const char *extra_config,
+		snd_pcm_stream_t stream,
+		int mode) {
 
 	char buffer[256];
 	snd_config_t *conf = NULL;
@@ -61,7 +66,8 @@ static int snd_pcm_open_bluealsa(snd_pcm_t **pcmp, const char *service, snd_pcm_
 			"  device \"12:34:56:78:9A:BC\"\n"
 			"  profile \"a2dp\"\n"
 			"  delay 0\n"
-			"}\n", service);
+			"  %s\n"
+			"}\n", service, extra_config);
 
 	if ((err = snd_config_top(&conf)) < 0)
 		goto fail;
@@ -161,13 +167,16 @@ static int test_pcm_open(pid_t *pid, snd_pcm_t **pcm, snd_pcm_stream_t stream) {
 	const char *service = "test";
 	if ((*pid = spawn_bluealsa_server(service, 1, true, false,
 					stream == SND_PCM_STREAM_PLAYBACK,
-					stream == SND_PCM_STREAM_CAPTURE)) == -1)
+					stream == SND_PCM_STREAM_CAPTURE,
+					false, false)) == -1)
 		return -1;
-	return snd_pcm_open_bluealsa(pcm, service, stream, 0);
+	return snd_pcm_open_bluealsa(pcm, service, "", stream, 0);
 }
 
 static int test_pcm_close(pid_t pid, snd_pcm_t *pcm) {
-	int rv = snd_pcm_close(pcm);
+	int rv = 0;
+	if (pcm != NULL)
+		rv = snd_pcm_close(pcm);
 	if (pid != -1) {
 		kill(pid, SIGTERM);
 		waitpid(pid, NULL, 0);
@@ -178,7 +187,7 @@ static int test_pcm_close(pid_t pid, snd_pcm_t *pcm) {
 static int16_t *test_sine_s16le(snd_pcm_uframes_t size) {
 	static size_t x = 0;
 	assert(ARRAYSIZE(buffer) >= size * pcm_channels);
-	x = snd_pcm_sine_s16le(buffer, size * pcm_channels, pcm_channels, x, 441.0 / pcm_sampling);
+	x = snd_pcm_sine_s16_2le(buffer, size, pcm_channels, x, 441.0 / pcm_sampling);
 	return buffer;
 }
 
@@ -249,10 +258,12 @@ START_TEST(test_capture_start) {
 	for (i = 0; i < buffer_size / period_size; i++)
 		ck_assert_int_eq(snd_pcm_readi(pcm, buffer, period_size), period_size);
 
-	/* after reading there should be no more than one period of data */
-	ck_assert_int_le(snd_pcm_avail(pcm), period_size);
+	/* after reading there should be no more than one period of data in buffer */
+	snd_pcm_sframes_t avail;
+	ck_assert_int_le((avail = snd_pcm_avail(pcm)), period_size);
+	/* but there may be more data in the FIFO */
 	ck_assert_int_eq(snd_pcm_delay(pcm, &delay), 0);
-	ck_assert_int_le(delay, period_size);
+	ck_assert_int_ge(delay, avail);
 
 	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
 
@@ -367,6 +378,45 @@ START_TEST(test_capture_overrun) {
 
 } END_TEST
 
+START_TEST(test_capture_poll) {
+	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+
+	unsigned int buffer_time = 200000;
+	unsigned int period_time = 25000;
+	snd_pcm_t *pcm = NULL;
+	pid_t pid = -1;
+
+	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_CAPTURE), 0);
+	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
+				&buffer_time, &period_time), 0);
+
+	struct pollfd pfds[8];
+	unsigned short revents;
+	int count = snd_pcm_poll_descriptors_count(pcm);
+	ck_assert_int_eq(snd_pcm_poll_descriptors(pcm, pfds, ARRAYSIZE(pfds)), count);
+
+	ck_assert_int_eq(snd_pcm_prepare(pcm), 0);
+	/* for a capture PCM just after prepare, the poll() call shall block
+	 * forever or at least the dispatched event shall be set to 0 */
+	ck_assert_int_ne(poll(pfds, count, 250), -1);
+	snd_pcm_poll_descriptors_revents(pcm, pfds, count, &revents);
+	ck_assert_int_eq(revents, 0);
+
+	/* make sure that further calls to poll() will actually block */
+	ck_assert_int_eq(poll(pfds, count, 250), 0);
+
+	ck_assert_int_eq(snd_pcm_start(pcm), 0);
+	do { /* started capture PCM shall not block forever */
+		ck_assert_int_gt(poll(pfds, count, -1), 0);
+		snd_pcm_poll_descriptors_revents(pcm, pfds, count, &revents);
+	} while (revents == 0);
+	/* we should get read event flag set */
+	ck_assert_int_eq(revents & POLLIN, POLLIN);
+
+	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+
+} END_TEST
+
 START_TEST(dump_playback) {
 	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
 
@@ -399,7 +449,11 @@ START_TEST(dump_playback) {
 
 } END_TEST
 
-START_TEST(test_playback_hw_constraints) {
+START_TEST(ba_test_playback_hw_constraints) {
+
+	if (pcm_device != NULL)
+		return;
+
 	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
 
 	/* hard-coded values used in the bluealsa-mock */
@@ -463,13 +517,69 @@ START_TEST(test_playback_hw_constraints) {
 	ck_assert_int_eq(d, 0);
 	snd_pcm_hw_params_any(pcm, params);
 	ck_assert_int_eq(snd_pcm_hw_params_set_buffer_time_last(pcm, params, &time, &d), 0);
-	ck_assert_int_eq(time, 95108934);
+	ck_assert_int_eq(time, 11888616);
 	ck_assert_int_eq(d, 1);
 
 	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
 
 } END_TEST
 
+START_TEST(ba_test_playback_extra_setup) {
+
+	if (pcm_device != NULL)
+		return;
+
+	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+
+	snd_pcm_t *pcm = NULL;
+	pid_t pid = -1;
+
+	const char *service = "test";
+	ck_assert_int_ne(pid = spawn_bluealsa_server(service, 1, true, false, true, false, false, false), -1);
+
+	ck_assert_int_eq(snd_pcm_open_bluealsa(&pcm, service, "codec \"sbc\"", SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(-1, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open_bluealsa(&pcm, service, "volume \"50+\"", SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(-1, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open_bluealsa(&pcm, service, "softvol true", SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(-1, pcm), 0);
+
+	ck_assert_int_eq(test_pcm_close(pid, NULL), 0);
+
+} END_TEST
+
+START_TEST(test_playback_hw_set_free) {
+	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+
+	unsigned int buffer_time = 200000;
+	unsigned int period_time = 25000;
+	snd_pcm_t *pcm = NULL;
+	pid_t pid = -1;
+	size_t i;
+
+	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+
+	for (i = 0; i < 5; i++) {
+		int set_hw_param_ret;
+		/* acquire Bluetooth transport */
+		if ((set_hw_param_ret = set_hw_params(pcm, pcm_format, pcm_channels,
+					pcm_sampling, &buffer_time, &period_time)) == -EBUSY) {
+			debug("Retrying snd_pcm_hw_params_set...");
+			/* do not treat busy as an error */
+			i--;
+			continue;
+		}
+		ck_assert_int_eq(set_hw_param_ret, 0);
+		/* release Bluetooth transport */
+		ck_assert_int_eq(snd_pcm_hw_free(pcm), 0);
+	}
+
+	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+
+} END_TEST
+
 START_TEST(test_playback_start) {
 	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
 
@@ -696,7 +806,7 @@ START_TEST(test_playback_underrun) {
 	/* after one and a half period time we shall be
 	 * able to write at least one period frames */
 	usleep(3 * period_time / 2);
-	ck_assert_int_gt(snd_pcm_avail(pcm), period_size);
+	ck_assert_int_ge(snd_pcm_avail(pcm), period_size);
 
 	/* allow under-run to occur */
 	usleep(buffer_time);
@@ -746,7 +856,7 @@ START_TEST(reference_playback_device_unplug) {
 	unsigned int period_time = 25000;
 	snd_pcm_sframes_t frames = 0;
 	snd_pcm_t *pcm = NULL;
-	struct pollfd pfds[2];
+	struct pollfd pfds[4];
 	unsigned short revents;
 
 	/* this test needs user-defined PCM device */
@@ -764,8 +874,8 @@ START_TEST(reference_playback_device_unplug) {
 
 	dumprv(frames);
 	dumprv(snd_pcm_poll_descriptors_count(pcm));
-	dumprv(snd_pcm_poll_descriptors(pcm, pfds, 4));
-	dumprv(snd_pcm_poll_descriptors_revents(pcm, pfds, 4, &revents));
+	dumprv(snd_pcm_poll_descriptors(pcm, pfds, ARRAYSIZE(pfds)));
+	dumprv(snd_pcm_poll_descriptors_revents(pcm, pfds, ARRAYSIZE(pfds), &revents));
 	dumprv(snd_pcm_prepare(pcm));
 	dumprv(snd_pcm_reset(pcm));
 	dumprv(snd_pcm_start(pcm));
@@ -782,7 +892,11 @@ START_TEST(reference_playback_device_unplug) {
 
 } END_TEST
 
-START_TEST(test_playback_device_unplug) {
+START_TEST(ba_test_playback_device_unplug) {
+
+	if (pcm_device != NULL)
+		return;
+
 	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
 
 	unsigned int buffer_time = 200000;
@@ -809,8 +923,8 @@ START_TEST(test_playback_device_unplug) {
 
 	ck_assert_int_eq(frames, -ENODEV);
 	ck_assert_int_eq(snd_pcm_poll_descriptors_count(pcm), 2);
-	ck_assert_int_eq(snd_pcm_poll_descriptors(pcm, pfds, 4), 2);
-	ck_assert_int_eq(snd_pcm_poll_descriptors_revents(pcm, pfds, 4, &revents), -ENODEV);
+	ck_assert_int_eq(snd_pcm_poll_descriptors(pcm, pfds, ARRAYSIZE(pfds)), 2);
+	ck_assert_int_eq(snd_pcm_poll_descriptors_revents(pcm, pfds, ARRAYSIZE(pfds), &revents), -ENODEV);
 	ck_assert_int_eq(snd_pcm_prepare(pcm), -ENODEV);
 	ck_assert_int_eq(snd_pcm_reset(pcm), 0);
 	ck_assert_int_eq(snd_pcm_start(pcm), -EBADFD);
@@ -868,53 +982,56 @@ int main(int argc, char *argv[]) {
 			return 1;
 		}
 
-	/* test-alsa-pcm and bluealsa-mock shall be placed in the same directory */
-	bluealsa_mock_path = dirname(argv[0]);
+	/* test-alsa-pcm and bluealsa-mock shall
+	 * be placed in the same directory */
+	char *argv_0 = strdup(argv[0]);
+	bluealsa_mock_path = dirname(argv_0);
 
 	Suite *s = suite_create(__FILE__);
-	TCase *tc = tcase_create(__FILE__);
 	SRunner *sr = srunner_create(s);
 
-	suite_add_tcase(s, tc);
-	tcase_set_timeout(tc, 6);
+	TCase *tc_capture = tcase_create("capture");
+	tcase_add_test(tc_capture, dump_capture);
+	tcase_add_test(tc_capture, test_capture_start);
+	tcase_add_test(tc_capture, test_capture_pause);
+	tcase_add_test(tc_capture, test_capture_overrun);
+	tcase_add_test(tc_capture, test_capture_poll);
+
+	TCase *tc_playback = tcase_create("playback");
+	tcase_add_test(tc_playback, dump_playback);
+	tcase_add_test(tc_playback, ba_test_playback_hw_constraints);
+	tcase_add_test(tc_playback, ba_test_playback_extra_setup);
+	tcase_add_test(tc_playback, test_playback_hw_set_free);
+	tcase_add_test(tc_playback, test_playback_start);
+	tcase_add_test(tc_playback, test_playback_drain);
+	tcase_add_test(tc_playback, test_playback_pause);
+	tcase_add_test(tc_playback, test_playback_reset);
+	tcase_add_test(tc_playback, test_playback_underrun);
+	tcase_add_test(tc_playback, ba_test_playback_device_unplug);
+
+	TCase *tc_unplug = tcase_create("unplug");
+	tcase_add_test(tc_unplug, reference_playback_device_unplug);
 
 	if (argc == optind) {
-		tcase_add_test(tc, dump_capture);
-		tcase_add_test(tc, test_capture_start);
-		tcase_add_test(tc, test_capture_overrun);
-		tcase_add_test(tc, dump_playback);
-		tcase_add_test(tc, test_playback_hw_constraints);
-		tcase_add_test(tc, test_playback_start);
-		tcase_add_test(tc, test_playback_drain);
-		tcase_add_test(tc, test_playback_pause);
-		tcase_add_test(tc, test_playback_reset);
-		tcase_add_test(tc, test_playback_underrun);
-		tcase_add_test(tc, test_playback_device_unplug);
+		suite_add_tcase(s, tc_capture);
+		suite_add_tcase(s, tc_playback);
 	}
 	else {
 		for (; optind < argc; optind++) {
-			if (strcmp(argv[optind], "capture") == 0) {
-				tcase_add_test(tc, dump_capture);
-				tcase_add_test(tc, test_capture_start);
-				tcase_add_test(tc, test_capture_pause);
-				tcase_add_test(tc, test_capture_overrun);
-			}
-			else if (strcmp(argv[optind], "playback") == 0) {
-				tcase_add_test(tc, dump_playback);
-				tcase_add_test(tc, test_playback_start);
-				tcase_add_test(tc, test_playback_drain);
-				tcase_add_test(tc, test_playback_pause);
-				tcase_add_test(tc, test_playback_reset);
-				tcase_add_test(tc, test_playback_underrun);
-			}
+			if (strcmp(argv[optind], "capture") == 0)
+				suite_add_tcase(s, tc_capture);
+			else if (strcmp(argv[optind], "playback") == 0)
+				suite_add_tcase(s, tc_playback);
 			else if (strcmp(argv[optind], "unplug") == 0)
-				tcase_add_test(tc, reference_playback_device_unplug);
+				suite_add_tcase(s, tc_unplug);
 		}
 	}
 
 	srunner_run_all(sr, CK_ENV);
 	int nf = srunner_ntests_failed(sr);
+
 	srunner_free(sr);
+	free(argv_0);
 
 	return nf == 0 ? 0 : 1;
 }
