@@ -1,6 +1,6 @@
 /*
  * BlueALSA - aplay.c
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
+ * Copyright (c) 2016-2021 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -12,13 +12,14 @@
 # include <config.h>
 #endif
 
+#include <assert.h>
 #include <errno.h>
 #include <getopt.h>
+#include <math.h>
 #include <poll.h>
 #include <pthread.h>
 #include <signal.h>
 #include <stdbool.h>
-#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -32,6 +33,7 @@
 #include "shared/defs.h"
 #include "shared/ffb.h"
 #include "shared/log.h"
+#include "alsa-mixer.h"
 #include "alsa-pcm.h"
 #include "dbus.h"
 
@@ -45,6 +47,9 @@ struct pcm_worker {
 	int ba_pcm_ctrl_fd;
 	/* opened playback PCM device */
 	snd_pcm_t *pcm;
+	/* mixer for volume control */
+	snd_mixer_t *mixer;
+	snd_mixer_elem_t *mixer_elem;
 	/* if true, playback is active */
 	bool active;
 	/* human-readable BT address */
@@ -55,6 +60,9 @@ static unsigned int verbose = 0;
 static bool list_bt_devices = false;
 static bool list_bt_pcms = false;
 static const char *pcm_device = "default";
+static const char *mixer_device = "default";
+static const char *mixer_elem_name = "Master";
+static unsigned int mixer_elem_index = 0;
 static bool ba_profile_a2dp = true;
 static bool ba_addr_any = false;
 static bdaddr_t *ba_addrs = NULL;
@@ -102,6 +110,22 @@ static int parse_bt_addresses(char *argv[], size_t count) {
 	return 0;
 }
 
+static const char *bluealsa_get_profile(const struct ba_pcm *pcm) {
+	switch (pcm->transport) {
+	case BA_PCM_TRANSPORT_A2DP_SOURCE:
+	case BA_PCM_TRANSPORT_A2DP_SINK:
+		return "A2DP";
+	case BA_PCM_TRANSPORT_HFP_AG:
+	case BA_PCM_TRANSPORT_HFP_HF:
+	case BA_PCM_TRANSPORT_HSP_AG:
+	case BA_PCM_TRANSPORT_HSP_HS:
+		return "SCO";
+	default:
+		error("Unknown transport: %#x", pcm->transport);
+		return "[...]";
+	}
+}
+
 static snd_pcm_format_t bluealsa_get_snd_pcm_format(const struct ba_pcm *pcm) {
 	switch (pcm->format) {
 	case 0x0108:
@@ -124,10 +148,10 @@ static void print_bt_device_list(void) {
 
 	static const struct {
 		const char *label;
-		unsigned int flag;
+		unsigned int mode;
 	} section[2] = {
-		{ "**** List of PLAYBACK Bluetooth Devices ****", BA_PCM_FLAG_SINK },
-		{ "**** List of CAPTURE Bluetooth Devices ****", BA_PCM_FLAG_SOURCE },
+		{ "**** List of PLAYBACK Bluetooth Devices ****", BA_PCM_MODE_SINK },
+		{ "**** List of CAPTURE Bluetooth Devices ****", BA_PCM_MODE_SOURCE },
 	};
 
 	const char *tmp;
@@ -140,7 +164,7 @@ static void print_bt_device_list(void) {
 			struct ba_pcm *pcm = &ba_pcms[ii];
 			struct bluez_device dev = { 0 };
 
-			if (!(pcm->flags & section[i].flag))
+			if (!(pcm->mode == section[i].mode))
 				continue;
 
 			if (strcmp(pcm->device_path, tmp) != 0) {
@@ -152,7 +176,7 @@ static void print_bt_device_list(void) {
 					dbus_error_free(&err);
 				}
 
-				char bt_addr[16];
+				char bt_addr[18];
 				ba2str(&dev.bt_addr, bt_addr);
 
 				printf("%s: %s [%s], %s%s\n",
@@ -162,7 +186,7 @@ static void print_bt_device_list(void) {
 			}
 
 			printf("  %s (%s): %s %d channel%s %d Hz\n",
-				pcm->flags & BA_PCM_FLAG_PROFILE_A2DP ? "A2DP" : "SCO",
+				bluealsa_get_profile(pcm),
 				pcm->codec,
 				snd_pcm_format_name(bluealsa_get_snd_pcm_format(pcm)),
 				pcm->channels, pcm->channels != 1 ? "s" : "",
@@ -191,7 +215,7 @@ static void print_bt_pcm_list(void) {
 			}
 		}
 
-		char bt_addr[16];
+		char bt_addr[18];
 		ba2str(&dev.bt_addr, bt_addr);
 
 		printf(
@@ -200,11 +224,11 @@ static void print_bt_pcm_list(void) {
 				"    %s (%s): %s %d channel%s %d Hz\n",
 			dbus_ba_service,
 			bt_addr,
-			pcm->flags & BA_PCM_FLAG_PROFILE_A2DP ? "a2dp" : "sco",
+			pcm->transport & BA_PCM_TRANSPORT_MASK_A2DP ? "a2dp" : "sco",
 			dev.name,
 			dev.trusted ? "trusted ": "", dev.icon,
-			pcm->flags & BA_PCM_FLAG_SINK ? "playback" : "capture",
-			pcm->flags & BA_PCM_FLAG_PROFILE_A2DP ? "A2DP" : "SCO",
+			pcm->mode == BA_PCM_MODE_SINK ? "playback" : "capture",
+			bluealsa_get_profile(pcm),
 			pcm->codec,
 			snd_pcm_format_name(bluealsa_get_snd_pcm_format(pcm)),
 			pcm->channels, pcm->channels != 1 ? "s" : "",
@@ -274,6 +298,109 @@ final:
 	return ret;
 }
 
+/**
+ * Synchronize BlueALSA PCM volume with ALSA mixer element. */
+static int pcm_worker_mixer_volume_sync(
+		struct pcm_worker *worker,
+		struct ba_pcm *ba_pcm) {
+
+	/* skip sync in case of software volume */
+	if (ba_pcm->soft_volume)
+		return 0;
+
+	snd_mixer_elem_t *elem = worker->mixer_elem;
+	if (elem == NULL)
+		return 0;
+
+	const int vmax = BA_PCM_VOLUME_MAX(ba_pcm);
+	long long volume_db_sum = 0;
+	bool muted = true;
+
+	snd_mixer_selem_channel_id_t ch;
+	for (ch = 0; snd_mixer_selem_has_playback_channel(elem, ch) == 1; ch++) {
+
+		long ch_volume_db;
+		int ch_switch;
+
+		int err;
+		if ((err = snd_mixer_selem_get_playback_dB(elem, 0, &ch_volume_db)) != 0 ||
+				(err = snd_mixer_selem_get_playback_switch(elem, 0, &ch_switch)) != 0) {
+			error("Couldn't get playback volume: %s", snd_strerror(err));
+			return -1;
+		}
+
+		volume_db_sum += ch_volume_db;
+		if (ch_switch == 1)
+			muted = false;
+
+	}
+
+	/* Safety check for undefined behavior from
+	 * out-of-bounds dB conversion. */
+	assert(volume_db_sum <= 0LL);
+
+	/* Convert dB to loudness using decibel formula and
+	 * round to the nearest integer. */
+	int volume = lround(pow(2, (0.01 * volume_db_sum / ch) / 10) * vmax);
+
+	ba_pcm->volume.ch1_muted = muted;
+	ba_pcm->volume.ch1_volume = volume;
+	ba_pcm->volume.ch2_muted = muted;
+	ba_pcm->volume.ch2_volume = volume;
+
+	DBusError err = DBUS_ERROR_INIT;
+	if (!bluealsa_dbus_pcm_update(&dbus_ctx, ba_pcm, BLUEALSA_PCM_VOLUME, &err)) {
+		error("Couldn't update PCM: %s", err.message);
+		dbus_error_free(&err);
+		return -1;
+	}
+
+	return 0;
+}
+
+/**
+ * Update ALSA mixer element according to BlueALSA PCM volume. */
+static int pcm_worker_mixer_volume_update(
+		struct pcm_worker *worker,
+		struct ba_pcm *ba_pcm) {
+
+	/* skip update in case of software volume */
+	if (ba_pcm->soft_volume)
+		return 0;
+
+	snd_mixer_elem_t *elem = worker->mixer_elem;
+	if (elem == NULL)
+		return 0;
+
+	/* User can connect BlueALSA PCM to mono, stereo or multi-channel output.
+	 * For mono input (audio from BlueALSA PCM), case case is simple: we are
+	 * changing all output channels at once. However, for stereo input it is
+	 * not possible to know how to control left/right volume unless there is
+	 * some kind of channel mapping. In order to simplify things, we will set
+	 * all channels to the average left-right volume. */
+
+	const int vmax = BA_PCM_VOLUME_MAX(ba_pcm);
+	int volume = ba_pcm->volume.ch1_volume;
+	int muted = ba_pcm->volume.ch1_muted;
+
+	if (ba_pcm->channels > 1) {
+		volume = (ba_pcm->volume.ch1_volume + ba_pcm->volume.ch2_volume) / 2;
+		muted = ba_pcm->volume.ch1_muted || ba_pcm->volume.ch2_muted;
+	}
+
+	/* convert loudness to dB using decibel formula */
+	long db = 10 * log2(1.0 * volume / vmax) * 100;
+
+	int err;
+	if ((err = snd_mixer_selem_set_playback_dB_all(elem, db, 0)) != 0 ||
+			(err = snd_mixer_selem_set_playback_switch_all(elem, !muted)) != 0) {
+		error("Couldn't set playback volume: %s", snd_strerror(err));
+		return -1;
+	}
+
+	return 0;
+}
+
 static void pcm_worker_routine_exit(struct pcm_worker *worker) {
 	if (worker->ba_pcm_fd != -1) {
 		close(worker->ba_pcm_fd);
@@ -287,6 +414,11 @@ static void pcm_worker_routine_exit(struct pcm_worker *worker) {
 		snd_pcm_close(worker->pcm);
 		worker->pcm = NULL;
 	}
+	if (worker->mixer != NULL) {
+		snd_mixer_close(worker->mixer);
+		worker->mixer_elem = NULL;
+		worker->mixer = NULL;
+	}
 	debug("Exiting PCM worker %s", worker->addr);
 }
 
@@ -311,7 +443,7 @@ static void *pcm_worker_routine(struct pcm_worker *w) {
 	}
 
 	DBusError err = DBUS_ERROR_INIT;
-	if (!bluealsa_dbus_open_pcm(&dbus_ctx, w->ba_pcm.pcm_path,
+	if (!bluealsa_dbus_pcm_open(&dbus_ctx, w->ba_pcm.pcm_path,
 				&w->ba_pcm_fd, &w->ba_pcm_ctrl_fd, &err)) {
 		error("Couldn't open PCM: %s", err.message);
 		dbus_error_free(&err);
@@ -420,6 +552,15 @@ static void *pcm_worker_routine(struct pcm_worker *w) {
 				continue;
 			}
 
+			if (alsa_mixer_open(&w->mixer, &w->mixer_elem,
+						mixer_device, mixer_elem_name, mixer_elem_index, &tmp) != 0) {
+				warn("Couldn't open mixer: %s", tmp);
+				free(tmp);
+			}
+
+			/* initial volume synchronization */
+			pcm_worker_mixer_volume_sync(w, &w->ba_pcm);
+
 			snd_pcm_get_params(w->pcm, &buffer_size, &period_size);
 			pcm_max_read_len = period_size * w->ba_pcm.channels * pcm_format_size;
 			pcm_open_retries = 0;
@@ -475,12 +616,52 @@ fail:
 	return NULL;
 }
 
-static int supervise_pcm_worker_start(struct ba_pcm *ba_pcm) {
+static bool pcm_hw_params_equal(
+		const struct ba_pcm *ba_pcm_1,
+		const struct ba_pcm *ba_pcm_2) {
+	if (ba_pcm_1->format != ba_pcm_2->format)
+		return false;
+	if (ba_pcm_1->channels != ba_pcm_2->channels)
+		return false;
+	if (ba_pcm_1->sampling != ba_pcm_2->sampling)
+		return false;
+	return true;
+}
+
+/**
+ * Stop the worker thread at workers[index]. */
+static void pcm_worker_stop(size_t index) {
+
+	/* Safety check for out-of-bounds read. */
+	assert(index < workers_count);
+
+	pthread_rwlock_wrlock(&workers_lock);
+
+	pthread_cancel(workers[index].thread);
+	pthread_join(workers[index].thread, NULL);
+
+	if (index != --workers_count)
+		/* Move the last worker in the array to position
+		 * index, to prevent any "gaps" in the array. */
+		memcpy(&workers[index], &workers[workers_count], sizeof(workers[index]));
+
+	pthread_rwlock_unlock(&workers_lock);
+
+}
+
+static struct pcm_worker *supervise_pcm_worker_start(const struct ba_pcm *ba_pcm) {
 
 	size_t i;
 	for (i = 0; i < workers_count; i++)
-		if (strcmp(workers[i].ba_pcm.pcm_path, ba_pcm->pcm_path) == 0)
-			return 0;
+		if (strcmp(workers[i].ba_pcm.pcm_path, ba_pcm->pcm_path) == 0) {
+			/* If the codec has changed after the device connected, then the
+			 * audio format may have changed. If it has, the worker thread
+			 * needs to be restarted. */
+			if (!pcm_hw_params_equal(&workers[i].ba_pcm, ba_pcm))
+				pcm_worker_stop(i);
+			else
+				return &workers[i];
+		}
 
 	pthread_rwlock_wrlock(&workers_lock);
 
@@ -492,7 +673,7 @@ static int supervise_pcm_worker_start(struct ba_pcm *ba_pcm) {
 			error("Couldn't (re)allocate memory for PCM workers: %s", strerror(ENOMEM));
 			workers = tmp;
 			pthread_rwlock_unlock(&workers_lock);
-			return -1;
+			return NULL;
 		}
 	}
 
@@ -512,41 +693,36 @@ static int supervise_pcm_worker_start(struct ba_pcm *ba_pcm) {
 					PTHREAD_ROUTINE(pcm_worker_routine), worker)) != 0) {
 		error("Couldn't create PCM worker %s: %s", worker->addr, strerror(errno));
 		workers_count--;
-		return -1;
+		return NULL;
 	}
 
-	return 0;
+	return worker;
 }
 
-static int supervise_pcm_worker_stop(struct ba_pcm *ba_pcm) {
+static struct pcm_worker *supervise_pcm_worker_stop(const struct ba_pcm *ba_pcm) {
 
 	size_t i;
 	for (i = 0; i < workers_count; i++)
-		if (strcmp(workers[i].ba_pcm.pcm_path, ba_pcm->pcm_path) == 0) {
-			pthread_rwlock_wrlock(&workers_lock);
-			pthread_cancel(workers[i].thread);
-			pthread_join(workers[i].thread, NULL);
-			memcpy(&workers[i], &workers[--workers_count], sizeof(workers[i]));
-			pthread_rwlock_unlock(&workers_lock);
-		}
+		if (strcmp(workers[i].ba_pcm.pcm_path, ba_pcm->pcm_path) == 0)
+			pcm_worker_stop(i);
 
-	return 0;
+	return NULL;
 }
 
-static int supervise_pcm_worker(struct ba_pcm *ba_pcm) {
+static struct pcm_worker *supervise_pcm_worker(const struct ba_pcm *ba_pcm) {
 
 	if (ba_pcm == NULL)
-		return -1;
+		return NULL;
 
-	if (!(ba_pcm->flags & BA_PCM_FLAG_SOURCE))
+	if (ba_pcm->mode != BA_PCM_MODE_SOURCE)
 		goto stop;
 
-	if ((ba_profile_a2dp && !(ba_pcm->flags & BA_PCM_FLAG_PROFILE_A2DP)) ||
-			(!ba_profile_a2dp && !(ba_pcm->flags & BA_PCM_FLAG_PROFILE_SCO)))
+	if ((ba_profile_a2dp && !(ba_pcm->transport & BA_PCM_TRANSPORT_MASK_A2DP)) ||
+			(!ba_profile_a2dp && !(ba_pcm->transport & BA_PCM_TRANSPORT_MASK_SCO)))
 		goto stop;
 
 	/* check whether SCO has selected codec */
-	if (ba_pcm->flags & BA_PCM_FLAG_PROFILE_SCO &&
+	if (ba_pcm->transport & BA_PCM_TRANSPORT_MASK_SCO &&
 			ba_pcm->sampling == 0) {
 		debug("Skipping SCO with codec not selected");
 		goto stop;
@@ -578,40 +754,51 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 	const char *signal = dbus_message_get_member(message);
 
 	DBusMessageIter iter;
-	struct ba_pcm *pcm;
+	struct pcm_worker *worker;
 
-	if (strcmp(interface, BLUEALSA_INTERFACE_MANAGER) == 0) {
+	if (strcmp(interface, DBUS_INTERFACE_OBJECT_MANAGER) == 0) {
 
-		if (strcmp(signal, "PCMAdded") == 0) {
+		if (strcmp(signal, "InterfacesAdded") == 0) {
+			if (!dbus_message_iter_init(message, &iter))
+				goto fail;
+			struct ba_pcm pcm;
+			DBusError err = DBUS_ERROR_INIT;
+			if (!bluealsa_dbus_message_iter_get_pcm(&iter, &err, &pcm)) {
+				error("Couldn't add new PCM: %s", err.message);
+				dbus_error_free(&err);
+				goto fail;
+			}
+			if (pcm.transport == BA_PCM_TRANSPORT_NONE)
+				goto fail;
 			struct ba_pcm *tmp = ba_pcms;
 			if ((ba_pcms = realloc(ba_pcms, (ba_pcms_count + 1) * sizeof(*ba_pcms))) == NULL) {
 				error("Couldn't add new PCM: %s", strerror(ENOMEM));
 				ba_pcms = tmp;
 				goto fail;
 			}
-			if (!dbus_message_iter_init(message, &iter) ||
-					!bluealsa_dbus_message_iter_get_pcm(&iter, NULL, &ba_pcms[ba_pcms_count])) {
-				error("Couldn't add new PCM: %s", "Invalid signal signature");
-				goto fail;
-			}
-			supervise_pcm_worker(&ba_pcms[ba_pcms_count++]);
+			memcpy(&ba_pcms[ba_pcms_count++], &pcm, sizeof(*ba_pcms));
+			supervise_pcm_worker(&pcm);
 			return DBUS_HANDLER_RESULT_HANDLED;
 		}
 
-		if (strcmp(signal, "PCMRemoved") == 0) {
+		if (strcmp(signal, "InterfacesRemoved") == 0) {
 			if (!dbus_message_iter_init(message, &iter) ||
 					dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_OBJECT_PATH) {
 				error("Couldn't remove PCM: %s", "Invalid signal signature");
 				goto fail;
 			}
 			dbus_message_iter_get_basic(&iter, &path);
-			supervise_pcm_worker_stop(get_ba_pcm(path));
+			struct ba_pcm *pcm;
+			if ((pcm = get_ba_pcm(path)) == NULL)
+				goto fail;
+			supervise_pcm_worker_stop(pcm);
 			return DBUS_HANDLER_RESULT_HANDLED;
 		}
 
 	}
 
 	if (strcmp(interface, DBUS_INTERFACE_PROPERTIES) == 0) {
+		struct ba_pcm *pcm;
 		if ((pcm = get_ba_pcm(path)) == NULL)
 			goto fail;
 		if (!dbus_message_iter_init(message, &iter) ||
@@ -623,7 +810,8 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 		dbus_message_iter_next(&iter);
 		if (!bluealsa_dbus_message_iter_get_pcm_props(&iter, NULL, pcm))
 			goto fail;
-		supervise_pcm_worker(pcm);
+		if ((worker = supervise_pcm_worker(pcm)) != NULL)
+			pcm_worker_mixer_volume_update(worker, pcm);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 
@@ -634,7 +822,7 @@ fail:
 int main(int argc, char *argv[]) {
 
 	int opt;
-	const char *opts = "hVvlLB:D:";
+	const char *opts = "hVvlLB:D:M:";
 	const struct option longopts[] = {
 		{ "help", no_argument, NULL, 'h' },
 		{ "version", no_argument, NULL, 'V' },
@@ -645,6 +833,9 @@ int main(int argc, char *argv[]) {
 		{ "pcm", required_argument, NULL, 'D' },
 		{ "pcm-buffer-time", required_argument, NULL, 3 },
 		{ "pcm-period-time", required_argument, NULL, 4 },
+		{ "mixer-device", required_argument, NULL, 'M' },
+		{ "mixer-name", required_argument, NULL, 6 },
+		{ "mixer-index", required_argument, NULL, 7 },
 		{ "profile-a2dp", no_argument, NULL, 1 },
 		{ "profile-sco", no_argument, NULL, 2 },
 		{ "single-audio", no_argument, NULL, 5 },
@@ -657,18 +848,21 @@ int main(int argc, char *argv[]) {
 			printf("Usage:\n"
 					"  %s [OPTION]... [BT-ADDR]...\n"
 					"\nOptions:\n"
-					"  -h, --help\t\tprint this help and exit\n"
-					"  -V, --version\t\tprint version and exit\n"
-					"  -v, --verbose\t\tmake output more verbose\n"
-					"  -l, --list-devices\tlist available BT audio devices\n"
-					"  -L, --list-pcms\tlist available BT audio PCMs\n"
-					"  -B, --dbus=NAME\tBlueALSA service name suffix\n"
-					"  -D, --pcm=NAME\tplayback PCM device to use\n"
-					"  --pcm-buffer-time=INT\tplayback PCM buffer time\n"
-					"  --pcm-period-time=INT\tplayback PCM period time\n"
-					"  --profile-a2dp\tuse A2DP profile (default)\n"
-					"  --profile-sco\t\tuse SCO profile\n"
-					"  --single-audio\tsingle audio mode\n"
+					"  -h, --help\t\t\tprint this help and exit\n"
+					"  -V, --version\t\t\tprint version and exit\n"
+					"  -v, --verbose\t\t\tmake output more verbose\n"
+					"  -l, --list-devices\t\tlist available BT audio devices\n"
+					"  -L, --list-pcms\t\tlist available BT audio PCMs\n"
+					"  -B, --dbus=NAME\t\tBlueALSA service name suffix\n"
+					"  -D, --pcm=NAME\t\tplayback PCM device to use\n"
+					"  --pcm-buffer-time=INT\t\tplayback PCM buffer time\n"
+					"  --pcm-period-time=INT\t\tplayback PCM period time\n"
+					"  -M, --mixer-device=NAME\tmixer device to use\n"
+					"  --mixer-name=NAME\t\tmixer element name\n"
+					"  --mixer-index=NUM\t\tmixer element channel index\n"
+					"  --profile-a2dp\t\tuse A2DP profile (default)\n"
+					"  --profile-sco\t\t\tuse SCO profile\n"
+					"  --single-audio\t\tsingle audio mode\n"
 					"\nNote:\n"
 					"If one wants to receive audio from more than one Bluetooth device, it is\n"
 					"possible to specify more than one MAC address. By specifying any/empty MAC\n"
@@ -694,10 +888,31 @@ int main(int argc, char *argv[]) {
 
 		case 'B' /* --dbus=NAME */ :
 			snprintf(dbus_ba_service, sizeof(dbus_ba_service), BLUEALSA_SERVICE ".%s", optarg);
+			if (!dbus_validate_bus_name(dbus_ba_service, NULL)) {
+				error("Invalid BlueALSA D-Bus service name: %s", dbus_ba_service);
+				return EXIT_FAILURE;
+			}
 			break;
+
 		case 'D' /* --pcm=NAME */ :
 			pcm_device = optarg;
 			break;
+		case 3 /* --pcm-buffer-time=INT */ :
+			pcm_buffer_time = atoi(optarg);
+			break;
+		case 4 /* --pcm-period-time=INT */ :
+			pcm_period_time = atoi(optarg);
+			break;
+
+		case 'M' /* --mixer-device=NAME */ :
+			mixer_device = optarg;
+			break;
+		case 6 /* --mixer-name=NAME */ :
+			mixer_elem_name = optarg;
+			break;
+		case 7 /* --mixer-index=NUM */ :
+			mixer_elem_index = atoi(optarg);
+			break;
 
 		case 1 /* --profile-a2dp */ :
 			ba_profile_a2dp = true;
@@ -706,13 +921,6 @@ int main(int argc, char *argv[]) {
 			ba_profile_a2dp = false;
 			break;
 
-		case 3 /* --pcm-buffer-time=INT */ :
-			pcm_buffer_time = atoi(optarg);
-			break;
-		case 4 /* --pcm-period-time=INT */ :
-			pcm_period_time = atoi(optarg);
-			break;
-
 		case 5 /* --single-audio */ :
 			pcm_mixer = false;
 			break;
@@ -768,10 +976,13 @@ int main(int argc, char *argv[]) {
 				"  PCM device: %s\n"
 				"  PCM buffer time: %u us\n"
 				"  PCM period time: %u us\n"
+				"  ALSA mixer device: %s\n"
+				"  ALSA mixer element: '%s',%u\n"
 				"  Bluetooth device(s): %s\n"
 				"  Profile: %s\n",
 				dbus_ba_service,
 				pcm_device, pcm_buffer_time, pcm_period_time,
+				mixer_device, mixer_elem_name, mixer_elem_index,
 				ba_addr_any ? "ANY" : &ba_str[2],
 				ba_profile_a2dp ? "A2DP" : "SCO");
 
@@ -779,9 +990,11 @@ int main(int argc, char *argv[]) {
 	}
 
 	bluealsa_dbus_connection_signal_match_add(&dbus_ctx,
-			dbus_ba_service, NULL, BLUEALSA_INTERFACE_MANAGER, "PCMAdded", NULL);
+			dbus_ba_service, NULL, DBUS_INTERFACE_OBJECT_MANAGER, "InterfacesAdded",
+			"path_namespace='/org/bluealsa'");
 	bluealsa_dbus_connection_signal_match_add(&dbus_ctx,
-			dbus_ba_service, NULL, BLUEALSA_INTERFACE_MANAGER, "PCMRemoved", NULL);
+			dbus_ba_service, NULL, DBUS_INTERFACE_OBJECT_MANAGER, "InterfacesRemoved",
+			"path_namespace='/org/bluealsa'");
 	bluealsa_dbus_connection_signal_match_add(&dbus_ctx,
 			dbus_ba_service, NULL, DBUS_INTERFACE_PROPERTIES, "PropertiesChanged",
 			"arg0='"BLUEALSA_INTERFACE_PCM"'");
